/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* The main program of the curses version of xconq. */

#include "conq.h"
#include "cconq.h"

int announced = FALSE;

char *announcemsg = NULL;

Side *dside = NULL;

UI dui;

/* The main program just runs through everything. */

main(argc, argv)
int argc;
char *argv[];
{
    int timeout;

    printf("\n              Welcome to Curses Xconq version %s\n\n",
	   version_string());
    init_library_path();
    print_any_news();
    clear_game_modules();
    init_data_structures();
    parse_command_line(argc, argv, 1);
    load_all_modules();
    check_game_validity();
    parse_command_line(argc, argv, 2);
    set_variants_from_options();
    parse_command_line(argc, argv, 3);
    set_players_from_options();
    make_trial_assignments();
    run_synth_methods();
    final_init();
    assign_players_to_sides();
    /* Run through game calcs, but don't let anything move yet. */
    run_game(0);
    init_display();
    init_interaction();
    redraw();
    init_sighandlers();
    ingame = TRUE;
    while (1) {
	timeout = run_game(10000);  /* assumes no real-timeness */
	maybe_handle_input(timeout);
    }
}

/* An init error needs to have the command re-run. */

init_error(str, a1, a2, a3)
char *str;
long a1, a2, a3;
{
    printf("\nError: ");
    printf(str, a1, a2, a3);
    printf("\n");
    fflush(stdout);
    exit(1);
}

/* A warning just gets displayed, no other action is taken. */

init_warning(str, a1, a2, a3)
char *str;
long a1, a2, a3;
{
    printf("Warning: ");
    printf(str, a1, a2, a3);
    printf("\n");
    fflush(stdout);
}

/* Nonempty display name not actually used, but needed to keep things */
/* straight. */

Player *
add_default_player()
{
    Player *player = add_player();

    /* (could try to pick up actual user name?) */
    player->name = getenv("USER");
    player->displayname = "ttyXX";
    return player;
}

/* Set up the basic user interface for a side. */

init_ui(side)
Side *side;
{
    if (side_wants_display(side)) {
	if (dside == NULL) {
	    side->ui = (UI *) xmalloc(sizeof(UI));
	    side->ui->active = FALSE;
	    dui.active = FALSE;
	    dui.followaction = 0;
	    dui.curhelpnode = NULL;
	    dui.mode = SURVEY;
	    dui.screenmode = normalwins;
	    dui.cmdarg = DFLTFLAG;
	    dui.nextwin = 0;
	    dside = side;
	} else {
	    /* an error */
	}
    }
}

/* Open display, create all the windows we'll need, do misc setup things,
   and initialize some globals to out-of-range values for recognition later. */

init_display()
{
    int i, s, n, x, y;
    int r, sx, sy;
    int rightwidth;

    if (dside == NULL) {
	fprintf(stderr, "No side wanted a display!\n");
	exit(1);
    }
    Dprintf("Will try to open screen as display...\n");
    /* Go through the standard curses setup sequence. */
    initscr();
    nonl();
    noecho();
    cbreak();
    clear();
    dside->ui->active = TRUE;
    dui.active = TRUE;
    /* Set up things shared by all the windows. */
    dui.nw = min(BUFSIZE, 60);
    dui.nh = 1;
    dui.text1[0] = '\0';
    dui.text2[0] = '\0';
    /* Compute the division of the screen. */
    dui.mw = (2 * COLS) / 3;
    dui.mh = LINES - 2 - 4;
    dui.totw = area.width * 2;
    if (!area.xwrap) dui.totw += (area.height / 2);
    dui.toth = area.height;
    dui.sh = numsides + 1;
    dui.lw = COLS - dui.mw - 1;
    dui.lh = LINES - dui.sh - 2;
    pick_a_focus(dside, &x, &y);
    dui.vcx = x;  dui.vcy = y;
    dui.lastvcx = dui.lastvcy = -1;
    /* Set default values for the display controls. */
    dui.drawterrain = TRUE;
    dui.drawunits = TRUE;
    dui.drawnames = FALSE;  /* they clutter up the screen */
    dui.drawweather = FALSE;
    dui.unit = NULL;
    dui.listtype = byrandorder;
    dui.listsides = allsides;
    dui.curunit = NULL;
    dui.curside = 0;
    dui.toplines = create_window(0, 0, dui.nw, 2);
    dui.date = create_window(dui.nw, 0, COLS - dui.nw, 2); 
    dui.closeup = create_window(0, 2, dui.mw, 4);
    dui.map = create_window(0, 2 + 4, dui.mw, dui.mh);
    dui.sides = create_window(dui.mw + 1, 2, dui.lw, dui.sh);
    dui.list = create_window(dui.mw + 1, 2 + dui.sh, dui.lw, dui.lh);
    /* Help window covers the screen usually. */
    dui.help = create_window(0, 0, COLS, LINES);
    set_scroll();
    Dprintf("Successfully opened display!\n");
}

init_interaction()
{
    Unit *unit;

    /* Look for and focus on an initial unit. */
    if ((unit = find_next_actor(dside, NULL)) != NULL) {
	make_current(unit);
	dui.mode = MOVE;
	show_cursor();
    } else {
	show_cursor();
	move_survey(dui.vcx, dui.vcy);
    }
}

set_scroll()
{
    if (dui.mw >= dui.totw) {
	dui.sx = (area.xwrap ? 0 : ((area.height / 2) * 2) / 2);
    }
    if (dui.mh >= dui.toth) {
	dui.sy = 0;
    }
}

set_map_viewport()
{
    /* Compute the size of the viewport. */
    dui.vw = min(area.width, dui.mw / 2 + 1);
    dui.vh = min(area.height, dui.mh + 1);
    /* Compute the bottom visible row. */
    dui.vy = (dui.toth - dui.sy) - dui.vh;
    /* Adjust to keep its value from being outside the area. */
    dui.vy = max(0, min(dui.vy, area.height - dui.vh));
    /* Compute the first visible column. */
    dui.vx = dui.sx / 2 - dui.vy / 2 - 1;
    DGprintf("Set %dx%d viewport at %d,%d\n", dui.vw, dui.vh, dui.vx, dui.vy);
}

center_on_focus()
{
    dui.sx = dui.vcx * 2 + (dui.vcy * 2) / 2 - dui.mw / 2 + 1;
    dui.sy = dui.toth - (dui.vcy + dui.mh / 2);
    dui.sx = max(0, min(dui.sx, dui.totw - dui.mw));
    dui.sy = max(0, min(dui.sy, dui.toth - dui.mh));
}

/* Windows in curses are just simple bounding boxes indexed by small
   integers. */

create_window(x, y, w, h)
int x, y, w, h;
{
    if (x + w > COLS) w = COLS - x;
    if (y + h > LINES) h = LINES - y;
    DGprintf("Creating %dx%d window at %d,%d\n", w, h, x, y);
    dui.wx[dui.nextwin] = x;  dui.wy[dui.nextwin] = y;
    dui.ww[dui.nextwin] = w;  dui.wh[dui.nextwin] = h;
    return (dui.nextwin++);
}

/* A run error is fatal. */

run_error(str, a1, a2, a3)
char *str;
long a1, a2, a3;
{
    close_displays();
    /* (should save the game or at least ask about it) */
    printf("\n");
    printf(str, a1, a2, a3);
    printf("\n");
    printf("Saving the game...");
    write_entire_game_state(savefile_name());
    printf(" done.\n");
    exit(1);
}

/* Runtime warnings are for when it's important to bug the player. */

run_warning(str, a1, a2, a3)
char *str;
long a1, a2, a3;
{
    /* (should put in topline and ask what to do) */
    printf("\n");
    printf(str, a1, a2, a3);
    printf("\n");
    fflush(stdout);
}

wait_for_char()
{
    show_cursor();
    dui.inpch = getch() & 0177;
    return dui.inpch;
}

/* Only kind of input is keystrokes. */

maybe_handle_input(timeout)
int timeout;
{
    char ch;

    if (active_display(dside)) {
	/* Select this or the next moving unit. (a hack) */
	if (dui.mode == MOVE) {
	    if (dui.curunit == NULL
		|| (dui.curunit->act && dui.curunit->act->acp <= 0)) {
		make_current(find_next_actor(dside, dui.curunit));
	    } else {
		make_current(dui.curunit);
	    }
	}
	show_cursor();
	dui.inpch = getch() & 0177;
	dui.gotkey = TRUE;
	/* Program will not work without following due to compiler bug. (?) */
	Dprintf("key is %c", dui.inpch); 
	/* Interpret the input. */
	interpret_input();
    }
    /* (Needs more work I think) */
    if (g_rt_per_side_max() > 0
	&& !side_lost(dside)
	&& dside->totaltimeused >= g_rt_per_side_max()) {
	side_loses(dside);
	notify("You ran out of time!!");
	beep();
    }
}

/* ^A does a "transparent redraw" */

#define REDRAW '\001'

/* The traditional direction characters. */

char dirchars[] = "ulnbhy";

interpret_input()
{
    int dir, terr;

    if (dui.inpch == REDRAW) {
	/* redraw everything but leave the interaction state unchanged */
	redraw();
    } else if (dui.mode == HELP) {
	interpret_help();
    } else if (isdigit(dui.inpch)) {
	if (dui.cmdarg == DFLTFLAG) {
	    dui.cmdarg = 0;
	} else {
	    dui.cmdarg *= 10;
	}
	dui.cmdarg += (dui.inpch - '0');
	sprintf(dui.text1, "Arg: %d", dui.cmdarg);
	show_toplines();
	return;
    } else {
	clear_toplines();
	if ((dir = iindex(dui.inpch, dirchars)) >= 0) {
	    if (dui.cmdarg == DFLTFLAG) dui.cmdarg = 1;
	    do_dir(dir, dui.cmdarg);
	} else if ((dir = iindex(lowercase(dui.inpch), dirchars)) >= 0) {
	    if (dui.cmdarg == DFLTFLAG) dui.cmdarg = dside->itertime;
	    if (dui.mode == SURVEY) dui.cmdarg = 10;
	    do_dir(dir, dui.cmdarg);
	} else {
	    execute_command();
	}
	/* Reset the arg so we don't get confused next time around */
	dui.cmdarg = DFLTFLAG;
    }
}

do_dir(dir, n)
int dir, n;
{
    int nx = wrap(dui.curx + n * dirx[dir]);
    int ny = dui.cury + n * diry[dir];
    Unit *unit2;

    if (dui.mode == SURVEY) {
	move_survey(nx, ny);
    } else if (dui.curunit == NULL) {
	beep(); /* or find a unit? */
    } else if ((unit2 = unit_at(nx, ny)) != NULL) {
	if (unit2->side != dui.curunit->side) {
	    prep_attack_action(dui.curunit, unit2, 100);
	} else if (can_occupy_cell(dui.curunit, nx, ny)) {
	    prep_move_action(dui.curunit, nx, ny, 0, dui.curunit);
	} else {
	    prep_enter_action(dui.curunit, dui.curunit, unit2);
	}
    } else {
	order_moveto(dui.curunit, nx, ny);
    }
}

move_survey(nx, ny)
int nx, ny;
{
    Unit *unit;

    if (inside_area(nx, ny)) {
	dui.curx = nx;  dui.cury = ny;
	dui.curunit = NULL;
	for_all_stack(nx, ny, unit) {
	    if (unit->side == dside) {
		make_current(unit);
		break;
	    }
	}
	put_on_screen(dui.curx, dui.cury);
    } else {
	beep();
    }
}

/* Ensure that given location is visible. */

put_on_screen(x, y)
int x, y;
{
    if (active_display(dside)) {
	if (!in_middle(x, y)) {
	    recenter(x, y);
	}
	show_closeup();
    }
}

/* Put the point x, y in the center of the map. */

recenter(x, y)
int x, y;
{
    if (!((x == dui.vcx) && (y == dui.vcy))) {
	dui.vcx = x;  dui.vcy = y;
	center_on_focus();
	set_map_viewport();
	show_map();
	show_cursor();
	refresh();
    }
}


/* Prompt for a yes/no answer with a settable default. */

ask_bool(question, dflt)
char *question;
int dflt;
{
    char ch;

    dui.prevmode = dui.mode;
    dui.mode = PROMPT;
    sprintf(dui.text1, "%s [%s]", question, (dflt ? "yn" : "ny"));
    show_toplines();
    show_cursor();
    ch = getch() & 0177;
    if (dflt ? (lowercase(ch) == 'n') : (lowercase(ch) == 'y'))
      dflt = !dflt;
    dui.mode = dui.prevmode;
    clear_toplines();
    return dflt;
}

/* Prompt for a type of a unit from player, maybe only allowing some types
   to be accepted.  Also allow specification of no unit type.  We do this
   by scanning the vector, building a string of chars and a vector of
   unit types, so as to be able to map back when done. */

ask_unit_type(prompt, possibles)
char *prompt;
short *possibles;
{
    char ch;
    int numtypes = 0, u, type;

    for_all_unit_types(u) {
	dui.bvec[u] = 0;
	if (possibles == NULL || possibles[u]) {
	    dui.bvec[u] = 1;
	    dui.ustr[numtypes] = utype_name_n(u, 1)[0];
	    dui.uvec[numtypes] = u;
	    numtypes++;
	}
    }
    if (numtypes == 0) {
	return (-1);
    } else {
	dui.prevmode = dui.mode;
	dui.mode = PROMPT;
	dui.ustr[numtypes] = '\0';
	sprintf(dui.text1, "%s [%s] ", prompt, dui.ustr);
	show_toplines();
    }
    while (1) {
	show_cursor();
	ch = (getch() & 0177);
	if (ch == ESCAPE) {
	    clear_toplines();
	    return (-1);
	}
	if (iindex(ch, dui.ustr) >= 0) {
	    clear_toplines();
	    return dui.uvec[iindex(ch, dui.ustr)];
	}
	beep();
    }
}

/* Do something with the char or unit type that the player entered. */

grok_unit_type()
{
    int i, type = NONUTYPE;

    if (dui.inpch == '?') {
	/* should do some kind of short list */
/*	help_unit_type();  */
    } else if (iindex(dui.inpch, dui.ustr) == -1) {
	notify("Unit type '%c' not in \"%s\"!",
	       dui.inpch, dui.ustr);
    } else {
	type = dui.uvec[iindex(dui.inpch, dui.ustr)];
    }
    if (type >= 0) {
	clear_toplines();
	for_all_unit_types(i) dui.bvec[i] = 0;
    }
    dui.mode = dui.prevmode;
    clear_toplines();
    return type;
}

/* Prompt for a type of a terrain from player, maybe only allowing some types
   to be accepted.  Also allow specification of no terrain type.  We do this
   by scanning the vector, building a string of chars and a vector of
   terrain types, so as to be able to map back when done. */

ask_terrain_type(prompt, possibles)
char *prompt;
short *possibles;
{
    int numtypes = 0, u, type;

    for_all_terrain_types(u) {
	dui.bvec[u] = 0;
	if (possibles == NULL || possibles[u]) {
	    dui.bvec[u] = 1;
	    dui.ustr[numtypes] = '?' /* utypes[u].uchar */;
	    dui.uvec[numtypes] = u;
	    numtypes++;
	}
    }
    if (numtypes == 0) {
	type = NONTTYPE;
    } else if (numtypes == 1) {
	type = dui.uvec[0];
	dui.bvec[type] = 0;
    } else {
	dui.ustr[numtypes] = '\0';
	sprintf(dui.text1, "%s [%s] ", prompt, dui.ustr);
	show_toplines();
	type = NONTTYPE;
    }
    return type;
}

/* Do something with the char or terrain type that the player entered. */

grok_terrain_type()
{
    int i, type = NONTTYPE;

    if (dui.inpch == '?') {
/*	help_terrain_type(); */
    } else if (dui.inpch == ESCAPE) {
	/* type is already correct */
    } else if (iindex(dui.inpch, dui.ustr) == -1) {
	notify("Terrain type '%c' not in \"%s\"!",
	       dui.inpch, dui.ustr);
    } else {
	type = 0 /* find_terrain_char(dui.inpch) */;
    }
    if (type >= 0) {
	clear_toplines();
	for_all_terrain_types(i) dui.bvec[i] = 0;
    }
    return type;
}

/* User is asked to pick a position on map.  This will iterate until the */
/* space bar designates the final position. */

ask_position(prompt, xp, yp)
char *prompt;
int *xp, *yp;
{
    char ch;
    int dir, nx, ny;

    sprintf(dui.text1, "%s [dir keys to move, space bar to set] - ",
	    prompt);
    show_toplines();
    save_cur();
    while (1) {
	show_cursor();
	ch = getch() & 0177;
	if (ch == ' ') {
	    *xp = dui.curx;  *yp = dui.cury;
	    restore_cur();
	    return TRUE;
	}
	if (ch == ESCAPE) {
	    restore_cur();
	    return FALSE;
	}
	if ((dir = iindex(ch, dirchars)) >= 0) {
	    point_in_dir(dui.curx, dui.cury, dir, &nx, &ny);
	    if (inside_area(nx, ny)) {
		dui.curx = nx;  dui.cury = ny;
		put_on_screen(dui.curx, dui.cury);
	    } else {
		beep();
	    }
	} else {
	    beep();
	}
    }
}

/* Save away the currently selected position. */

save_cur()
{
    dui.tmpcurx = dui.curx;  dui.tmpcury = dui.cury;
    dui.tmpcurunit = dui.curunit;
}

/* Restore the saved "cur" slots. */

restore_cur()
{
    dui.curx = dui.tmpcurx;  dui.cury = dui.tmpcury;
    dui.curunit = dui.tmpcurunit;
}

/* Read a string from the prompt window.  Deletion is allowed, and a
   cursor is displayed. */

char *
ask_string(prompt, dflt)
char *prompt, *dflt;
{
    char ch;

    sprintf(dui.text1, "%s ", prompt);
    dui.reqstrbeg = strlen(dui.text1);
    if (dflt != NULL) {
	strcat(dui.text1, dflt);
    }
    dui.reqstrend = strlen(dui.text1);
    dui.prevmode = dui.mode;
    dui.mode = PROMPT;
    show_toplines();
    while (1) {
	show_cursor();
	ch = getch() & 0177;
	switch (ch) {
	  case '\r':
	  case '\n':
	    /* We're done, put everything back to normal. */
	    dui.mode = dui.prevmode;
	    clear_toplines();
	    show_cursor();
	    /* Return a copy of the part of the buffer that was typed into. */
	    return copy_string(dui.text1 + dui.reqstrbeg);
	  case ESCAPE:
	    beep();
	    dui.mode = dui.prevmode;
	    clear_toplines();
	    show_cursor();
	    return NULL;
	  case BACKSPACE:
	  case 0x7f:
	    if (dui.reqstrend > dui.reqstrbeg) {
		--dui.reqstrend;
	    } else {
		beep();
	    }
	    break;
	  default:
	    if (dui.reqstrend < BUFSIZE-2) {
		/* Insert the character. */
		(dui.text1)[dui.reqstrend++] = ch;
	    } else {
		beep();
	    }
	}
	/* Always make sure we're terminated. */
	(dui.text1)[dui.reqstrend] = '\0';
	show_toplines();
    }
}

ask_side(prompt, dflt)
Side *dflt;
char *prompt;
{
    sprintf(dui.text1, "%s ", prompt);
    dui.reqstrbeg = strlen(dui.text1);
    if (dflt != NULL) {
	strcat(dui.text1, dflt->name);
    }
    dui.reqstrend = strlen(dui.text1);
    dui.prevmode = dui.mode;
    dui.mode = PROMPT;
    show_toplines();
    show_cursor();
}

Side *
grok_side()
{
    return NULL;
}

order_movetounit() {}

make_current(unit)
Unit *unit;
{
    if (unit != NULL) {
	dui.curx = unit->x;  dui.cury = unit->y;
	dui.curunit = unit;
	show_closeup();
    }
    show_cursor();
}

/* Add 'P' for printing/saving to file (which?). */
/* (Need individual screen scrolling also) */
/* (Be able to put *all* help info into a single file?) */

interpret_help()
{
    HelpNode *prevhelpnode = dui.curhelpnode;

    switch (dui.inpch) {
      case ' ':
	break;
      case ESCAPE:
      case 'q':
      case 'Q':
	dui.screenmode = normalwins;
	dui.mode = SURVEY;
	redraw();
	break;
      case 'n':
	dui.curhelpnode = dui.curhelpnode->next;
	break;
      case 'p':
	dui.curhelpnode = dui.curhelpnode->prev;
	break;
      case 'b':
	break;
      case 't':
	dui.curhelpnode = firsthelpnode;
	break;
      default:
	beep();
	break;
    }
    if (prevhelpnode != dui.curhelpnode) {
	dui.helpstring = get_help_text(dui.curhelpnode);
	show_help();
    }
}

/* Reading is usually pretty fast, so don't do anything here. */

announce_read_progress() {}

/* Announce the start of a time-consuming computation. */

announce_lengthy_process(msg)
char *msg;
{
    if (n_seconds_elapsed(1)) {
	printf("%s; ", msg);
	fflush(stdout);
	announced = TRUE;
    } else {
	announcemsg = msg;  /* had better be a static string... */
    }
}

announce_progress(percentdone)
int percentdone;
{
    if (n_seconds_elapsed(1)) {
	if (announcemsg) {
	    printf("%s; ", announcemsg);
	    announcemsg = NULL;
	}
	if (percentdone > 0) {
	    printf(" %d%%", percentdone);
	    fflush(stdout);
	    announced = TRUE;
	}
    }
}

finish_lengthy_process()
{
    if (announced) {
	printf("; done.\n");
	announced = FALSE;
    }
}

fatal_error(clue)
char *clue;
{
    abort();
}

/* This is the basic routine to get out of the program - we assume any
   game hacking has been taken care of already. */

exit_cconq(side)
Side *side;
{
    close_displays();
    /* Leave any commentary on the console. */
    printf("%s\n", exit_commentary(side));
    exit(0);
}

show_clocks() {}

flush_display_buffers(side)
Side *side;
{
    if (active_display(side)) {
	refresh();
    }
}

abort_init()
{
    fprintf(stderr, "Fatal errors encountered during game setup, exiting\n");
    exit(1);
}

/* Draw an individual detailed hex, as a row of one, on all maps. */

update_cell_display(side, x, y, flushit)
Side *side;
int x, y;
int flushit;
{
    if (active_display(side)) {
	draw_row(x, y, 1);
	if (flushit) refresh();
    }
}

/* Curses updates efficiently, so don't need special routine. */

update_side_display(side, side2)
Side *side, *side2;
{
    if (active_display(side)) {
	show_side_list();
	refresh();
    }
}

/* Just change the part of the unit list relevant to the given unit. */

update_unit_display(side, unit)
Side *side;
Unit *unit;
{
    if (active_display(side) && unit != NULL) {
	if (dui.curunit == unit) {
	    show_closeup();
	    show_map();
	}
	show_list();
	refresh();
    }
}

update_turn_display(side, rightnow)
Side *side;
int rightnow;
{
    if (active_display(side)) {
	show_game_date();
	show_cursor();
	refresh();
    }
}

update_action_display(side, rightnow)
Side *side;
int rightnow;
{
    if (active_display(side)) {
	/* what to do? */
    }
}

update_action_result_display(side, unit, rslt, rightnow)
Side *side;
Unit *unit;
int rslt, rightnow;
{
    extern HevtDefn hevtdefns[];

    if (active_display(side)) {
	notify("%s %s %s!", unit_desig(unit),
	       action_desig(&(unit->act->nextaction)), hevtdefns[rslt].name);
    }
}

update_event_display()
{
}

update_fire_at_display()
{
}

/* Updates to clock need to be sure that display changes immediately. */

update_clock_display(side, rightnow)
Side *side;
int rightnow;
{
    if (active_display(side)) {
	show_clocks();
	show_cursor();
	refresh();
    }
}

update_everything()
{
    redraw();
}

/* Send a message to everybody who's got a screen. */

/*VARARGS*/
notify_all(control, a1, a2, a3, a4, a5, a6)
char *control, *a1, *a2, *a3, *a4, *a5, *a6;
{
    if (active_display(dside)) {
	notify(control, a1, a2, a3, a4, a5, a6);
    }
}


close_displays()
{
    Side *side;

    for_all_sides(side) if (active_display(side)) close_display(side);
}
