(game-module "ww2-div-pac"
  (title "ww2-division-pacific")
  (blurb "Defns for divisional (theater) level WWII in the Pacific")
)

(unit-type a-inf (image-name "inf-div") (possible-sides "allied"))
(unit-type a-armor (image-name "armor-div") (possible-sides "allied"))

(unit-type j-inf (image-name "inf-div") (possible-sides "japanese"))
(unit-type j-armor (image-name "armor-div") (possible-sides "japanese"))

(unit-type para (image-name "para-div") (possible-sides "allied"))
(unit-type hq (image-name "hq"))

(define ground-types (a-inf a-armor j-inf j-armor para hq))

(unit-type aa (image-name "ap"))
(unit-type dd (image-name "dd"))
(unit-type ss (image-name "sub"))
(unit-type cl (image-name "ca"))
(unit-type ca (image-name "ca"))
(unit-type cve (image-name "cv"))
(unit-type cvl (image-name "cv"))
(unit-type cvs (image-name "cv"))
(unit-type cv (image-name "cv"))
(unit-type bc (image-name "bb"))
(unit-type bb (image-name "bb"))

(define small-ship-types (aa dd ss))

(define large-ship-types (cl ca cve))

(define capital-ship-types (cvl cvs cv bc bb))

(define ship-types (aa dd ss cl ca cve cvl cvs cv bc bb))

(unit-type airfield (image-name "airbase"))
(unit-type airbase)
(unit-type port)
(unit-type town (image-name "town20"))
(unit-type city (image-name "city20"))

(define place-types (airfield airbase port town city))

;(material-type organization
;  (help "important to do things, acquired from HQs"))
(material-type aircraft
  (help "attached aircraft used for support"))

(terrain-type sea)
(terrain-type atoll)
(terrain-type clear)
(terrain-type jungle (image-name "forest"))
(terrain-type hills)
(terrain-type mountains)
(terrain-type road (subtype connection))
(terrain-type reef (subtype border) (color "gray"))
(terrain-type river (subtype border))

(define water (sea atoll))
(define land (clear jungle mountains))

(add water elevation-min 0)

(add water elevation-max 0)

;;; Static relationships.

(table vanishes-on
  (ground-types sea true)
  (ship-types land true)
  (place-types sea true)
  ((town city) atoll true)
  )

(add sea capacity 100)

(add land capacity 4)

(add aa capacity 1)

(table unit-size-as-occupant
  (u* aa 99)
  (ground-types aa 1)
  )

;;; Action parameters.

(add ground-types acp-per-turn 4)
(add ship-types acp-per-turn 60)

;;; Movement.

(add ground-types speed 4)
(add ship-types speed 60)

(table mp-to-enter-terrain
  (ground-types sea 99)
  (ship-types land 99)
  (ship-types atoll 5)
  )

(table mp-to-leave-terrain
  (ship-types atoll 5)
  )

;;; Combat.

(add ground-types hp-max    12)
(add ground-types parts-max 12)

;aa dd ss cl ca cve cvl cvs cv bc bb

(add ship-types hp-max    (6 6 6 4 6 4 4 4 5 5 6))
(add ship-types parts-max (6 6 6 2 2 2 1 1 1 1 1))

(add place-types hp-max (10 20 20 40 80))

(table acp-to-attack
  (ship-types u* 10)
  )

(table acp-to-defend
  (u* ship-types 10)
  )

(table hit-chance
  (ship-types ship-types 50)
  )

(table damage
  (ship-types ship-types 1)
  )

(table capture-chance
  (ground-types airfield 100)
  (ground-types (airbase port) 90)
  (ground-types (town city) 50)
  )

;;; Defaults for testing.

;;; Some defns for the fractal percentile generator.

(set alt-blob-density 10000)
(set alt-blob-size 20)
(set alt-smoothing 6)
(set wet-blob-density 2000)
(set wet-blob-size 100)

(add t* alt-percentile-min (  0  90  90  90  97  98 0 0 0))
(add t* alt-percentile-max ( 90  91  97  97  98 100 0 0 0))
(add t* wet-percentile-min (  0   0   0  50   0   0 0 0 0))
(add t* wet-percentile-max (100 100  50 100 100 100 0 0 0))

;;; River generation.

;(add (plains forest mountains) river-chance (5000 10000 10000))

;(set river-sink-terrain shallows)

;(table terrain-interaction (river (sea shallows) non-terrain))

;(add (sea shallows ice) extensions ((water true)))

(set edge-terrain sea)

(add u* start-with 1)  ; one of everything

(set country-radius-min 4)
(set country-separation-min 20)
(set country-separation-max 60)

(add (sea clear) country-terrain-min (10 5))

(table favored-terrain
  (u* t* 0)
  (ground-types land 100)
  (ship-types sea 100)
  (place-types land 100)
  )

(side-library set
  ((name "Allies") (adjective "Allied") (emblem-name "flag-usa")
   (class "allied") (names-locked true))
  ((name "Japan") (adjective "Japanese") (emblem-name "flag-japan")
   (class "japanese") (names-locked true))
)

(include "ng-ships")

;; need to make side-specific eventually
(add ca namer "american-ca-names")
(add cv namer "american-cv-names")
(add bb namer "american-bb-names")

(world 800 (year-length 52))

(game-module (design-notes (
  "The map-scale is 30 miles/hex, game time is 1 week/turn."
  "One unit of aircraft material is one plane."
  ""
  "To Do:"
  ""
  "Adapt old pacific map."
  ""
  "Could have one port and one airfield share a hex, but no more."
  ""
  "Units with aircraft should have a long vision range."
  ""
  "Add fuel as material, scenarios could predefine certain places to have lots."
  )))
