/* Copyright (c) 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Mac designer handling. */

#include "conq.h"
#include "mac.h"

#ifdef DESIGNERS

int designwinw = 120;

WindowPtr designwin = nil;

Handle menubuttonhandle = nil;

/* This is the height of each design too. */

short deltdim = 36;

/* Variables that keep track of what the designer is doing. */

enum tooltype tooltype = notool;
short curutype = 0;
short curttype = 0;
short curbgttype = 0;
short cursidenumber = 0;
short curfid = 0;
GeoFeature *curfeature = NULL;
char *curfeaturetypename;
char *curfeaturename;
short curelevation = 0;
short curbrushradius = 0;

enable_designing(forsure)
int forsure;
{
	extern int compromised;

	if (dside == NULL || dside->designer) return;
	if (!forsure && !compromised) {
		switch (CautionAlert(aConfirmDesign, nil)) {
			case aiConfirmDesignOK:
				break;
			case aiConfirmDesignCancel:
				return;
		}
	}
	/* Actually change designer status, this will call back to alter all displays. */
	become_designer(dside);
	/* Create and add the designer's palette. */
	if (designwin == nil) {
		create_design_window();
	}
	if (designwin != nil) {
		position_design_window();
		ShowWindow(designwin);
/*		SelectWindow(designwin); */
	}
}

disable_designing()
{
	if (dside == NULL || !dside->designer) return;
	/* Hide (but don't destroy) the designer's palette. */
	if (designwin != nil) {
		HideWindow(designwin);
	}
	/* Actually change designer status, this will call back to alter all displays. */
	become_nondesigner(dside);
}

CursHandle paintcursors[10];
CursHandle bordpaintor;
CursHandle connpaintor;
CursHandle coatpaintor;

init_design_cursors()
{
	paintcursors[terraintool] = GetCursor(cCell);
	bordpaintor = GetCursor(cBord);
	connpaintor = GetCursor(cConn);
/*	coatpaintor = GetCursor(cCoat);  */
	paintcursors[unittool] = GetCursor(cUnit);
	paintcursors[peopletool] = GetCursor(cPeople);
	paintcursors[materialtool] = GetCursor(cMaterial);
	paintcursors[featuretool] = GetCursor(cFeature);
	paintcursors[elevationtool] = GetCursor(cElevation);
	paintcursors[temperaturetool] = GetCursor(cTemperature);
	paintcursors[cloudstool] = GetCursor(cClouds);
	paintcursors[windstool] = GetCursor(cWinds);
}

/* Adjust the cursor for some (actually, most) of the designer tools. */

CursPtr
adjust_designer_cursor(mouse, region)
Point mouse;
RgnHandle region;
{
	Map *map;
	CursPtr curs = &qd.arrow;

	switch (tooltype) {
		case terraintool:
			switch (t_subtype(curttype)) {
				case cellsubtype:
					curs = *(paintcursors[tooltype]);
					break;
				case bordersubtype:
					curs = *bordpaintor;
					break;
				case connectionsubtype:
					curs = *connpaintor;
					break;
				case coatingsubtype:
					break;
				default:
					terrain_subtype_warning("cursor adjust", curttype);
					break;
			}
			break;
		default:
			curs = *(paintcursors[tooltype]);
			break;
	}
	return curs;
}

/* Create the designer tool window. */

ControlHandle elevmenu = nil;

create_design_window()
{
	if (hasColorQD) {
		designwin = GetNewCWindow(wDesign, NULL, (WindowPtr) -1L);
	} else {
		designwin = GetNewWindow(wDesign, NULL, (WindowPtr) -1L);
	}
	add_window_menu_item("Design", designwin);  /* until this becomes a windoid */
	/* Make the current side be the one with the display. */
	cursidenumber = side_number(dside);
	menubuttonhandle = GetNamedResource('SICN', "\pmenubutton");
/*	elevmenu = GetNewControl(211, designwin); */
	SizeWindow(designwin, designwinw, 10 * deltdim - 1, 1);
	position_design_window();
	init_design_cursors();
	ShowWindow(designwin);
}

/* Try to put the palette alongside the frontmost window. */

position_design_window()
{
	Point pt;
	WindowPtr win;
	GrafPtr oldport;

	if ((win = FrontWindow()) != nil) {
		GetPort(&oldport);
		SetPort(win);
		SetPt(&pt, win->portRect.right + 3, win->portRect.top);
		LocalToGlobal(&pt);
		SetPort(oldport);
	} else {
		SetPt(&pt, 500, 50);
	}
	/* (should make sure is not off the edge of the screen) */
	if (pt.h + designwinw > 640 /* is off screen entirely */) {
		SetPt(&pt, 500, 50);
	}
	MoveWindow(designwin, pt.h, pt.v, TRUE);
}

draw_design_window()
{
	int i;
	GrafPtr oldport;

	if (!active_display(dside) || designwin == nil) return;
	GetPort(&oldport);
	SetPort(designwin);
	for (i = 0; i < 10; ++i) {
		draw_design_window_tool(i);
	}
	DrawControls(designwin);
	SetPort(oldport);
}

draw_design_window_tool(tool)
enum tooltype tool;
{
	int enabled = TRUE, paintable = TRUE, haspopup = FALSE;
	char *toolname = NULL, *auxtoolname = NULL;
	Rect tmprect, imagerect;
	Str255 tmpstr;

	SetRect(&tmprect, 0, tool * deltdim, designwinw, (tool + 1) * deltdim);
	EraseRect(&tmprect);
	/* Confine the image to a square subrect on the left side of the window. */
	imagerect = tmprect;
	imagerect.right = deltdim;
	switch (tool) {
		case notool:
			toolname = "Normal";
			if (curbrushradius > 0) {
				sprintf(spbuf, "%d", curbrushradius);
				c2p(spbuf, tmpstr);
				MoveTo(4, tmprect.top + deltdim / 2 + 5);
				DrawString(tmpstr);
			}
			break;
		case terraintool:
			toolname = t_type_name(curttype);
			InsetRect(&imagerect, (deltdim - hws[4]) / 2, (deltdim - hhs[4]) / 2);
			/* Only do bg terrain type if painting cell terrain. */
			if (t_subtype(curttype) == cellsubtype) {
				auxtoolname = t_type_name(curbgttype);
				OffsetRect(&imagerect, 3, 3);
				draw_terrain_sample(imagerect, curbgttype);
				OffsetRect(&imagerect, -6, -6);
			}
			draw_terrain_sample(imagerect, curttype);
			haspopup = TRUE;
			break;
		case unittool:
			toolname = u_type_name(curutype);
			InsetRect(&imagerect, (deltdim - 32) / 2, (deltdim - 32) / 2);
		 	draw_unit_image(designwin, imagerect.left, imagerect.top, 32, 32,
		 					curutype, side_n(cursidenumber), 0);
			/* Gray out the unit if not allowed for the current side. */
			paintable = type_allowed_on_side(curutype, side_n(cursidenumber));
			haspopup = TRUE;
			break;
		case peopletool:
			shortest_side_title(side_n(cursidenumber), spbuf);
			toolname = spbuf;
			InsetRect(&imagerect, (deltdim - 16) / 2, (deltdim - 16) / 2);
			draw_side_emblem(designwin, imagerect.left, imagerect.top, 16, 16,
							 side_n(cursidenumber));
			haspopup = TRUE;
			break;
		case materialtool:
			if (nummtypes > 0 /* and terrain capacities > 0 */) {
			} else {
				toolname = "Material";
				enabled = FALSE;
			}
			break;
		case featuretool:
			if (features_defined()) {
				if (curfeature == NULL) {
					curfeature = (GeoFeature *) find_geo_feature(curfid);
				}
				if (curfeature != NULL) {
					toolname = curfeature->typename;
					auxtoolname = curfeature->name;
				}
				sprintf(spbuf, "#%d", curfid);
				c2p(spbuf, tmpstr);
				MoveTo(4, tmprect.top + deltdim / 2 + 5);
				DrawString(tmpstr);
			} else {
				toolname = "Feature";
				enabled = FALSE;
			}
			break;
		case elevationtool:
			if (!world_is_flat()) {
				sprintf(spbuf, "%d", curelevation);
				c2p(spbuf, tmpstr);
				MoveTo(4, tmprect.top + deltdim / 2 + 5);
				DrawString(tmpstr);
			} else {
				toolname = "Elevation";
				enabled = FALSE;
			}
			break;
		case temperaturetool:
			toolname = "Temperature";
			enabled = FALSE;
			break;
		case cloudstool:
			toolname = "Clouds";
			enabled = FALSE;
			break;
		case windstool:
			toolname = "Winds";
			enabled = FALSE;
			break;
		default:
			/* ??? */
			break;
	}
	if (toolname != NULL) {
		c2p(toolname, tmpstr);
		MoveTo(deltdim, tmprect.top + (auxtoolname != NULL ? deltdim / 4 + 4
														   : deltdim / 2 + 5));
		DrawString(tmpstr);
	}
	if (auxtoolname != NULL) {
		c2p(auxtoolname, tmpstr);
		MoveTo(deltdim, tmprect.top + (deltdim * 3) / 4);
		DrawString(tmpstr);
	}
	if (!paintable) {
		gray_out_rect(&tmprect);
	}
	if (haspopup) {
		plot_sicn(designwin, tmprect.right - 20, tmprect.top + deltdim / 2 - 8,
				  menubuttonhandle);
	}
	if (!enabled) {
		gray_out_rect(&tmprect);
	}
	/* Draw a gray dividing line. */
	PenPat(QD(gray));
	MoveTo(0, tmprect.bottom - 1);  Line(tmprect.right, 0);
	PenNormal();
	/* Highlight the currently selected tool with a heavy outline rect. */
	if (tool == tooltype) {
		tmprect.bottom -= 1;
		InvertRect(&tmprect);
		InsetRect(&tmprect, 3, 3);
		InvertRect(&tmprect);
	}
}

/* This draws a type of terrain in a way that indicates its subtype. */

draw_terrain_sample(tmprect, t)
Rect tmprect;
int t;
{
	int dir;

	switch (t_subtype(t)) {
		case cellsubtype:
			draw_hex_polygon(tmprect.left, tmprect.top, 4, FALSE, t, -1);
			break;
		case bordersubtype:
			for_all_directions(dir) {
				draw_border_line(tmprect.left, tmprect.top, dir, 4, t);
			}
			break;
		case connectionsubtype:
			for_all_directions(dir) {
				draw_connection_line_half(tmprect.left, tmprect.top, dir, 4, t);
			}
			break;
		case coatingsubtype:
			draw_hex_polygon(tmprect.left, tmprect.top, 4, FALSE, t, -1);
			/* Make it a 50% pattern. */
			gray_out_rect(&tmprect);
			break;
		default:
			run_warning("Garbage %d subtype %d?", t, t_subtype(t));
			break;
	}
}

/* Respond to a mouse down in the designer's window. */

/* This macro implements cycling of a variable through a set of consecutive
   values, with direction controlled by the shift key. */

#define OPTION_CYCLE(var, lim, mods)  \
  ((var) = ((var) + ((mods) & shiftKey ? -1 : 1) + (lim)) % (lim))

do_mouse_down_design(mouse, mods)
Point mouse;
int mods;
{
	int oldtool, newutype, newbgttype;
	GeoFeature *feature;
	Point popupcorner;
	Rect tmprect;
	ControlHandle control;
	long ttypechoice, utypechoice, sidechoice;
	extern int nextfid;

	/* (should make popup controls for all the popups) */
	tmprect.left = 0;  tmprect.right = designwinw;
	oldtool = tooltype;
	/* Any click in the tool's area selects it. */
	tooltype = mouse.v / deltdim;
	switch (tooltype) {
		case notool:
			if (mods & optionKey) {
				/* Option-click and Option-Shift-click circulate brush sizes. */
				OPTION_CYCLE(curbrushradius, 99, mods);
			}
			break;
		case terraintool:
			if (mods & optionKey) {
				/* Option-click and Option-Shift-click circulate around all the types. */
				OPTION_CYCLE(curttype, numttypes, mods);
			} else if ((mods & cmdKey) && t_subtype(curttype) == cellsubtype) {
				/* Cmd-click and Cmd-Shift-click circulate around all the bg types. */
				newbgttype = curbgttype;
				do {
					OPTION_CYCLE(newbgttype, numttypes, mods);
					if (newbgttype == curbgttype) break;
				} while (t_subtype(newbgttype) != cellsubtype);
				curbgttype = newbgttype;
			} else if (mouse.h > tmprect.right - 20) {
				if (ttypemenu != nil) {
					tmprect.top = deltdim;  tmprect.bottom = tmprect.top + deltdim;
					popupcorner.h = tmprect.right;  popupcorner.v = tmprect.top;
					LocalToGlobal(&popupcorner);
					build_terrain_type_menu();
					ttypechoice =
						PopUpMenuSelect(ttypemenu, popupcorner.v, popupcorner.h,
										curttype + 1);
					InvertRect(&tmprect);
					if (LoWord(ttypechoice) > 0) {
						curttype = LoWord(ttypechoice) - 1;
					}
				}
			}
			break;
		case unittool:
			if (mods & optionKey) {
				/* Option-click and Option-Shift-click circulate around all the types
				   allowed for the current side. */
				newutype = curutype;
				do {
					OPTION_CYCLE(newutype, numutypes, mods);
					if (newutype == curutype) break;
				} while (!type_allowed_on_side(newutype, side_n(cursidenumber)));
				curutype = newutype;
			} else if (mouse.h > tmprect.right - 20) {
				if (utypemenu != nil) {
					tmprect.top = 2 * deltdim;  tmprect.bottom = tmprect.top + deltdim;
					InvertRect(&tmprect);
					popupcorner.h = tmprect.right;  popupcorner.v = tmprect.top;
					LocalToGlobal(&popupcorner);
					build_unit_type_menu();
					mark_allowed_unit_types();
					utypechoice =
						PopUpMenuSelect(utypemenu, popupcorner.v, popupcorner.h, curutype + 1);
					InvertRect(&tmprect);
					if (LoWord(utypechoice) > 0) {
						curutype = LoWord(utypechoice) - 1;
					}
				}
			}
			break;
		case peopletool:
			if (mods & optionKey) {
				/* Option-click and Option-Shift-click circulate around all the sides. */
				OPTION_CYCLE(cursidenumber, numsides + 1, mods);
			} else if (mouse.h > tmprect.right - 20) {
				if (sidemenu != nil) {
					tmprect.top = 3 * deltdim;  tmprect.bottom = tmprect.top + deltdim;
					InvertRect(&tmprect);
					popupcorner.h = tmprect.right;  popupcorner.v = tmprect.top;
					LocalToGlobal(&popupcorner);
					build_side_menu();
					mark_allowed_sides();
					sidechoice =
						PopUpMenuSelect(sidemenu, popupcorner.v, popupcorner.h, cursidenumber);
					InvertRect(&tmprect);
					if (LoWord(sidechoice) > 0) {
						cursidenumber = LoWord(sidechoice);
						if (cursidenumber > numsides) cursidenumber = 0;;
					}
				}
			}
			break;
		case materialtool:
			break;
		case featuretool:
			if (mods & optionKey && nextfid > 1) {
				/* Option-click and Option-Shift-click cycles around the features. */
				OPTION_CYCLE(curfid, (nextfid - 1), mods);
				curfeature = (GeoFeature *) find_geo_feature(curfid);
			} else if (mouse.h > tmprect.right - 20) {
				if ((feature = create_feature("?type?", "?name?")) != NULL) {
					curfeature = feature;
					curfid = feature->id;
				}
			}
			/* should make create/rename buttons for this */
			break;
		case elevationtool:
			if (mods & optionKey) {
				/* Option-click and Option-Shift-click adjusts the elevation. */
				OPTION_CYCLE(curelevation, 9999, mods);
			} else if (mouse.h > tmprect.right - 20) {
#if 0
				FindControl(mouse, designwin, &control);
				if (control == elevmenu) {
					TrackControl(control, mouse, (void *) -1);
					printf("%d\n", GetCtlValue(control));
				}
#endif
			}
			break;
		case temperaturetool:
			break;
		case cloudstool:
			break;
		case windstool:
			break;
		default:
			break;
	}
	/* Draw the old and new tools. */
	if (oldtool != tooltype) {
		draw_design_window_tool(oldtool);
	}
	draw_design_window_tool(tooltype);
	DrawControls(designwin);
}

mark_allowed_unit_types()
{
	Side *side = side_n(cursidenumber);
	int u;

	for_all_unit_types(u) {
		EnableItem(utypemenu, u + 1);
		SetItemMark(utypemenu, u + 1,
					(type_allowed_on_side(u, side) ? diamondMark : noMark));
	}
}

mark_allowed_sides()
{
	Side *side;

	for_all_sides(side) {
		EnableItem(sidemenu, side_number(side));
		SetItemMark(sidemenu, side_number(side),
					(type_allowed_on_side(curutype, side) ? diamondMark : noMark));
	}
	EnableItem(sidemenu, numsides + 1);
	SetItemMark(sidemenu, numsides + 1,
				(type_allowed_on_side(curutype, NULL) ? diamondMark : noMark));
}

#define set_flag_from_ditem(di,place)  \
  GetDItem(win, (di), &itemtype, &itemhandle, &itemrect);  \
  (place) = GetCtlValue((ControlHandle) itemhandle); 

#define put_number_into_ditem(di,num)  \
  GetDItem(win, (di), &itemtype, &itemhandle, &itemrect);  \
  NumToString((num), tmpstr);  \
  SetIText(itemhandle, tmpstr);

#define get_number_from_ditem(di,place)  \
  GetDItem(win, (di), &itemtype, &itemhandle, &itemrect);  \
  GetIText(itemhandle, tmpstr);  \
  StringToNum(tmpstr, &(place));

#define get_string_from_ditem(di,buf)  \
  GetDItem(win, (di), &itemtype, &itemhandle, &itemrect);  \
  GetIText(itemhandle, tmpstr);  \
  strncpy((buf), tmpstr+1, tmpstr[0]); /* this is basically p2c */  \
  (buf)[tmpstr[0]] = '\0';

/* This allows the designer to choose which parts of a game to write out. */

designer_save_dialog()
{
	int done = FALSE;
	short ditem;
	char namebuf[BUFSIZE];
	Str255 tmpstr; 
	WindowPtr win;
	Module *module;
	short itemtype;  Handle itemhandle;  Rect itemrect;
	
	win = GetNewDialog(dDesignerSave, NULL, (DialogPtr) -1L);
	module = create_game_module(NULL);
	if (module == NULL) { SysBeep(20); return; }
	init_module_reshape(module);
	/* Only rarely does the designer not want to compress all the area layers. */
	module->compresslayers = TRUE;
	GetDItem(win, diDesignerSaveCompress, &itemtype, &itemhandle, &itemrect);
	SetCtlValue((ControlHandle) itemhandle, module->compresslayers);
	while (!done) {
		ModalDialog(NULL, &ditem);
		switch (ditem) {
			case diDesignerSaveOK:
				get_string_from_ditem(diDesignerSaveName, namebuf);
				set_flag_from_ditem(diDesignerSaveTypes, module->deftypes);
				set_flag_from_ditem(diDesignerSaveTables, module->deftables);
				set_flag_from_ditem(diDesignerSaveGlobals, module->defglobals);
				set_flag_from_ditem(diDesignerSaveWorld, module->defworld);
				set_flag_from_ditem(diDesignerSaveAreas, module->defareas);
				set_flag_from_ditem(diDesignerSaveAreaTerrain, module->defareaterrain);
				set_flag_from_ditem(diDesignerSaveAreaMisc, module->defareamisc);
				set_flag_from_ditem(diDesignerSaveAreaWeather, module->defareaweather);
				set_flag_from_ditem(diDesignerSaveAreaMaterial, module->defareamaterial);
				set_flag_from_ditem(diDesignerSaveSides, module->defsides);
				set_flag_from_ditem(diDesignerSaveSideViews, module->defsideviews);
				set_flag_from_ditem(diDesignerSavePlayers, module->defplayers);
				set_flag_from_ditem(diDesignerSaveUnits, module->defunits);
				set_flag_from_ditem(diDesignerSaveUnitProps, module->defunitprops);
				set_flag_from_ditem(diDesignerSaveUnitMoves, module->defunitacts);
				set_flag_from_ditem(diDesignerSaveUnitPlans, module->defunitplans);
				set_flag_from_ditem(diDesignerSaveScoring, module->defscoring);
				set_flag_from_ditem(diDesignerSaveHistory, module->defhistory);
				set_flag_from_ditem(diDesignerSaveCompress, module->compresslayers);
				/* (should do sfputfile using namebuf to seed) */
				sprintf(spbuf, "%s.g", namebuf);
				module->filename = copy_string(spbuf);
				if (!write_game_module(module)) {
					SysBeep(20);  /* (should make a better warning) */
					break;
				}
				/* Fall through to next case. */
			case diDesignerSaveCancel:
				done = TRUE;
				break;
			case diDesignerSaveTypes:
			case diDesignerSaveTables:
			case diDesignerSaveGlobals:
			case diDesignerSaveWorld:
			case diDesignerSaveAreas:
			case diDesignerSaveAreaTerrain:
			case diDesignerSaveAreaMisc:
			case diDesignerSaveAreaWeather:
			case diDesignerSaveAreaMaterial:
			case diDesignerSaveSides:
			case diDesignerSaveSideNames:
			case diDesignerSaveSideProps:
			case diDesignerSaveSideViews:
			case diDesignerSavePlayers:
			case diDesignerSaveUnits:
			case diDesignerSaveUnitProps:
			case diDesignerSaveUnitMoves:
			case diDesignerSaveUnitPlans:
			case diDesignerSaveScoring:
			case diDesignerSaveHistory:
			case diDesignerSaveCompress:
				/* Toggle check boxes. */
				GetDItem(win, ditem, &itemtype, &itemhandle, &itemrect);
				SetCtlValue((ControlHandle) itemhandle,
							!GetCtlValue((ControlHandle) itemhandle));
				break;
			case 4 /*diDesignerSaveModule*/:
				/* (should bring up a dialog to set module properties) */
				break;
			case diDesignerSaveReshape:
				/* Bring up *another* modal dialog. */
				designer_reshape_dialog(module);
				break;
		}
	}
	DisposDialog(win);
}

/* A special dialog that allows for saving a world of different dimensions than is
   currently being used in the game. */

designer_reshape_dialog(module)
Module *module;
{
	int done = FALSE;
	short ditem;
	WindowPtr win;
	Str255 tmpstr;
	short itemtype;  Handle itemhandle;  Rect itemrect;
	
	win = GetNewDialog(dDesignerReshape, NULL, (DialogPtr) -1L);
	while (!done) {
  		put_number_into_ditem(diDesignerReshapeOrigWidth, area.width);
  		put_number_into_ditem(diDesignerReshapeOrigHeight, area.height);
  		put_number_into_ditem(diDesignerReshapeOrigWorld, world.circumference);
  		put_number_into_ditem(diDesignerReshapeOrigSubWidth, module->subareawidth);
  		put_number_into_ditem(diDesignerReshapeOrigSubHeight, module->subareaheight);
  		put_number_into_ditem(diDesignerReshapeOrigSubX, module->subareax);
  		put_number_into_ditem(diDesignerReshapeOrigSubY, module->subareay);
  		put_number_into_ditem(diDesignerReshapeOutputSubWidth, module->finalsubareawidth);
  		put_number_into_ditem(diDesignerReshapeOutputSubHeight, module->finalsubareaheight);
  		put_number_into_ditem(diDesignerReshapeOutputSubX, module->finalsubareax);
  		put_number_into_ditem(diDesignerReshapeOutputSubY, module->finalsubareay);
  		put_number_into_ditem(diDesignerReshapeOutputWidth, module->finalwidth);
  		put_number_into_ditem(diDesignerReshapeOutputHeight, module->finalheight);
  		put_number_into_ditem(diDesignerReshapeOutputWorld, module->finalcircumference);
/*  		put_number_into_ditem(diDesignerReshapeFillTerrain, module->fillterrain); */
		ModalDialog(NULL, &ditem);
		switch (ditem) {
			case diDesignerReshapeOK:
		  		get_number_from_ditem(diDesignerReshapeOrigSubWidth, module->subareawidth);
		  		get_number_from_ditem(diDesignerReshapeOrigSubHeight, module->subareaheight);
		  		get_number_from_ditem(diDesignerReshapeOrigSubX, module->subareax);
		  		get_number_from_ditem(diDesignerReshapeOrigSubY, module->subareay);
		  		get_number_from_ditem(diDesignerReshapeOutputSubWidth, module->finalsubareawidth);
		  		get_number_from_ditem(diDesignerReshapeOutputSubHeight, module->finalsubareaheight);
		  		get_number_from_ditem(diDesignerReshapeOutputSubX, module->finalsubareax);
		  		get_number_from_ditem(diDesignerReshapeOutputSubY, module->finalsubareay);
		  		get_number_from_ditem(diDesignerReshapeOutputWidth, module->finalwidth);
		  		get_number_from_ditem(diDesignerReshapeOutputHeight, module->finalheight);
		  		get_number_from_ditem(diDesignerReshapeOutputWorld, module->finalcircumference);
/*		  		get_number_from_ditem(diDesignerReshapeFillTerrain, module->fillterrain);  */
				/* Fall through to next case. */
			case diDesignerReshapeCancel:
				done = TRUE;
				break;
		}
	}
	DisposDialog(win);
}

#endif /* DESIGNERS */
