/* Copyright (c) 1992-1993  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Map graphics for the Mac interface. */

#include "conq.h"
#include "mac.h"

int mayseeall = FALSE;

int weseeall = FALSE;

RGBColor graycolor, blackcolor;

#define hexagon_adjust(m) (area.xwrap ? 0 : ((area.height / 2) * (m)->hw) / 2)

#define terrain_visible(x, y)  \
  (g_see_all() || dside->designer || weseeall || terrain_view(dside, wrapx(x), y) != UNSEEN)

#define units_visible(x, y)  \
  (g_see_all() || dside->designer || weseeall || cover(dside, wrap(x), y) >= 1)

#define bords_to_draw(m) (numbordtypes > 0 && bwid[(m)->power] > 0)

#define conns_to_draw(m) (numconntypes > 0 && cwid[(m)->power] > 0)

#define draw_any_materials(m) (0)

int conwid = 32;

int tophgt = 16;

int mapnum = 1;

int nummaps = 0;

PicHandle tlcontrols = nil;
PicHandle blcontrols = nil;

char *mouseover = "something";

/* Draw an individual detailed hex, as a row of one, on all maps. */
  
update_cell_display(side, x, y, rightnow)
Side *side;
int x, y, rightnow;
{
	GrafPtr oldport;
	Map *map;
	RgnHandle tmprgn;
	
	if (active_display(side)) {
		GetPort(&oldport);
		for_all_maps(map) {
			/* (should only do if x,y is actually visible...) */
			/* Set up the drawing context. */
			SetPort(map->window);
			tmprgn = NewRgn();
			GetClip(tmprgn);
			ClipRect(&(map->contentrect));
			/* Draw the cell. */
			draw_row(map, x, y, 1, TRUE);
			/* Draw any selections that are here. */
			draw_selections_at(map, x, y);
			SetClip(tmprgn);
			DisposeRgn(tmprgn);
		}
		SetPort(oldport);
	}
}

/* Create a new map window and all its paraphernalia. */

Map *
create_map(power)
int power;
{
	int m, sx, sy, i, x, y;
	Unit *unit;
	Rect vscrollrect, hscrollrect;
	Map *map = (Map *) xmalloc(sizeof(Map));

	DGprintf("Creating map, mag power %d\n", power);
	set_map_power(map, power);
	/* Pick an appropriate focus of the view. */
	pick_a_focus(dside, &x, &y);
	map->vcx = x;  map->vcy = y;
	/* Set default values for the display controls. */
	map->toph = 0;
	map->drawterrain = TRUE;
	map->drawhexpats = TRUE;  /* should get from prefs */
	map->drawunits = TRUE;
	map->drawnames = defaultdrawnames;
	map->drawpeople = FALSE;
	map->drawelevations = FALSE;
	for_all_material_types(m) {
		map->drawmaterials[m] = FALSE;
	}
	map->drawtemp = FALSE;
	map->drawweather = FALSE;
	map->drawlighting = TRUE;
	map->drawgrid = defaultdrawgrid;
	map->drawothermaps = TRUE;
	map->numselections = 0;
	/* Newest map goes on the front of the list. */
	map->next = maplist;
	maplist = map;
	if (hasColorQD) {
		map->window = GetNewCWindow(wMap, nil, (WindowPtr) -1L);
	} else {
		map->window = GetNewWindow(wMap, nil, (WindowPtr) -1L);
	}
	ShowWindow(map->window);
	SetPort(map->window);
	sprintf(spbuf, "Map %d", mapnum++);
	add_window_menu_item(spbuf, map->window);
	set_content_rect(map);
	center_on_focus(map);
	set_map_viewport(map);
	/* Make the scrollbars. */
	vscrollrect = map->window->portRect;
	vscrollrect.top -= 1;
	vscrollrect.bottom -= sbarwid - 1;
	vscrollrect.left = vscrollrect.right - sbarwid;
	vscrollrect.right += 1;
	map->vscrollbar =
		NewControl(map->window, &vscrollrect, "\p", 1,
			 map->sy, 0, max(0, map->totsh - map->pxh), scrollBarProc, 0L);
	hscrollrect = map->window->portRect;
	hscrollrect.top = hscrollrect.bottom - sbarwid;
	hscrollrect.bottom += 1;
	hscrollrect.left += conwid + 1;
	hscrollrect.right -= sbarwid - 1;
	map->hscrollbar =
		NewControl(map->window, &hscrollrect, "\p", 1,
			 map->sx, 0, max(0, map->totsw - map->pxw), scrollBarProc, 0L);
	set_map_scrollbars(map);
	++nummaps;
	return map;
}

/* Compute the content part of the map window. */

set_content_rect(map)
Map *map;
{
	map->contentrect = map->window->portRect;
	map->contentrect.left += conwid;  map->contentrect.top += map->toph;
	map->contentrect.right -= sbarwid;  map->contentrect.bottom -= sbarwid;
	map->pxw = (map->contentrect.right - map->contentrect.left);
	map->pxh = (map->contentrect.bottom - map->contentrect.top);
}

/* Set vcx/vcy to point to the center of the displayed map. */

focus_on_center(map)
Map *map;
{
	map->vcy = (map->totsh - (map->sy + map->pxh / 2)) / map->hch;
	map->vcx = map->sx / map->hw - (map->vcy / 2) + map->vh / 2;
	/* Restrict the focus to be *inside* the area. */
	map->vcy = max(1, min(map->vcy, area.height - 2));
	if (area.xwrap) {
		map->vcx = wrapx(map->vcx);
	} else {
		map->vcx = max(1, min(map->vcx, area.width - 2));
		if (map->vcx + map->vcy < area.height / 2 + 1)
			map->vcx = area.height / 2 + 1;
		if (map->vcx + map->vcy > area.width + area.height / 2 - 1)
			map->vcx = area.width + area.height / 2 - 1;
	}
}

/* Put vcx/vcy in the middle of the map. */

center_on_focus(map)
Map *map;
{
	map->sx = map->vcx * map->hw + (map->vcy * map->hw) / 2 - map->pxw / 2 + map->hw / 2;
	map->sy = map->totsh - (map->vcy * map->hch + map->pxh / 2 + map->hh / 2);
	map->sx = max(0, min(map->sx, map->totsw - map->pxw));
	map->sy = max(0, min(map->sy, map->totsh - map->pxh));
	DGprintf("Mapwin at %d,%d, focused at %d,%d\n", map->sx, map->sy, map->vcx, map->vcy);
}

/* Adjust the appearance and thumb of the scroll bars to reflect the map. */

set_map_scrollbars(map)
Map *map;
{
	if (map->pxw < (map->totsw - hexagon_adjust(map))) {
		SetCtlMin(map->hscrollbar, hexagon_adjust(map));
		/* Wrapped-around areas need extra room to look at cells on the seam. */
		SetCtlMax(map->hscrollbar, map->totsw - (area.xwrap ? 0 : map->pxw));
		SetCtlValue(map->hscrollbar, map->sx);
		HiliteControl(map->hscrollbar, 0);
	} else {
		/* Force sx to a default value and disable the scrollbar. */
		map->sx = hexagon_adjust(map);
		HiliteControl(map->hscrollbar, 255);
	}
	if (map->pxh < map->totsh) {
		SetCtlMax(map->vscrollbar, map->totsh - map->pxh);
		HiliteControl(map->vscrollbar, 0);
	} else {
		map->sy = 0;
		HiliteControl(map->vscrollbar, 255);
	}
}

/* Given a magnification power, update misc map parms from arrays. */

set_map_power(map, power)
Map *map;
int power;
{
	map->power = power;
	map->mag = mags[power];
	map->hw = hws[power];  map->hh = hhs[power];  map->hch = hcs[power];
	map->uw = uws[power];  map->uh = uhs[power];
	/* Calculate and cache the size in pixels of the whole area. */
	map->totsw = area.width * map->hw;
	/* A hexagon area also bases at 0,0, so needs to be wider. */
	map->totsw += hexagon_adjust(map);
	/* Height is based on center-to-center height, plus an adjustment. */
	map->totsh = area.height * map->hch + (map->hh - map->hch);
}

/* Given a map with already-set pixel size and position, compute
   reasonable values for its size and position in area coordinates.
   This is really just caching, can be run at any time. */

set_map_viewport(map)
Map *map;
{
	/* Compute the size of the viewport.  Make sure it will extend past the edges
	   of the window, so that partial hexes will fill in around the edges. */
	map->vw = min(area.width, map->pxw / map->hw + 2);
	map->vh = min(area.height, map->pxh / map->hch + 1);
	/* Compute the bottom visible row. */
	map->vy = ((map->totsh - map->sy) / map->hch) - map->vh;
	/* Now adjust the bottom row so it doesn't go outside the area. */
	if (map->vy < 0) map->vy = 0;
	if (map->vy > area.height - map->vh) map->vy = area.height - map->vh;
	/* Compute the leftmost "column". */
	map->vx = map->sx / map->hw - map->vy / 2 - 1;
	DGprintf("Set %dx%d viewport at %d,%d\n", map->vw, map->vh, map->vx, map->vy);
}

/* Given a window, find the map that it belongs to. */

Map *
map_from_window(window)
WindowPtr window;
{
	Map *map;
	
	if (dside == NULL) return NULL;
	for_all_maps(map) {
		if (map->window == window) return map;
	}
	return NULL;
}

/* Set the size of the map window and position its scrollbars correctly. */

grow_map(map, w, h)
Map *map;
int w, h;
{
/*	EraseRect(&map->window->portRect); */
	SizeWindow(map->window, w, h, 1);
	adjust_map_decor(map);
}

/* Map zooming actually does "rightsizing" if possible. */

zoom_window_map(map, part)
Map *map;
short part;
{
	/* (should look at actual size of main screen...) */
	if (area.width * map->hw < 640 || area.height * map->hch < 480) {
		grow_map(map,
				 min(640, area.width * map->hw + conwid + sbarwid + 1),
				 min(480, area.height * map->hch + sbarwid + map->toph + 1));
		/* should move the map to make it all visible if necessary */
	} else {
		EraseRect(&map->window->portRect);
		ZoomWindow(map->window, part, true);
		adjust_map_decor(map);
	}
}

/* Move and size the controls to be correct for the map. */

adjust_map_decor(map)
Map *map;
{				
	int w, h;

	w = map->window->portRect.right - map->window->portRect.left;
	h = map->window->portRect.bottom - map->window->portRect.top;
/*	HideControl(map->hscrollbar); */
	MoveControl(map->hscrollbar, conwid, h - sbarwid);
	SizeControl(map->hscrollbar, w - conwid - sbarwid + 1, sbarwid + 1);
/*	HideControl(map->vscrollbar); */
	MoveControl(map->vscrollbar, w - sbarwid, -1);
	SizeControl(map->vscrollbar, sbarwid + 1, h - sbarwid + 1 + 1);
	set_content_rect(map);
	InvalRect(&map->window->portRect);
}

/* Decide whether given location is away from the edge of the map's window. */

in_middle(map, x, y)
Map *map;
int x, y;
{
	int sx, sy;
	
	xform(map, x, y, &sx, &sy);
	return (between(20, sx, map->pxw - 20) && between(20, sy, map->pxh - 20));
}

/* Given a map and a cell, compute the pixel coords of the cell's UL corner.
   This is the core routine that relates cells and pixels. */

xform(map, x, y, sxp, syp)
Map *map;
int x, y, *sxp, *syp;
{
	if (in_area(x, y)) {
		*sxp = x * map->hw + (y * map->hw) / 2 - map->sx;
		*syp = (map->totsh - (map->hh + y * map->hch)) - map->sy;
	} else {
		/* Always complain about this, indicates bugs. */
		run_warning("attempting to xform %d,%d", x, y);
		/* But it's not fatal, so return a semi-plausible position. */
		*sxp = *syp = 1;
	}
	/* Adjust for the control panel and topline displays. */
	*sxp += conwid;  *syp += map->toph;
}

xform_unit(map, unit, sxp, syp, swp, shp)
Map *map;
Unit *unit;
int *sxp, *syp, *swp, *shp;
{
	int num, n, sq, sx1, sy1, sw1, sh1, x = unit->x, y = unit->y;
	Unit *unit2;

	if (unit->transport == NULL) {
		xform(map, x, y, sxp, syp);
		num = num_units_at(x, y);
		if (num == 0) { *swp = *shp = *sxp = *syp = 1; return; }
		/* Figure out our position in this cell's stack list. */
		n = 0;
		for_all_stack(x, y, unit2) {
			if (unit2 == unit) break;
			++n;
		}
		sq = isqrt(num);
		if (sq * sq < num) ++sq;
		*swp = map->uw / sq;  *shp = map->uh / sq;
		*sxp += (map->hw - map->uw) / 2 + *swp * (n / sq);
		*syp += (map->hh - map->uh) / 2 + *shp * (n % sq);
	} else {
		/* Go up the transport chain to get the bounds for this unit. */
		xform_unit(map, unit->transport, &sx1, &sy1, &sw1, &sh1);
		xform_occupant(map, unit->transport, unit, sx1, sy1, sw1, sh1, sxp, syp, swp, shp);
	}
}

xform_unit_self(map, unit, sxp, syp, swp, shp)
Map *map;
Unit *unit;
int *sxp, *syp, *swp, *shp;
{
	int num, n, sq, sx1, sy1, sw1, sh1, x = unit->x, y = unit->y;
	Unit *unit2;

	if (unit->transport == NULL) {
		if (unit->occupant == NULL) {
			xform_unit(map, unit, sxp, syp, swp, shp);
		} else {
			xform_unit(map, unit, &sx1, &sy1, &sw1, &sh1);
			xform_occupant(map, unit, unit, sx1, sy1, sw1, sh1, sxp, syp, swp, shp);
		}
	} else {
		xform_unit(map, unit->transport, &sx1, &sy1, &sw1, &sh1);
		xform_occupant(map, unit->transport, unit, sx1, sy1, sw1, sh1, sxp, syp, swp, shp);
	}
}

xform_occupant(map, transport, unit, sx, sy, sw, sh, sxp, syp, swp, shp)
Map *map;
Unit *transport, *unit;
int sx, sy, sw, sh, *sxp, *syp, *swp, *shp;
{
	int num, n, sq;
	Unit *unit2;

	if ((num = num_occupants(transport)) <= 0) {
		run_warning("Zero occupants?");
		*swp = *shp = *sxp = *syp = 1;
		return;
	}
	++num;
	/* Figure out the position of this unit among all the occupants. */
	/* The transport is counted too, as number 0. */
	if (transport == unit) {
		n = 0;
	} else {
		n = 1;
		for_all_occupants(transport, unit2) {
			if (unit2 == unit) break;
			++n;
		}
	}
	/* Compute how the box will be subdivided. */
	sq = isqrt(num);
	if (sq * sq < num) ++sq;
	*swp = sw / sq;  *shp = sh / sq;
	*sxp = sx + *swp * (n / sq);  *syp = sy + *shp * (n % sq);
}

/* Un-transform screen coordinates (as supplied by mouse perhaps) into
   map coordinates.  This routine doesn't yet
   pay attention to hexagon boundaries, using only box-shaped areas,
   but in practice (for small hexes only) this seems to be OK. */

nearest_cell(map, sx, sy, xp, yp)
Map *map;
int sx, sy, *xp, *yp;
{
	/* Subtract offset due to map decor. */
	sx -= conwid;
	sy -= map->toph;
	/* Flip the raw y and then scale to hex coords. */
	*yp = (map->totsh - (map->sy + sy)) / map->hch;
	/* Scale adjusted x to hex coord. */
	*xp = (sx + map->sx - (*yp * map->hw) / 2) / map->hw;
	/* If the magnification of the map is large enough that the top and bottom
	   edges of a hex are visibly sloping, then we have to take those edges
	   int account, and accurately. */
	if ((map->hh - map->hch) / 2 > 1) {
		/* (should adjust according to hex boundaries correctly here) */
	}
	/* Wrap coords as usual. */
	if (area.xwrap) *xp = wrapx(*xp);
	DGprintf("Pixel %d,%d -> hex %d,%d\n", sx, sy, *xp, *yp); 
}

nearest_boundary(map, sx, sy, xp, yp, dirp)
Map *map;
int sx, sy, *xp, *yp, *dirp;
{
	int sx2, sy2, ydelta, hexslope;

	/* Get the nearest cell... */
	nearest_cell(map, sx, sy, xp, yp);
	/* ... and xform it back to get the pixel coords. */ 
	xform(map, *xp, *yp, &sx2, &sy2);
	ydelta = sy - sy2;
	hexslope = (map->hh - map->hch) / 2;
	if (sx - sx2 > map->hw / 2) {
		*dirp = ((ydelta < hexslope) ? NE : (ydelta > map->hch ? SE : EAST));
	} else {
		*dirp = ((ydelta < hexslope) ? NW : (ydelta > map->hch ? SW : WEST));
	}
	DGprintf("Pixel %d,%d -> hex %d,%d dir %d\n", sx, sy, *xp, *yp, *dirp); 
}

Unit *
find_unit_or_occ(map, unit, usx, usy, usw, ush, sx, sy)
Map *map;
Unit *unit;
int usx, usy, usw, ush, sx, sy;
{
	int usx1, usy1, usw1, ush1;
	Unit *occ, *rslt;

	/* See if the point might be over an occupant. */
	if (unit->occupant != NULL) {
		for_all_occupants(unit, occ) {
			xform_unit(map, occ, &usx1, &usy1, &usw1, &ush1);
			if ((rslt = find_unit_or_occ(map, occ, usx1, usy1, usw1, ush1, sx, sy))) {
				return rslt;
			}
		}
	}
	/* Otherwise see if it could be the unit itself.  This has the effect of
	   "giving" the transport everything in its box that is not in an occ. */
	xform_unit(map, unit, &usx1, &usy1, &usw1, &ush1);
	if (between(usx1, sx, usx1 + usw1) && between(usy1, sy, usy1 + ush1)) {
		return unit;
	}
	return NULL;
}

Unit *
find_unit_at(map, x, y, sx, sy)
Map *map;
int x, y, sx, sy;
{
	int usx, usy, usw, ush;
	Unit *unit, *rslt;
	
	for_all_stack(x, y, unit) {
		xform_unit(map, unit, &usx, &usy, &usw, &ush);
		if ((rslt = find_unit_or_occ(map, unit, usx, usy, usw, ush, sx, sy))) {
			return rslt;
		}
	}
	return NULL;
}

nearest_unit(map, sx, sy, unitp)
Map *map;
int sx, sy;
Unit **unitp;
{
	int x, y;

	nearest_cell(map, sx, sy, &x, &y);
	if (map->power > 4) {
		*unitp = find_unit_at(map, x, y, sx, sy);
	} else {
		*unitp = unit_at(x, y);
	}
	DGprintf("Pixel %d,%d -> unit %s\n", sx, sy, unit_desig(*unitp)); 
}

erase_map(map)
Map *map;
{
	Rect tmprect;
	WindowPtr mapwin = map->window;
	RgnHandle tmprgn;

	BackPat(QD(gray));
	EraseRect(&(mapwin->portRect));
}

/* Display a map and all of its paraphernalia. */

draw_map(map)
Map *map;
{
	Rect tmprect;
	WindowPtr mapwin = map->window;
	RgnHandle tmprgn;

	/* Draw control panel and topline before clipping to inner part of window. */
	draw_control_panel(map);
	if (map->toph > 0) draw_top_line(map);
	tmprgn = NewRgn();
	GetClip(tmprgn);
	ClipRect(&(map->contentrect));
	/* Calculate shapes and sizes. */
	set_map_viewport(map); /* flush eventually */
	DGprintf("Displaying map at %d,%d, size %d,%d, focus %d,%d\n",
		map->vx, map->vy, map->vw, map->vh, map->vcx, map->vcy);
	/* If part of the window is entirely outside the world, we draw its shape on
	   top of gray, otherwise window starts out all white. */
	if (area.width * map->hw < 32000) {
		FillRect(&(map->contentrect), QD(gray));
	} else {
		if (strcmp(g_grid_color(), "gray") == 0) {
			if (hasColorQD) {
				graycolor.red = graycolor.green = graycolor.blue = 32768;
				RGBForeColor(&graycolor);
				PaintRect(&(map->contentrect));
				blackcolor.red = blackcolor.green = blackcolor.blue = 0;
				RGBForeColor(&blackcolor);
			} else {
				FillRect(&(map->contentrect), QD(gray));
			}
		} else {
			FillRect(&(map->contentrect), QD(white));
		}
	}
	set_map_scrollbars(map);
	draw_area_background(map);
	draw_map_content(map);
	if (map->drawothermaps) draw_other_maps(map);
	draw_selections(map);
#ifdef DEBUGGING
	/* Indicate where the focus is. */
	if (DebugG) {
		int sx, sy;

		xform(map, map->vcx, map->vcy, &sx, &sy);
		SetRect(&tmprect, sx - 4, sx - 4, sx + map->hw + 4, sy + map->hh + 4);
		InvertOval(&tmprect);
		InsetRect(&tmprect, 2, 2);
		InvertOval(&tmprect);
	}
#endif /* DEBUGGING */
	SetClip(tmprgn);
	DisposeRgn(tmprgn);
}

/* Draw the actual map data. */

draw_map_content(map)
Map *map;
{
	int y1, y2, y, x1, x2, xx1, yy1, xx2, yy2;
	int halfheight = area.height / 2;
	Rect bbox = (*(map->window->visRgn))->rgnBBox;	

	/* Compute top and bottom rows to be displayed. */
	y1 = min(map->vy + map->vh, area.height - 1);
	y2 = map->vy;
	nearest_cell(map, bbox.left, bbox.top, &xx1, &yy1);
	nearest_cell(map, bbox.right, bbox.bottom, &xx2, &yy2);
	DGprintf("Map is %d - %d, Update is %d - %d\n", y2, y1, yy2, yy1);
	if (between(y2, yy1, y1)) y1 = yy1; 
	if (between(y2, yy2, y1)) y2 = yy2; 
	for (y = y1; y >= y2; --y) {
		/* Adjust the right and left bounds to fill the viewport as
		   much as possible, without going too far (the drawing code
		   will clip, but clipped drawing is still expensive). */
		/* could test by drawing viewport rect as lines... */
		x1 = map->vx - (y - map->vy) / 2;
		x2 = x1 + map->vw + 2 /* bleah, shouldn't be necessary */;
		/* If the area doesn't wrap, then we might have to stop
		   drawing before we reach the edge of the viewport. */
		if (!area.xwrap) {
			/* Truncate x's to stay within the area. */
			x1 = max(0, min(x1, area.width-1));
			x2 = max(0, min(x2, area.width));
			/* If this row is entirely in the NE corner, don't draw anything. */
			if (x1 + y > area.width + halfheight)
			  continue;
			/* If this row is entirely in the SW corner, don't draw anything. */
			if (x2 + y < halfheight)
			  continue;
			/* If the row ends up in the NE corner, shorten it. */
			if (x2 + y > area.width + halfheight)
			  x2 = area.width + halfheight - y;
			/* If the row starts out in the SW corner, shorten it. */
			if (x1 + y < halfheight)
			  x1 = halfheight - y;
		}
		draw_row(map, x1, y, x2 - x1, FALSE);
	}
}

/* This draws a hexagon or rectangle that covers the total area of the area, whether
   seen or not. */
 
draw_area_background(map)
Map *map;
{
	int sx, sy, x, y;
	int llx, lly, lrx, lry, rx, ry, urx, ury, ulx, uly, lx, ly;
	PolyHandle poly;
	Rect arearect;

	if (area.width * map->hw < 32000) {
		if (area.xwrap) {
			xform(map, 0, 0, &sx, &sy);
			arearect.left = 0;  arearect.top = 0;
			arearect.right = area.width * map->hw;  arearect.bottom = sy;
			if (strcmp(g_grid_color(), "gray") == 0) {
				if (hasColorQD) {
					graycolor.red = graycolor.green = graycolor.blue = 32768;
					RGBForeColor(&graycolor);
					PaintRect(&arearect);
					blackcolor.red = blackcolor.green = blackcolor.blue = 0;
					RGBForeColor(&blackcolor);
				} else {
					FillRect(&arearect, QD(gray));
				}
			} else {
				FillRect(&arearect, QD(white));
			}
		} else {
			/* (should make once and save) */
			poly = OpenPoly();		
			xform(map, 0 + area.height/2, 0, &llx, &lly);
			MoveTo(llx, lly);
			xform(map, area.width-1, 0, &lrx, &lry);
		 	LineTo(lrx, lry);
			xform(map, area.width-1, area.height/2, &rx, &ry);
			LineTo(rx, ry);
 			xform(map, area.width-1 - area.height/2, area.height-1, &urx, &ury);
			LineTo(urx, ury);
 			xform(map, 0, area.height-1, &ulx, &uly);
			LineTo(ulx, uly);
 			xform(map, 0, area.height/2, &lx, &ly);
			LineTo(lx, ly);
			LineTo(llx, lly);
			ClosePoly();
			OffsetPoly(poly, map->hw/2, map->hh/2);
			if (strcmp(g_grid_color(), "gray") == 0) {
				if (hasColorQD) {
					graycolor.red = graycolor.green = graycolor.blue = 32768;
					RGBForeColor(&graycolor);
					PaintPoly(poly);
					blackcolor.red = blackcolor.green = blackcolor.blue = 0;
					RGBForeColor(&blackcolor);
				} else {
					FillPoly(poly, QD(gray));
				}
			} else {
				FillPoly(poly, QD(white));
			}
		}
#if 0
		/* This doesn't look very good. */
		for (x = 0; x < area.width; ++x) {
			xform(map, x, 0, &sx, &sy);
			draw_border_line(sx, sy, SW, map->power, -1);
			draw_border_line(sx, sy, SE, map->power, -2);
		}
		PenNormal();
#endif
	}
}

/* Draw the map control panel as a pair of PICTs. */

draw_control_panel(map)
Map *map;
{
	int winhgt;
	Rect tmprect;

	winhgt = (map->window->portRect).bottom - (map->window->portRect).top;
	SetRect(&tmprect, 0, 0, conwid, winhgt);
	FillRect(&tmprect, QD(white));
	MoveTo(conwid - 1, 0);  Line(0, winhgt);
	if (tlcontrols == nil) {
		tlcontrols = (PicHandle) GetResource('PICT', pMapControlsTL);
	}
	if (tlcontrols != nil) {
		SetRect(&tmprect, 0, 0,
				picture_width(tlcontrols), picture_height(tlcontrols));
		DrawPicture(tlcontrols, &tmprect);
	}
	if (blcontrols == nil) {
		blcontrols = (PicHandle) GetResource('PICT', pMapControlsBL);
	}
	if (blcontrols != nil) {
		SetRect(&tmprect, 0, winhgt - picture_height(blcontrols),
				picture_width(blcontrols), winhgt);
		DrawPicture(blcontrols, &tmprect);
	}
	/* (should modify appearance of top left arrow buttons to reflect abilities) */
	/* Draw state of bottom left control buttons. */
	if (map->power <= 0) {
		SetRect(&tmprect, 0, winhgt - 15, 15, winhgt);
		gray_out_rect(&tmprect);
	}
	if (map->power >= NUMPOWERS - 1) {
		SetRect(&tmprect, 16, winhgt - 15, 30, winhgt);
		gray_out_rect(&tmprect);
	}
}

draw_top_line(map)
Map *map;
{
	Rect tmprect;
	Str255 tmpstr;

	/* Clear the whole topline area. */
	SetRect(&tmprect, conwid, 0, conwid + map->pxw, tophgt);
	FillRect(&tmprect, QD(white));
	/* Draw a line dividing this from the map content. */
	MoveTo(conwid, tmprect.bottom - 1);  Line(map->pxw, 0);
	/* Draw the current date. */
	TextSize(10);
	MoveTo(tmprect.right - 60, tmprect.top + 11);
	c2p(absolute_date_string(curturn), tmpstr);
	DrawString(tmpstr);
	/* Draw description of what the mouse is over. */
	MoveTo(conwid + 10, tmprect.top + 11);
	c2p(mouseover, tmpstr);
	DrawString(tmpstr);
}

/* Draw an indication of the position of other maps relative to this one. */

draw_other_maps(map)
Map *map;
{
	Map *map2;

	for_all_maps(map2) {
		if (map != map2 /* && reasonable to show? */) {
			draw_other_map(map, map2);
		}
	}
}

draw_related_maps(map)
Map *map;
{
	Map *map2;
	GrafPtr oldport;

	for_all_maps(map2) {
		if (map != map2 && map2->drawothermaps /* && reasonable to show? */) {
			GetPort(&oldport);
			SetPort(map2->window);
			/* (also clipping?) */
			draw_other_map(map2, map);
			SetPort(oldport);
		}
	}
}

draw_other_map(map, map2)
Map *map, *map2;
{
	int sx, sy, sw, sh;
	Rect tmprect;

	sx = map2->sx - map->sx;
	sy = map2->sy - map->sy;
	sw = (map2->pxw * map->hw) / map2->hw;
	sh = (map2->pxh * map->hch) / map2->hch;
	SetRect(&tmprect, sx, sy, sx + sw, sy + sh);
	PenSize(2, 2);
	FrameRect(&tmprect);
	PenNormal();
}

/* x0 may be negative here. */

draw_row(map, x0, y0, len, clearit)
Map *map;
int x0, y0, len, clearit;
{
	int empty = TRUE;
	char ch;
	int i = 0, x, sx, sy, b, c;

	if (!between(0, y0, area.height-1)) return;

	/* Examine row to see if we can skip it entirely. */
	for (x = x0; x < x0 + len; ++x) {
		if (terrain_visible(x, y0)) { empty = FALSE;  break; }
		/* (only need following test if border types possible) */
		if (terrain_visible(x, y0 + 1)) { empty = FALSE;  break; }
	}
	if (empty) return;
	/* The terrain always comes first. */
	if (map->drawterrain) {
		draw_terrain_row(map, x0, y0, len);
	}
	if (map->drawterrain && any_aux_terrain_defined()) {
		if (bords_to_draw(map)) {
			/* The relative ordering of these is quite important.  Note that
			   each should be prepared to run independently also, since the
			   other displays might have been turned off. */
			PenSize(bwid[map->power], bwid[map->power]);
			for_all_terrain_types(b) {
				if (t_is_border(b) && aux_terrain_defined(b)) {
					for (x = x0; x < x0 + len; ++x) {
						/* Nothing inside the edge cells of the area, so skip them. */
						if (!inside_area(x, y0)) continue;
						draw_ul_borders(map, x, y0, b);
					}
				}
			}
		}
		if (conns_to_draw(map)) {
			PenSize(cwid[map->power], cwid[map->power]);
			for_all_terrain_types(c) {
				if (t_is_connection(c) && aux_terrain_defined(c)) {
					for (x = x0; x < x0 + len; ++x) {
						/* Nothing inside the edge cells of the area, so skip them. */
						if (!inside_area(x, y0)) continue;
						draw_connections(map, x, y0, c);
					}
				}
			}
		}
		PenSize(1, 1);
	}
	if (draw_any_materials(map) /* and any to draw, etc */) {
		for (x = x0; x < x0 + len; ++x) {
			draw_materials(map, x, y0);
		}
	}
	/* (should be global to entire map drawing somehow?) */
	if (map->drawnames) {
		for (x = x0; x < x0 + len; ++x) {
			if (!inside_area(x, y0)) continue;
			draw_legend(map, x, y0);
		}
	}
	if (people_sides_defined() && map->hw >= 4) {
		for (x = x0; x < x0 + len; ++x) {
			draw_people(map, x, y0);
		}
	}
	if (map->drawunits) {
		for (x = x0; x < x0 + len; ++x) {
			draw_units(map, x, y0);
		}
	}
}

int tmpdrawlighting;

hex_overlay(x, y)
int x, y;
{
	extern int sunx, suny;

	if (!tmpdrawlighting) return (-1);
	return (((world.daylength == 1)
			 || (distance(wrapx(x), y, sunx, suny) < world.circumference / 2)) ? -1 : 0);
}

int
hex_pattern(x, y, power)
int x, y, power;
{
	return (terrain_visible(x, y) ? terrain_at(wrapx(x), y) : -1);
}

#define dontdraw 0
#define useblocks 1
#define usepictures 2
#define usepolygons 3

hex_style(x, y, power)
int x, y, power;
{
	return (terrain_visible(x, y)  ?
			(power >= 4 ? usepolygons : useblocks) :
			dontdraw);
}

hex_update(map, x, y)
Map *map;
int x, y;
{
	int sx, sy;
	Rect tmprect;

	if (!in_area(x, y)) return FALSE;
	xform(map, x, y, &sx, &sy);
	SetRect(&tmprect, sx, sy, sx + map->hw, sy + map->hh);
	return (RectInRgn(&tmprect, map->window->visRgn));
}

/* Draw an entire row of terrain, possibly with a single rectangle fill. */
 
/* x0 may be negative. */

draw_terrain_row(map, x0, y0, len)
Map *map;
int x0, y0, len;
{
	int empty;
	int x, x1, t, sx, sy, w, h, p = map->power;
	int dogrid = map->drawgrid, dofill = map->drawhexpats;
	int i = 0;
	int /* enum whattouse */ style, segstyle;
	int pat, segpat;
	int over, segover;
	int update, segupdate;
	Rect tmprect, rect;
	Image *timg;
	
	tmpdrawlighting = map->drawlighting;
	x1 = x0;
	segstyle = hex_style(x0, y0, p);
	segpat = hex_pattern(x0, y0, p);
	segover = hex_overlay(x0, y0);
	segupdate = hex_update(map, x0, y0);
	for (x = x0; x < x0 + len + 1; ++x) {
		style = hex_style(x, y0, p);
		pat = hex_pattern(x, y0, p);
		over = hex_overlay(x, y0);
		update = hex_update(map, x, y0);
		/* Decide if the run is over and we need to dump some output. */
		if (x == x0+len
			|| style != segstyle
			|| pat != segpat
			|| over != segover
			|| update != segupdate
			|| segstyle == usepictures
			|| segstyle == usepolygons) {
			/* don't draw anything that would match the window's bg */
			if (segupdate && segstyle != dontdraw) {
				xform(map, x1, y0, &sx, &sy);
				switch (segstyle) {
					case useblocks:
						draw_hex_block(sx, sy, i, map->power, segpat, segover);
						break;
					case usepictures:
						break;
					case usepolygons:
						draw_hex_polygon(sx, sy, map->power, dogrid, segpat, segover);
						/* Assume that only polygon scale can fit an elev number. */
						if (elevations_defined() && map->drawelevations) {
							draw_elevation(sx, sy, map->power, elev_at(x1, y0));
						}
						if (DebugG && !g_see_all()) {
							sx += (map->power > 5 ? map->hw/2 : 2);
							sy += map->hh/2;
							sprintf(spbuf, "*%d*", cover(dside, x1, y0));
							draw_legend_text(sx, sy, map->hh/2, spbuf, 0);
						}
				}
			}
			i = 0;
			x1 = x;
			segstyle = style;
			segpat = pat;
			segover = over;
			segupdate = update;
		}
		++i;
	}
}

/* Indicate the elevation of the given location, textually for now. */

draw_elevation(sx, sy, power, elev)
int sx, sy, power, elev;
{
	/* Decide whether to center value in hex or put close to left edge (so it fits). */
	sx += (power > 5 ? hws[power]/2 : 2);
	sy += hhs[power]/2;
	if (elev != 0) {
		sprintf(spbuf, "%d", elev);
		draw_legend_text(sx, sy, hhs[power]/2, spbuf, 0);
	}
	/* (Also draw contour lines eventually) */
}

/* Draw three borders of the given hex. */

#define any_borders_at(x, y, b) (aux_terrain_at(x, y, b) != 0)

draw_ul_borders(map, x, y, b)
Map *map;
int x, y, b;
{
	int sx, sy;
	
 	if (!terrain_visible(x, y) || !any_borders_at(x, y, b)) return;
 	xform(map, x, y, &sx, &sy);
	if (border_at(x, y, WEST, b)
		&& seen_border(dside, x, y, WEST)) {
		draw_border_line(sx, sy, WEST, map->power, b);
	}
	if (border_at(x, y, NW, b)
		&& seen_border(dside, x, y, NW)) {
		draw_border_line(sx, sy, NW,   map->power, b);
	}
	if (border_at(x, y, NE, b)
		&& seen_border(dside, x, y, NE)) {
		draw_border_line(sx, sy, NE,   map->power, b);
	}
}

draw_all_borders(map, x, y, b)
Map *map;
int x, y, b;
{
	int sx, sy;
	
 	if (!terrain_visible(x, y) || !any_borders_at(x, y, b)) return;
	draw_ul_borders(map, x, y, b);
	xform(map, x, y, &sx, &sy);
	if (border_at(x, y, EAST, b)
		&& seen_border(dside, x, y, EAST)) {
		draw_border_line(sx, sy, EAST, map->power, b);
	}
	if (border_at(x, y, SE,   b)
		&& seen_border(dside, x, y, SE)) {
		draw_border_line(sx, sy, SE,   map->power, b);
	}
	if (border_at(x, y, SW,   b)
		&& seen_border(dside, x, y, SW)) {
		draw_border_line(sx, sy, SW,   map->power, b);
	}
}

/* Draw three connections of the given hex. */

/* Actually this draws six half-connections.  It (should) also only draws the
   connection if the underlying terrain is different. */

#define any_connections_at(x, y, c) (aux_terrain_at(x, y, c) != 0)

draw_connections(map, x, y, c)
Map *map;
int x, y, c;
{
	int sx, sy, dir, t = terrain_at(x, y);
	
	if (!terrain_visible(x, y) || !any_connections_at(x, y, c)) return;
	xform(map, x, y, &sx, &sy);
	for_all_directions(dir) {
		if (connection_at(x, y, dir, c)) {
			draw_connection_line_half(sx, sy, dir, map->power, c);
		}
	}
}

/* Draw all the units visible in the given cell. */

draw_units(map, x, y)
Map *map;
int x, y;
{
	short uview;
	int sx, sy, sw, sh, u, s;
	int terr = terrain_at(wrapx(x), y);
	Side *side2;
	Unit *unit;
	int num;

	if (units_visible(x, y) && (num = num_units_at(x, y)) > 0) {
		if (map->uw <= 16) {
			xform(map, x, y, &sx, &sy);
			/* Adjust to unit part of cell. */
			sx += (map->hw - map->uw) / 2;  sy += (map->hh - map->uh) / 2;
			if ((unit = unit_at(x, y)) != NULL) {
				draw_unit_image(map->window, sx, sy, map->uw, map->uh,
								unit->type, side_number(unit->side), !completed(unit));
				if (map->drawnames) draw_unit_name(unit, sx, sy, map->uw, map->uh); 
				/* (should indicate presence of other units in some way) */
			}
		} else {
			for_all_stack(x, y, unit) {
				xform_unit(map, unit, &sx, &sy, &sw, &sh);
				draw_unit_and_occs(map, unit, sx, sy, sw, sh);
			}
		}
	} else {
		if ((uview = unit_view(dside, x, y)) != EMPTY) {
			u = vtype(uview);
			s = vside(uview);
			/* a hack, this should never happen in the first place */
			if (s == side_number(dside)) {
				return;
			}
			xform(map, x, y, &sx, &sy);
			/* Adjust to unit part of cell. */
			sx += (map->hw - map->uw) / 2;  sy += (map->hh - map->uh) / 2;
			draw_unit_image(map->window, sx, sy, map->uw, map->uh, u, s, 0);
		}
	}
}

draw_unit_and_occs(map, unit, sx, sy, sw, sh)
Map *map;
Unit *unit;
int sx, sy, sw, sh;
{
	int u = unit->type, s = side_number(unit->side), sx2, sy2, sw2, sh2;
	Unit *occ;
	Rect tmprect;

	if (unit->occupant == NULL) {
		draw_unit_image(map->window, sx, sy, sw, sh, u, s, !completed(unit));
		if (map->drawnames) draw_unit_name(unit, sx, sy, sw, sh); 
	} else {
		xform_occupant(map, unit, unit, sx, sy, sw, sh, &sx2, &sy2, &sw2, &sh2);
		/* Draw a sort of "grouping rectangle". */
		SetRect(&tmprect, sx, sy, sx+sw, sy+sh);
		FillRect(&tmprect, QD(white));
		FrameRect(&tmprect);
		draw_unit_image(map->window, sx2, sy2, sw2, sh2, u, s, !completed(unit));
		if (map->drawnames) draw_unit_name(unit, sx2, sy2, sw2, sh2); 
		for_all_occupants(unit, occ) {
			xform_occupant(map, unit, occ, sx, sy, sw, sh, &sx2, &sy2, &sw2, &sh2);
			draw_unit_and_occs(map, occ, sx2, sy2, sw2, sh2);
		}
	}
}

/* Draw a unit's name or number. */

draw_unit_name(unit, sx, sy, sw, sh)
Unit *unit;
int sx, sy, sw, sh;
{
	char legend[BUFSIZE];

	if (unit->name != NULL) {
		strcpy(legend, unit->name);
	} else if (unit->number > 0) {
		sprintf(legend, "%d%s", unit->number, ordinal_suffix(unit->number));
	} else {
		strcpy(legend, "");
	}
	if (strlen(legend) > 0) {
		draw_legend_text(sx + sw + 1, sy + sh/2, sh, legend, 1);
	}
}

/* Indicate what kind of people are living in the given hex. */

/* This can look at edge cells, which should always have a pop of NOBODY. */

draw_people(map, x, y)
Map *map;
int x, y;
{
	int pop, sx, sy, sw, sh, ex, ey, ew, eh, dir, x1, y1, pop1, bordercell = FALSE;
	Side *side2;
	
	if (terrain_visible(x, y)) {
		pop = people_side_at(wrapx(x), y);
		side2 = side_n(pop);
		xform(map, x, y, &sx, &sy);
		/* Decide which edges are borders of the country. */
		for_all_directions(dir) {
			if (point_in_dir(x, y, dir, &x1, &y1)) {
				pop1 = people_side_at(wrapx(x1), y1);
				if (pop != pop1) {
					if (pop1 != NOBODY) {
						draw_country_border_line(sx, sy, dir, map->power);
					} else {
						/* should draw in gray instead? */
						draw_country_border_line(sx, sy, dir, map->power);
					}
					bordercell = TRUE;
				}
			}
		}
		/* Draw an emblem for the people in the cell. */
		if (map->drawpeople && bordercell) {
			sw = hws[map->power];  sh = hhs[map->power];
			ew = min(sw, max(8, sw / 4));  eh = min(sh, max(8, sh / 4));
			ex = sx + sw / 2 - ew / 2;  ey = sy + sh / 2 - eh / 2;
			draw_side_emblem(map->window, ex, ey, ew, eh, side2);
		}
	}
}

/* This draws a small set of bar charts, one for each material type. */

draw_materials(map, x, y)
Map *map;
int x, y;
{
	int m, sx, sy, mx, my, mw;
	Rect graphrect;
	
	if (nummtypes == 0) return;
	mw = map->uw / nummtypes /* should be count of displayed materials... */;
	if (mw < 1) return;
	xform(map, x, y, &sx, &sy);
	mx = sx + (map->hw - map->uw) / 2;  my = sy + (map->hh - map->uh) / 2;
	for_all_material_types(m) {
		if (map->drawmaterials[m]) {
			SetRect(&graphrect, mx + m * mw, my, mx + (m + 1) * mw, my + map->uh);
			FrameRect(&graphrect);
		}
	}
}

/* Draw any text that should be associated with this hex. */

/* (could precompute what the string will lap over and move or truncate str),
   should be deterministic for each mag, so redraw doesn't scramble */
/* do geofeatures, label at a hex with nothing else, and declared as the
   feature's "center" */

draw_legend(map, x, y)
Map *map;
int x, y;
{
	int xx, numhexes = 1, color, sx, sy, pixlen;
	char legend[BUFSIZE];
	Unit *unit;
	extern GeoFeature *feature_at();
	GeoFeature *feature;

	/* Draw the name of a terrain feature. */
	/* (should limit to one cell of feature, preferably centering on quasi-centroid) */	
	if (terrain_visible(x, y)) {
		if ((feature = feature_at(x, y)) != NULL) {
			if (1 /* feature->size > 0  make sure this is calced! */) {
				if (1 /* center of feature */) {
					strcpy(legend, feature_name_at(x, y));
					xform(map, x, y, &sx, &sy);
					draw_legend_text(sx + map->hw/2, sy + map->hh/2, map->uh, legend, 0);
				}
			}
		}
	}
}

draw_blast(unit, es, hit)
Unit *unit;
Side *es;
int hit;
{
	int ux = unit->x, uy = unit->y, sx, sy, sw, sh, i, btype;
	Side *us = unit->side, *side;
	Map *map;
	GrafPtr oldport;
	RgnHandle tmprgn;
	
	btype = ((hit >= unit->hp) ? 2 : ((hit > 0) ? 1 : 0));
	if (active_display(dside) && units_visible(ux, uy)) {
		GetPort(&oldport);
		for_all_maps(map) {
			/* (should only do if x,y is actually visible...) */
			/* Set up the drawing context. */
			SetPort(map->window);
			tmprgn = NewRgn();
			GetClip(tmprgn);
			ClipRect(&(map->contentrect));
			xform_unit(map, unit, &sx, &sy, &sw, &sh);
			draw_blast_image(map->window, sx, sy, sw, sh, btype);
			SetClip(tmprgn);
			DisposeRgn(tmprgn);
		}
		SetPort(oldport);
	}
}

/* Draw all the selections of all the units. */

draw_selections(map)
Map *map;
{
	int sx, sy, sx2, sy2, i;
	GrafPtr oldport;
	Rect tmprect;
	RgnHandle tmprgn;
	Unit *unit;

 	GetPort(&oldport);
	SetPort(map->window);
	tmprgn = NewRgn();
	GetClip(tmprgn);
	ClipRect(&(map->contentrect));
	for (i = 0; i < map->numselections; ++i) {
		unit = map->selections[i];
		draw_selected_unit(map, unit);
	}
	SetClip(tmprgn);
	DisposeRgn(tmprgn);
	SetPort(oldport);
}

/* Draw all the selected units in the given cell. */

draw_selections_at(map, x, y)
Map *map;
int x, y;
{
	int i;
	GrafPtr oldport;
	RgnHandle tmprgn;
	Unit *unit;

 	GetPort(&oldport);
	SetPort(map->window);
	tmprgn = NewRgn();
	GetClip(tmprgn);
	ClipRect(&(map->contentrect));
	for (i = 0; i < map->numselections; ++i) {
		unit = map->selections[i];
		if (unit && unit->x == x && unit->y == y) {
			draw_selected_unit(map, unit);
		}
	}
	SetClip(tmprgn);
	DisposeRgn(tmprgn);
	SetPort(oldport);
}

draw_selected_unit_setport(map, unit)
Map *map;
Unit *unit;
{
	int i;
	GrafPtr oldport;
	RgnHandle tmprgn;

 	GetPort(&oldport);
	SetPort(map->window);
	tmprgn = NewRgn();
	GetClip(tmprgn);
	ClipRect(&(map->contentrect));
	draw_selected_unit(map, unit);
	SetClip(tmprgn);
	DisposeRgn(tmprgn);
	SetPort(oldport);
}

/* Draw a single selected unit on the given map.  Assumes that grafport already set. */

draw_selected_unit(map, unit)
Map *map;
Unit *unit;
{
	int sx, sy, sw, sh;
	Rect tmprect;

	if (unit && unit->x >= 0 && unit->y >= 0 /* should be "is_visible" */) {
		if (map->power >= 5) {
			xform_unit_self(map, unit, &sx, &sy, &sw, &sh);
		} else {
			xform(map, unit->x, unit->y, &sx, &sy);
			/* Adjust to unit part of cell. */
			sx += (map->hw - map->uw) / 2;  sy += (map->hh - map->uh) / 2;
			sw = map->uw;  sh = map->uh;
		}
		/* Be sure the selected unit is drawn. */
		draw_unit_image(map->window, sx, sy, sw, sh,
						unit->type, side_number(unit->side), !completed(unit));
		/* Draw a highlighting rectangle. */
		SetRect(&tmprect, sx + 1, sy + 1, sx + sw - 1, sy + sh - 1);
		/* Black is for units that can still act, dark gray for actors, gray if the
		   unit can't do anything. */
		PenPat((unit->act && unit->act->initacp > 0) ?
				((unit->act->acp > 0) ? QD(black) : QD(dkGray)) : QD(gray));
		PenSize(2, 2);
		FrameRect(&tmprect);
		PenNormal();
#if 0
		/* (should do somewhere else) */
		if (unit->plan
			&& unit->plan->tasks
			&& unit->plan->tasks->type == MOVETO_TASK) {
			Task *task = unit->plan->tasks;
			xform(map, task->x, task->y, &sx2, &sy2);
			MoveTo(sx+map->hw/2, sy+map->hh/2);
			LineTo(sx2+map->hw/2, sy2+map->hh/2);
		}
#endif
		DGprintf("show selection of %s at %d,%d\n", unit_desig(unit), unit->x, unit->y);
	}
}

erase_selections(map)
Map *map;
{
	int sx, sy, sx2, sy2, i;
	GrafPtr oldport;
	RgnHandle tmprgn;
	Unit *unit;

 	GetPort(&oldport);
	SetPort(map->window);
	tmprgn = NewRgn();
	GetClip(tmprgn);
	ClipRect(&(map->contentrect));
	for (i = 0; i < map->numselections; ++i) {
		unit = map->selections[i];
		draw_unselected_unit(map, unit);
	}
	SetClip(tmprgn);
	DisposeRgn(tmprgn);
	SetPort(oldport);
}

erase_selection(map, unit)
Map *map;
Unit *unit;
{
	GrafPtr oldport;
	RgnHandle tmprgn;

 	GetPort(&oldport);
	SetPort(map->window);
	tmprgn = NewRgn();
	GetClip(tmprgn);
	ClipRect(&(map->contentrect));
	draw_unselected_unit(map, unit);
	SetClip(tmprgn);
	DisposeRgn(tmprgn);
	SetPort(oldport);
}

draw_unselected_unit(map, unit)
Map *map;
Unit *unit;
{
	if (unit != NULL && inside_area(unit->x, unit->y) /* should be "is_visible" */) {
		draw_row(map, unit->x, unit->y, 1, TRUE);
	}
}

force_map_update(map)
Map *map;
{
	force_update(map->window);
}

/* Remove and destroy the map object. */

destroy_map(map)
Map *map;
{
	Map *map2;
	
	if (maplist == map) {
		maplist = map->next;
	} else {
		for_all_maps(map2) {
			if (map2->next == map) {
				map2->next = map->next;
			}
		}
	}
	/* (should destroy substructs) */
	free(map);
}

/* (should be able to mention borders and conns also) */

oneliner(map, sx, sy)
Map *map;
int sx, sy;
{
	int x, y, t, uview, u, s;
	Unit *unit;

	nearest_cell(map, sx, sy, &x, &y);
	t = terrain_at(x, y);
	if (!in_area(x, y)) {
		sprintf(tmpbuf, "(nothing)");
		return;
	} else if (!terrain_visible(x, y)) {
		sprintf(tmpbuf, "(unknown)");
	} else if (!units_visible(x, y)) {
		if ((uview = unit_view(dside, x, y)) != EMPTY) {
			u = vtype(uview);  s = vside(uview);
			sprintf(tmpbuf, "%s %s (in %s)",
					side_adjective(side_n(s)), u_type_name(u), t_type_name(t));
		} else {
			sprintf(tmpbuf, "%s", t_type_name(t));
		}
	} else {
		nearest_unit(map, sx, sy, &unit);
		if (unit != NULL) {
			sprintf(tmpbuf, "%s %s (in %s)",
					side_name(unit->side), u_type_name(unit->type), t_type_name(t));
		} else {
			sprintf(tmpbuf, "%s", t_type_name(t));
		}
	}
	sprintf(tmpbuf+strlen(tmpbuf), " (%d,%d)", x, y);
}

activate_map(map, activate)
Map *map;
int activate;
{
	Rect growRect;

	if (activate) {
		HiliteControl(map->vscrollbar, 0);
		HiliteControl(map->hscrollbar, 0);
#if 0
		/* the controls must be redrawn on activation: */
		(*(map->vscrollbar))->contrlVis = 255;
		(*(map->hscrollbar))->contrlVis = 255;
		InvalRect(&(*(map->vscrollbar))->contrlRect);
		InvalRect(&(*(map->hscrollbar))->contrlRect);
#endif
		/* The growbox needs to be redrawn on activation. */
		growRect = map->window->portRect;
		/* adjust for the scrollbars */
		growRect.top = growRect.bottom - sbarwid;
		growRect.left = growRect.right - sbarwid;
		InvalRect(&growRect);
	} else {
		/* the controls must be hidden on deactivation: */
		HiliteControl(map->vscrollbar, 255);
		HiliteControl(map->hscrollbar, 255);
/*		HideControl(map->vscrollbar);
		HideControl(map->hscrollbar); */
		/* the growbox should be changed immediately on deactivation: */
		DrawGrowIcon(map->window);
	}
}

