/* Copyright (c) 1989, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Interpret the forms in a game module. */

/* Syntax is el cheapo Lisp. */

#include "conq.h"

extern int compromised;

struct a_key {
    char *name;
    enum keywords key;
    short value;
} keywordtable[] = {

#undef  DEF_KWD
#define DEF_KWD(CODE,NAME,VALUE)  { NAME, CODE, VALUE },

#include "keyword.def"

    { NULL, 0 }
};

keyword_value(str)
char *str;
{
    int i;

    /* (should do a radix search first, then switch to exhaustive) */
    for (i = 0; keywordtable[i].name != NULL; ++i) {
	if (strcmp(str, keywordtable[i].name) == 0) return keywordtable[i].key;
    }
    return (-1);
}

char *
value_keyword(k)
enum keywords k;
{
    return keywordtable[k].name;
}

#define TYPEPROP(TYPES, N, DEFNS, I, TYPE)  \
  ((TYPE *) &(((char *) (&(TYPES[N])))[DEFNS[I].offset]))[0]

/* This is a generic syntax check and escape. */

#define SYNTAX(X,TEST,MSG)  \
  if (!(TEST)) { syntax_error((X), (MSG));  return; }
  
syntax_error(x, msg)
Obj *x;
char *msg;
{
    char foobuf[BUFSIZE];

    sprintlisp(foobuf, x);
    init_warning("syntax error in %s - %s", foobuf, msg);
}

#define TYPECHECK(PRED,X,MSG)  \
  if (!PRED(X)) { type_error((X), (MSG));  return; }

type_error(x, msg)
Obj *x;
char *msg;
{
    char foobuf[BUFSIZE];

    sprintlisp(foobuf, x);
    init_warning("type error in %s - %s", foobuf, msg);
}

#define PARSE_PROPERTY(BDG,NAME,VAL)  \
  SYNTAX(bdg, consp(bdg) && symbolp(car(bdg)), "property binding");  \
  (NAME) = c_string(car(BDG));  \
  (VAL) = cadr(BDG);

extern Obj *synthlist;
extern Obj *sidenamelist;

/* These variables indicate whether new types can still be defined.
   Once a table or list of types is manipulated, these are turned off. */

int canaddutype = TRUE;
int canaddmtype = TRUE;
int canaddttype = TRUE;

/* True if game will start up in the middle of a turn. */

int midturnrestore = FALSE;

/* This is the basic interpreter of a form appearing in a module. */

interp_form(module, form)
Module *module;
Obj *form;
{
    Obj *thecar, *rest;
    char *name, *str;

    if (consp(form) && symbolp(thecar = car(form))) {
	name = c_string(thecar);
	if (Debug) {
	    /* If in a module, report the line number(s) of a form. */
	    if (module != NULL) {
		Dprintf("Line %d", module->startlineno);
		if (module->endlineno != module->startlineno) {
		    Dprintf("-%d", module->endlineno);
		}
	    }
	    Dprintf(": (%s ", name);
	    Dprintlisp(cadr(form));
	    if (cddr(form) != lispnil) {
		Dprintf(" ");
		Dprintlisp(caddr(form));
		if (cdr(cddr(form)) != lispnil) {
		    Dprintf(" ...");
		}
	    }
	    Dprintf(")\n");
	}
	switch (keyword_value(name)) {
	  case K_GAME_MODULE:
	    interp_game_module(form, module);
	    break;
#ifndef SPECIAL
	  case K_UNIT_TYPE:
	    interp_utype(form);  break;
	  case K_MATERIAL_TYPE:
	    interp_mtype(form);  break;
	  case K_TERRAIN_TYPE:
	    interp_ttype(form);  break;
	  case K_TABLE:
	    interp_table(form, TRUE);  break;
          case K_DEFINE:
	    interp_new_variable(form);  break;
	  case K_SET:
	    /* alternate syntax for defn? */
	    interp_variable(form);  break;
#if 0
	  case K_APPEND:
	    break;
	  case K_REMOVE:
	    break;
#endif
	  case K_ADD:
	    add_stuff(form);  break;
#endif /* n SPECIAL */
	  case K_WORLD:
	    interp_world(form);  break;
	  case K_AREA:
	    interp_area(form);  break;
	  case K_SIDE:
	    interp_side(form, NULL);  break;
	  case K_INDEPENDENT_UNITS:
	    interp_side(form, indepside);  break;
	  case K_PLAYER:
	    interp_player(form);  break;
	  case K_AGREEMENT:
	    interp_agreement(form);  break;
	  case K_INCLUDE:
	    include_module(form, module);  break;
	  case K_SCOREKEEPER:
	    interp_scorekeeper(form);  break;
	  case K_EVT:
	    interp_history(form);  break;
	  case K_UNIT:
	    /* This is for when the unit name matches a keyword */
	    if (consp(cdr(form))) {
		interp_unit(cdr(form));  break;
	    } else {
		/* broken actually */
	    }
	    break;
	  case K_UNIT_DEFAULTS:
	    /* defines how to interpret succeeding unit defns */
	  case K_SYNTHESIS_METHODS:
	    interp_synth_setter(form);  break;
	  case K_SIDE_LIBRARY:
	    interp_side_library(form);  break;
	  case K_NAMER:
	    interp_namer(form);  break;
	  case K_PRINT:
#ifdef USE_CONSOLE
	    printlisp(cadr(form));
	    if (symbolp(cadr(form))) {
		if (boundp(cadr(form))) {
		    printf(" -> ");
		    printlisp(symbol_value(cadr(form)));
		} else {
		    printf(" <unbound>");
		}
	    }
	    printf("\n");
#else
	    /* should do some sort of random user alert thingy */
#endif /* USE_CONSOLE */
	    break;
	  default:
	    if (utype_from_name(name) != NONUTYPE) {
		interp_unit(form);
	    } else {
		useless_form_warning(module, form);
	    }
	}
    } else {
	useless_form_warning(module, form);
    }
}

useless_form_warning(module, form)
Module *module;
Obj *form;
{
    char buf[BUFSIZE];

    sprintlisp(buf, form);
    if (module) {
	init_warning("%d-%d: A useless form: %s",
		     module->startlineno, module->endlineno, buf);
    } else {
	init_warning("Useless input form: %s", buf);
    }
}

/* Inclusion is half-module-like, not strictly textual. */

include_module(form, module)
Obj *form;
Module *module;
{
    char *name;
    Module *submodule;

    if (!(symbolp(cadr(form)) || stringp(cadr(form)))) {
	/* syntax error */
	return;
    }
    name = c_string(cadr(form));
    Dprintf("Including \"%s\" ...\n", name);
    submodule = add_game_module(name, module);
    load_game_module(submodule);
    Dprintf("... Done including \"%s\".\n", name);
}

/* Digest the form defining the module as a whole. */

interp_game_module(form, module)
Obj *form;
Module *module;
{
    char *name = NULL, *propname;
    Obj *props = cdr(form), *bdg, *tmp, *propval;

    if (module == NULL) return;  /* why is this here? */

    if (stringp(car(props))) {
	name = c_string(car(props));
	props = cdr(props);
    }
    if (name != NULL) {
	if (module->name == NULL || strlen(module->name) == 0) {
	    module->name = name;
	} else {
	    if (strcmp(name, module->name) != 0) {
		init_warning(
   "Module's name \"%s\" does not match declared name \"%s\", ignoring declared name",
			     module->name, name);
	    }
	}
    }
    for (; props != lispnil; props = cdr(props)) {
	bdg = car(props);
	PARSE_PROPERTY(bdg, propname, propval);
	switch (keyword_value(propname)) {
	  case K_TITLE:
	    module->title = c_string(propval);  break;
	  case K_BLURB:
	    module->blurb = c_string(propval);  break;
	  case K_PICTURE_NAME:
	    module->picturename = c_string(propval);  break;
	  case K_DEFAULT_INCLUDE:
	    module->defaultinclude = c_string(propval);  break;
	  case K_INSTRUCTIONS:
	    /* The instructions are a list of strings. */
	    module->instructions = cadr(bdg);  break;
	  case K_VARIANTS:
	    /* The variants are a list of lists. */
	    module->variants = cdr(bdg);  break;
	  case K_NOTES:
	    /* The player notes are a list of strings. */
	    module->notes = cadr(bdg);  break;
	  case K_DESIGN_NOTES:
	    /* The design notes are a list of strings. */
	    module->designnotes = cadr(bdg);  break;
	  case K_VERSION:
	    module->version = c_string(propval);  break;
	  case K_PROGRAM_VERSION:
	    module->programversion = c_string(propval);  break;
	  default:
	    unknown_property("game module", module->name, propname);
	}
    }
    /* Should be smarter about earlier vs later versions. */
    if (module->programversion != NULL
	&& strcmp(module->programversion, version) != 0) {
	/* This should become some sort of alert on some systems. */
	init_warning(
	      "Ahem, the module \"%s\" is claimed to be for version \"%s\",",
		     module->name, module->programversion);
	init_warning("   but you are running version \"%s\".", version);
	init_warning("Continuing anyway, but beware!\n");
    }
}

/* The following code is unneeded if all the types have been compiled in. */

#ifndef SPECIAL

/* Create a new type of unit and fill in info about it. */

interp_utype(form)
Obj *form;
{
    int u;
    Obj *name = cadr(form), *utype;

    TYPECHECK(symbolp, name, "unit-type name not a symbol");
    if (!canaddutype)
      init_warning("Should not be defining more unit types");
    if (numutypes < MAXUTYPES) {
	u = numutypes++;
	utype = new_utype(u);
	/* Set default values for the unit type's props first. */
	/* Any default type name shouldn't confuse the code below. */
	default_unit_type(u);
	setq(name, utype);
	/* Set the values of random props. */
	fill_in_utype(u, cddr(form));
	/* If no internal type name string given, use the defined name. */
	if (empty_string(u_internal_name(u))) {
	    set_u_internal_name(u, c_string(name));
	}
	if (empty_string(u_type_name(u))) {
	    set_u_type_name(u, u_internal_name(u));
	}
    } else {
	too_many_types("unit", MAXUTYPES, name);
    }
    /* Blast any cached list of types. */
    makunbound(intern_symbol("u*"));
    eval_symbol(intern_symbol("u*"));
}

/* Trudge through assorted props, filling them in. */

fill_in_utype(u, list)
int u;
Obj *list;
{
    int i, found;
    char *propname;
    Obj *bdg, *propsym, *val;

    for ( ; list != lispnil; list = cdr(list)) {
	bdg = car(list);
	PARSE_PROPERTY(bdg, propname, val);
	found = FALSE;
	for (i = 0; utypedefns[i].name != NULL; ++i) {
	    if (strcmp(propname, utypedefns[i].name) == 0) {
		if (utypedefns[i].intgetter) {
		    TYPEPROP(utypes, u, utypedefns, i, short) = c_number(val);
		} else if (utypedefns[i].strgetter) {
		    TYPEPROP(utypes, u, utypedefns, i, char *) = c_string(val);
		} else {
		    TYPEPROP(utypes, u, utypedefns, i, Obj *) = cadr(bdg);
		}
		found = TRUE;
		break;
	    }
	}
	if (!found) unknown_property("unit type", u_type_name(u), propname);
    }
}

/* Create a new type of material or modify the props of pre-existing. */

interp_mtype(form)
Obj *form;
{
    int m;
    Obj *name = cadr(form), *mtype;
    char *namestring;
    
    TYPECHECK(symbolp, name, "material-type name not a symbol");
    if (!canaddmtype)
      init_warning("Should not be defining more material types");
    if (nummtypes < MAXMTYPES) {
	m = nummtypes++;
	mtype = new_mtype(m);
	/* Set default values for the material type's properties first. */
	default_material_type(m);
	setq(name, mtype);
	/* Set the values of random props. */
	fill_in_mtype(m, cddr(form));
	/* If no type name string given, use the defined name. */
	if (empty_string(m_type_name(m))) {
	    namestring = c_string(name);
	    set_m_type_name(m, namestring);
	}
    } else {
	too_many_types("material", MAXMTYPES, name);
    }
    /* Blast and remake any cached list of types. */
    makunbound(intern_symbol("m*"));
    eval_symbol(intern_symbol("m*"));
}

/* Go through a list of prop name/value pairs and fill in the
   material type description from them. */

fill_in_mtype(m, list)
int m;
Obj *list;
{
    int i, found;
    Obj *bdg, *val;
    char *propname;

    for ( ; list != lispnil; list = cdr(list)) {
	bdg = car(list);
	PARSE_PROPERTY(bdg, propname, val);
	found = FALSE;
	for (i = 0; mtypedefns[i].name != NULL; ++i) {
	    if (strcmp(propname, mtypedefns[i].name) == 0) {
		if (mtypedefns[i].intgetter) {
		    TYPEPROP(mtypes, m, mtypedefns, i, short) = c_number(val);
		} else if (mtypedefns[i].strgetter) {
		    TYPEPROP(mtypes, m, mtypedefns, i, char *) = c_string(val);
		} else {
		    TYPEPROP(mtypes, m, mtypedefns, i, Obj *) = cadr(bdg);
		}
		found = TRUE;
		break;
	    }
	}
	if (!found) unknown_property("material type", m_type_name(m), propname);
    }
}

/* Declare a new type of terrain and fill in info about it. */

interp_ttype(form)
Obj *form;
{
    int t;
    Obj *name = cadr(form), *ttype;

    TYPECHECK(symbolp, name, "terrain-type name not a symbol");
    if (!canaddttype)
      init_warning("Should not be defining more terrain types");
    if (numttypes < MAXTTYPES) {
	t = numttypes++;
	ttype = new_ttype(t);
	/* Set default values for the terrain type's props first. */
	default_terrain_type(t);
	setq(name, ttype);
	/* If no type name string given, use the defined name. */
	if (empty_string(t_type_name(t))) {
	    set_t_type_name(t, c_string(name));
	}
	/* Set the values of random props. */
	fill_in_ttype(t, cddr(form));
    } else {
	too_many_types("terrain", MAXTTYPES, name);
    }
    /* Blast and remake any cached list of all types. */
    makunbound(intern_symbol("t*"));
    eval_symbol(intern_symbol("t*"));
}

/* Go through a list of prop name/value pairs and fill in the
   terrain type description from them. */

fill_in_ttype(t, list)
int t;
Obj *list;
{
    int i, found;
    char *propname;
    Obj *bdg, *val;

    for ( ; list != lispnil; list = cdr(list)) {
	bdg = car(list);
	PARSE_PROPERTY(bdg, propname, val);
	found = FALSE;
	for (i = 0; ttypedefns[i].name != NULL; ++i) {
	    if (strcmp(propname, ttypedefns[i].name) == 0) {
		if (ttypedefns[i].intgetter) {
		    TYPEPROP(ttypes, t, ttypedefns, i, short) = c_number(val);
		} else if (ttypedefns[i].strgetter) {
		    TYPEPROP(ttypes, t, ttypedefns, i, char *) = c_string(val);
		} else {
		    TYPEPROP(ttypes, t, ttypedefns, i, Obj *) = cadr(bdg);
		}
		found = TRUE;
		break;
	    }
	}
	if (!found) unknown_property("terrain type", t_type_name(t), propname);
    }
    /* Recalculate the count of subtypes. */
    count_terrain_subtypes();
}

/* Fill in a table. */

interp_table(form, reset)
Obj *form;
int reset;
{
    int i, found, lim1, lim2;
    Obj *formsym = cadr(form), *body = cddr(form);
    char *tablename;

    TYPECHECK(symbolp, formsym, "table name not a symbol");
    tablename = c_string(formsym);
    found = FALSE;
    for (i = 0; tabledefns[i].name != NULL; ++i) {
	if (strcmp(tablename, tabledefns[i].name) == 0) {
	    allocate_table(i, reset);
	    update_table(i, body, FALSE);
	    found = TRUE;
	    break;
	}
    }
    if (!found) init_warning( "Undefined table \"%s\"", tablename);
}

/* Given a table and a list of value-setting clauses, fill in the table. */

#define INDEXP(typ, x) \
  ((typ == UTYP) ? utypep(x) : ((typ == MTYP) ? mtypep(x) : ttypep(x)))

#define nonlist(x) (!consp(x) && x != lispnil)

#define CHECK_INDEX_1(tbl, x)  \
  if (!INDEXP(tabledefns[tbl].index1, (x))) {  \
      init_warning("index 1 to table %s has wrong type",  \
		   tabledefns[tbl].name);  \
      return;  \
  }

#define CHECK_INDEX_2(tbl, x)  \
  if (!INDEXP(tabledefns[tbl].index2, (x))) {  \
      init_warning("index 2 to table %s has wrong type",  \
		   tabledefns[tbl].name);  \
      return;  \
  }


update_table(tbl, clauses, init)
int tbl, init;
Obj *clauses;
{
    int i, j, num, lim1, lim2;
    Obj *clause, *indexes1, *indexes2, *values, *tmp1, *tmp2, *value, *subvalue;

    lim1 = numtypes_from_index_type(tabledefns[tbl].index1);
    lim2 = numtypes_from_index_type(tabledefns[tbl].index2);
    for ( ; clauses != lispnil; clauses = cdr(clauses)) {
	clause = car(clauses);
	switch (clause->type) {
	  case SYMBOL:
	    clause = eval_symbol(clause);
	    TYPECHECK(numberp, clause, "table clause does not eval to number");
	    /* Now treat it as a number. */
	  case NUMBER:
	    /* a constant value - blast over everything */
	    num = c_number(clause);
	    for (i = 0; i < lim1 * lim2; ++i) {
		(*(tabledefns[tbl].table))[i] = num;
	    }
	    break;
	  case CONS:
	    /* Evaluate the three parts of a clause. */
	    indexes1 = eval(car(clause));
	    indexes2 = eval(cadr(clause));
	    values = eval(caddr(clause));
	    if (nonlist(indexes1)) {
		CHECK_INDEX_1(tbl, indexes1);
		i = c_number(indexes1);
		if (nonlist(indexes2) ) {
		    CHECK_INDEX_2(tbl, indexes2);
		    j = c_number(indexes2);
		    value = (consp(values) ? car(values) : values);
		    TYPECHECK(numberp, value, "value in table must be a number");
		    num = c_number(value);
		    (*(tabledefns[tbl].table))[lim2 * i + j] = num;
		} else {
		    for (tmp2 = indexes2; tmp2 != lispnil; tmp2 = cdr(tmp2)) {
			CHECK_INDEX_2(tbl, car(tmp2));
			j = c_number(car(tmp2));
			value = (consp(values) ? car(values) : values);
			TYPECHECK(numberp, value, "value in table must be a number");
			num = c_number(value);
			(*(tabledefns[tbl].table))[lim2 * i + j] = num;
			if (consp(values) && cdr(values) != lispnil) {
			    values = cdr(values);
			}
		    }
		}
	    } else {
		for (tmp1 = indexes1; tmp1 != lispnil; tmp1 = cdr(tmp1)) {
		    CHECK_INDEX_1(tbl, car(tmp1));
		    i = c_number(car(tmp1));
		    value = (consp(values) ? car(values) : values);
		    if (nonlist(indexes2)) {
			CHECK_INDEX_2(tbl, indexes2);
			j = c_number(indexes2);
			TYPECHECK(numberp, value, "value in table must be a number");
			num = c_number(value);
			(*(tabledefns[tbl].table))[lim2 * i + j] = num;
		    } else {
		    	for (tmp2 = indexes2; tmp2 != lispnil; tmp2 = cdr(tmp2)) {
			    CHECK_INDEX_2(tbl, car(tmp2));
			    j = c_number(car(tmp2));
			    subvalue = (consp(value) ? car(value) : value);
			    TYPECHECK(numberp, subvalue, "value in table must be a number");
			    num = c_number(subvalue);
			    (*(tabledefns[tbl].table))[lim2 * i + j] = num;
			    if (consp(value)) {
			    	value = cdr(value);
			    }
		        }
		    }
		    if (consp(values)) {
			values = cdr(values);
		    }
		}
	    }
	    break;
	  case STRING:
	    break; /* error? */
	  default:
	    /* who knows? */
	    break;
	}
    }
}

/* Use this to declare a new variable, warning if it is bound already. */

interp_new_variable(form)
Obj *form;
{
    Obj *var = cadr(form);
    Obj *val = caddr(form);

    if (symbolp(var)) {
	if (boundp(var)) {
	    init_warning("Symbol \"%s\" has been bound already, overwriting",
			 c_string(var));
	}
	setq(var, eval(val));
    } else {
	init_warning("Can't define a non-symbol!");
    }
}

/* Set the binding of an existing known variable. */

interp_variable(form)
Obj *form;
{
    Obj *var = cadr(form);
    Obj *val = eval(caddr(form));
    char *name;

    if (!symbolp(var)) {
	init_warning("Can't set a non-symbol!");
	return;
    }
    name = c_string(var);

#undef  DEF_VAR_I
#define DEF_VAR_I(STR,fname,SETFNAME,doc,var,lo,dflt,hi)  \
    if (strcmp(name, STR) == 0)  \
      { SETFNAME(c_number(val));  return; }
#undef  DEF_VAR_S
#define DEF_VAR_S(STR,fname,SETFNAME,doc,var,dflt)  \
    if (strcmp(name, STR) == 0)  \
      { SETFNAME(c_string(val));  return; }
#undef  DEF_VAR_L
#define DEF_VAR_L(STR,fname,SETFNAME,doc,var)  \
    if (strcmp(name, STR) == 0)  \
      { SETFNAME(val);  return; }

#include "gvar.def"

    /* Try as a random symbol. */
    if (symbolp(var) && boundp(var)) {
	setq(var, val);
	return;
    }
    /* Out of luck. */
    init_warning("Can't set unknown global named \"%s\"", name);
}

/* General function to augment objects. */

add_stuff(form)
Obj *form;
{
    Obj *subclass = cadr(form);

    if (symbolp(subclass) && keyword_value(c_string(subclass)) == K_TABLE) {
	interp_table(cdr(form), FALSE);
    } else {
	subclass = eval(subclass);
	if (utypep(subclass)) {
	    add_to_utypes(cons(subclass, lispnil),
			  car(cddr(form)),
			  eval(cdr(cddr(form))));
	} else if (consp(subclass) && utypep(car(subclass))) {
	    add_to_utypes(subclass,
			  car(cddr(form)),
			  eval(cadr(cddr(form))));
	} else if (mtypep(subclass)) {
	    add_to_mtypes(cons(subclass, lispnil),
			  car(cddr(form)),
			  eval(cdr(cddr(form))));
	} else if (consp(subclass) && mtypep(car(subclass))) {
	    add_to_mtypes(subclass,
			  car(cddr(form)),
			  eval(cadr(cddr(form))));
	} else if (ttypep(subclass)) {
	    add_to_ttypes(cons(subclass, lispnil),
			  car(cddr(form)),
			  eval(cdr(cddr(form))));
	} else if (consp(subclass) && ttypep(car(subclass))) {
	    add_to_ttypes(subclass,
			  car(cddr(form)),
			  eval(cadr(cddr(form))));
	} else {
	    char buf[BUFSIZE];

	    sprintlisp(buf, form);
	    init_warning("What *is* this? - %s", buf);
	}
    }
}

add_to_utypes(types, prop, values)
Obj *types, *prop, *values;
{
    Obj *lis1, *lis2;

    if (utypep(types)) {
	fill_in_utype(types->v.num,
		      cons(cons(prop, cons(values, lispnil)), lispnil));
    } else if (consp(types)) {
	if (consp(values)) {
	    if (length(types) != length(values)) {
	    	sprintlisp(spbuf, prop);
	    	init_warning("non-matching lists for property %s", spbuf);
	    }
	    for (lis1 = types, lis2 = values;
		 lis1 != lispnil && lis2 != lispnil;
		 lis1 = cdr(lis1), lis2 = cdr(lis2)) {
		TYPECHECK(utypep, car(lis1), "not a unit type");
		fill_in_utype(car(lis1)->v.num,
			      cons(cons(prop, cons(car(lis2), lispnil)),
				   lispnil));
	    }
	} else {
	    for (lis1 = types; lis1 != lispnil; lis1 = cdr(lis1)) {
		TYPECHECK(utypep, car(lis1), "not a unit type");
		fill_in_utype(car(lis1)->v.num,
			      cons(cons(prop, cons(values, lispnil)),
				   lispnil));
	    }
	}
    }
}

add_to_mtypes(types, prop, values)
Obj *types, *prop, *values;
{
    Obj *lis1, *lis2;

    if (mtypep(types)) {
	fill_in_mtype(types->v.num,
		      cons(cons(prop, cons(values, lispnil)), lispnil));
    } else if (consp(types)) {
	if (consp(values)) {
	    if (length(types) != length(values)) {
	    	sprintlisp(spbuf, prop);
	    	init_warning("non-matching lists for property %s", spbuf);
	    }
	    for (lis1 = types, lis2 = values;
		 lis1 != lispnil && lis2 != lispnil;
		 lis1 = cdr(lis1), lis2 = cdr(lis2)) {
		TYPECHECK(mtypep, car(lis1), "not a unit type");
		fill_in_mtype(car(lis1)->v.num,
			      cons(cons(prop, cons(car(lis2), lispnil)),
				   lispnil));
	    }
	} else {
	    for (lis1 = types; lis1 != lispnil; lis1 = cdr(lis1)) {
		TYPECHECK(mtypep, car(lis1), "not a unit type");
		fill_in_mtype(car(lis1)->v.num,
			      cons(cons(prop, cons(values, lispnil)),
				   lispnil));
	    }
	}
    }
}

add_to_ttypes(types, prop, values)
Obj *types, *prop, *values;
{
    Obj *lis1, *lis2;

    if (ttypep(types)) {
	fill_in_ttype(types->v.num,
		      cons(cons(prop, cons(values, lispnil)), lispnil));
    } else if (consp(types)) {
	if (consp(values)) {
	    if (length(types) != length(values)) {
	    	sprintlisp(spbuf, prop);
	    	init_warning("non-matching lists for property %s", spbuf);
	    }
	    for (lis1 = types, lis2 = values;
		 lis1 != lispnil && lis2 != lispnil;
		 lis1 = cdr(lis1), lis2 = cdr(lis2)) {
		TYPECHECK(ttypep, car(lis1), "not a terrain type");
		fill_in_ttype(car(lis1)->v.num,
			      cons(cons(prop, cons(car(lis2), lispnil)),
				   lispnil));
	    }
	} else {
	    for (lis1 = types; lis1 != lispnil; lis1 = cdr(lis1)) {
		TYPECHECK(ttypep, car(lis1), "not a terrain type");
		fill_in_ttype(car(lis1)->v.num,
			      cons(cons(prop, cons(values, lispnil)),
				   lispnil));
	    }
	}
    }
}

#endif /* n SPECIAL */

/* Interpret a world-specifying form. */

interp_world(form)
Obj *form;
{
    int newworld = FALSE;
    Obj *props, *bdg, *propval;
    char *propname;

    props = cdr(form);
    if (symbolp(car(props))) {
	/* This is the id of the world (eventually). */
	props = cdr(props);
    }
    if (numberp(car(props))) {
    	set_world_circumference(c_number(car(props)), TRUE);
	props = cdr(props);
    }
    for ( ; props != lispnil; props = cdr(props)) {
	bdg = car(props);
	PARSE_PROPERTY(bdg, propname, propval);
	switch (keyword_value(propname)) {
	  case K_CIRCUMFERENCE:
	    world.circumference = c_number(propval);  break;
	  case K_DAY_LENGTH:
	    world.daylength = c_number(propval);  break;
	  case K_YEAR_LENGTH:
	    world.yearlength = c_number(propval);  break;
	  default:
	    unknown_property("world", "", propname);
	}
    }
}

/* Only one area, of fixed size.  Created anew if shape/size is supplied, else
   just modified. */

interp_area(form)
Obj *form;
{
    int newarea = FALSE, newwidth = 0, newheight = 0;
    Obj *props, *bdg, *propval;
    char *propname;

    props = cdr(form);
    /* (eventually this will be an id or name) */
    if (symbolp(car(props))) {
	props = cdr(props);
	newarea = TRUE;
    }
    if (numberp(car(props))) {
	newwidth = newheight = c_number(car(props));
	props = cdr(props);
	newarea = TRUE;
    }
    if (numberp(car(props))) {
	newheight = c_number(car(props));
	props = cdr(props);
	newarea = TRUE;
    }
    if (newarea) set_area_shape(newwidth, newheight, TRUE);
    for ( ; props != lispnil; props = cdr(props)) {
	bdg = car(props);
	PARSE_PROPERTY(bdg, propname, propval);
	switch (keyword_value(propname)) {
	  case K_WIDTH:
	    area.width = c_number(propval);  break;
	  case K_HEIGHT:
	    area.height = c_number(propval);  break;
	  case K_LATITUDE:
	    area.latitude = c_number(propval);  break;
	  case K_LONGITUDE:
	    area.longitude = c_number(propval);  break;
	  case K_CELL_WIDTH:
	    area.cellwidth = c_number(propval);  break;
	  case K_TERRAIN:
	    fill_in_terrain(cdr(bdg));  break;
	  case K_AUX_TERRAIN:
	    fill_in_aux_terrain(cdr(bdg));  break;
	  case K_PEOPLE_SIDES:
	    fill_in_people_sides(cdr(bdg));  break;
	  case K_FEATURES:
	    fill_in_features(cdr(bdg));  break;
	  case K_ELEVATIONS:
	    fill_in_elevations(cdr(bdg));  break;
	  case K_MATERIAL:
	    fill_in_cell_material(cdr(bdg));  break;
	  case K_TEMPERATURES:
	    fill_in_temperatures(cdr(bdg));  break;
	  default:
	    unknown_property("area", "", propname);
	}
    }
}

/* The general plan of reading is similar for all layers - create a blank
   layer if necessary, then call read_layer and pass it a function that will
   actually put a value into a cell of the layer.  We need to define those
   functions because most of the setters are macros, and because we can do
   some extra error checking. */

/* Read the area terrain. */

int numbadterrain = 0;

fn_set_terrain_at(x, y, val)
int x, y, val;
{
    /* It's important not to put bad values into the terrain layer. */
    if (!is_terrain_type(val)) {
    	/* Only warn the first few times, just count thereafter. */
    	if (numbadterrain < 10) {
	    init_warning("Unknown terrain type (%d) at %d,%d, substituting %s",
			 val, x, y, t_type_name(0));
	}
	val = 0;
	++numbadterrain;
    }
    set_terrain_at(x, y, val);
}

fill_in_terrain(contents)
Obj *contents;
{
    /* We must have some terrain types or we're screwed. */
    if (numttypes == 0) load_standard_game();
    numbadterrain = 0;
    /* Make sure the terrain layer exists. */
    if (!terrain_defined()) allocate_area_terrain();
    read_layer(contents, fn_set_terrain_at);
    if (numbadterrain > 0) {
    	init_warning("%d occurrences of unknown terrain in all.", numbadterrain);
    }
}

/* Read a layer of auxiliary terrain. */

fn_set_aux_terrain_at(x, y, val)
int x, y, val;
{
    /* (should check values here too?) */
    set_aux_terrain_at(x, y, tmpttype, val);
}

fill_in_aux_terrain(contents)
Obj *contents;
{
    int t;
    Obj *typesym = car(contents), *typeval;

    if (symbolp(typesym) && ttypep(typeval = eval(typesym))) {
	t = c_number(typeval);
	contents = cdr(contents);
	/* Make sure aux terrain space exists, but don't overwrite. */
	allocate_area_aux_terrain(t);
	tmpttype = t;
	read_layer(contents, fn_set_aux_terrain_at);
	/* Ensure that borders and connections have all their bits
	   correctly set. */
	patch_linear_terrain(t);
    } else {
	/* not a valid aux terrain type */
    }
}

fn_set_people_side_at(x, y, val)
int x, y, val;
{
    set_people_side_at(x, y, val);
}

fill_in_people_sides(contents)
Obj *contents;
{
    /* Make sure the people sides layer exists. */
    if (!people_sides_defined()) allocate_area_people_sides();
    read_layer(contents, fn_set_people_side_at);
}

/* This should recompute size etc of all these features too. */

fn_set_raw_feature_at(x, y, val)
int x, y, val;
{
    set_raw_feature_at(x, y, val);
}

fill_in_features(contents)
Obj *contents;
{
    Obj *featspec, *flist;
    GeoFeature *feat;

    init_geo_features();
    for (flist = car(contents); flist != lispnil; flist = cdr(flist)) {
	featspec = car(flist);
	feat = create_feature(c_string(cadr(featspec)),
			      c_string(caddr(featspec)));
	feat->id = c_number(car(featspec));
    }
    read_layer(cdr(contents), fn_set_raw_feature_at);
}

fn_set_elevation_at(x, y, val)
int x, y, val;
{
    set_elev_at(x, y, val);
}

fill_in_elevations(contents)
Obj *contents;
{
    /* Make sure the elevation layer exists. */
    if (!elevations_defined()) allocate_area_elevations();
    read_layer(contents, fn_set_elevation_at);
}

fn_set_material_at(x, y, val)
int x, y, val;
{
    set_material_at(x, y, tmpmtype, val);
}

fill_in_cell_material(contents)
Obj *contents;
{
    int m;
    Obj *typesym = car(contents), *typeval;

    if (symbolp(typesym) && mtypep(typeval = eval(typesym))) {
	m = c_number(typeval);
	contents = cdr(contents);
	/* Make sure material space exists. */
	allocate_area_material(m);
	tmpmtype = m;
	read_layer(contents, fn_set_material_at);
    } else {
	/* not a valid material type spec, should warn */
    }
}

fn_set_temperature_at(x, y, val)
int x, y, val;
{
    set_temperature_at(x, y, val);
}

fill_in_temperatures(contents)
Obj *contents;
{
    /* Make sure the temperature layer exists. */
    if (!temperatures_defined()) allocate_area_temperatures();
    read_layer(contents, fn_set_temperature_at);
}

fill_in_clouds(contents)
Obj *contents;
{
	init_warning("can't read clouds layer yet");
}

fill_in_winds(contents)
Obj *contents;
{
	init_warning("can't read winds layer yet");
}

/* Interpret a side spec. */

interp_side(form, side)
Obj *form;
Side *side;
{
    int id = -1;
    Obj *ident = lispnil, *props = cdr(form);

    /* See if there's an optional side identifier and pick it off. */
    if (props != lispnil && !consp(car(props))) {
	ident = car(props);
	props = cdr(props);
    }
    if (numberp(ident)) {
	id = c_number(ident);
	side = side_n(id);
    } else {
	/* We want to create a new side. */
    }
    if (side == NULL) {
	side = create_side();
    }
    if (id >= 0) {
    	/* (should worry about id conflicts) */
    	side->id = id;
    }
    fill_in_side(side, props, FALSE);
    Dprintf("  Got side %s\n", side_desig(side));
}

/* Given a side, fill in some of its props. */

fill_in_side(side, props, userdata)
Side *side;
Obj *props;
int userdata;
{
    int s, u, numval = 0;
    char *propname, *strval = NULL;
    Obj *bdg, *tmp, *propval;

    for (; props != lispnil; props = cdr(props)) {
	bdg = car(props);
	PARSE_PROPERTY(bdg, propname, propval);
	if (numberp(propval)) numval = c_number(propval);
	if (stringp(propval)) strval = c_string(propval);
	/* (should check if user-settable property by looking at keyword flag) */
	switch (keyword_value(propname)) {
	  case K_NAME:
	    check_name_uniqueness(side, strval, "name");
	    side->name = strval;  break;
	  case K_LONG_NAME:
	    check_name_uniqueness(side, strval, "long name");
	    side->longname = strval;  break;
	  case K_SHORT_NAME:
	    check_name_uniqueness(side, strval, "short name");
	    side->shortname = strval;  break;
	  case K_NOUN:
	    check_name_uniqueness(side, strval, "noun");
	    side->noun = strval;  break;
	  case K_PLURAL_NOUN:
	    check_name_uniqueness(side, strval, "plural noun");
	    side->pluralnoun = strval;  break;
	  case K_ADJECTIVE:
	    check_name_uniqueness(side, strval, "adjective");
	    side->adjective = strval;  break;
          /* Several synonyms are allowed for specifying colors. */
	  case K_COLOR:
	  case K_COLORS:
	  case K_COLOR_SCHEME:
	    side->colorscheme = strval;  break;
	  case K_EMBLEM_NAME:
	    side->emblemname = strval;  break;
	  case K_UNIT_NAMERS:
	    for (tmp = cdr(bdg); tmp != lispnil; tmp = cdr(tmp)) {
/*		side->tech[u++] = c_number(car(tmp)); */
	    }
	    break;
	  case K_FEATURE_NAMERS:
	    for (tmp = cdr(bdg); tmp != lispnil; tmp = cdr(tmp)) {
/*		side->tech[u++] = c_number(car(tmp)); */
	    }
	    break;

	  case K_TASK_LIMIT:
	    side->tasklimit = numval;  break;
	  case K_RESPECT_NEUTRALITY:
	    side->respectneutrality = numval;  break;
	  case K_REAL_TIMEOUT:
	    side->realtimeout = numval;  break;
	  case K_START_BEEP_TIME:
	    side->startbeeptime = numval;  break;
	  case K_WILLING_TO_DRAW:
	    side->willingtodraw = numval;  break;
	  case K_TRUSTS:
	    s = 0;
	    for (tmp = cdr(bdg); tmp != lispnil; tmp = cdr(tmp)) {
		if (s > g_sides_max()) break;
		side->trusts[s++] = c_number(car(tmp));
	    }
	    break;
	  case K_TRADES:
	    s = 0;
	    for (tmp = cdr(bdg); tmp != lispnil; tmp = cdr(tmp)) {
		if (s > g_sides_max()) break;
		side->trades[s++] = c_number(car(tmp));
	    }
	    break;
	  case K_DOCTRINES:
	    read_utype_doctrine(side, cadr(bdg), cddr(bdg));  break;
	  case K_UI_PREFS:
	    /* should pass to interface routine */
	    break;

	  case K_NAMES_LOCKED:
	    side->nameslocked = numval;  break;
	  case K_CLASS:
/*	    if (userdata) { init_warning("No permission to set property");  break; } */
	    side->sideclass = strval;  break;
	  case K_ACTIVE:
	    side->ingame = numval;  break;
	  case K_STATUS:
	    side->status = numval;  break;
	  case K_PLAYER:
	    /* This will be replaced by the actual pointer later. */
	    side->player = (Player *) numval;  break;
	  case K_CONTROLLED_BY:
	    /* This will be replaced by the actual pointer later. */
	    side->controlledby = (Side *) numval;  break;
	  case K_SELF_UNIT:
	    /* This will be replaced by the actual pointer later. */
	    side->selfunit = (Unit *) numval;  break;
	  case K_TURN_TIME_USED:
	    side->turntimeused = numval;  break;
	  case K_TOTAL_TIME_USED:
	    side->totaltimeused = numval;  break;
	  case K_TIMEOUTS:
	    side->timeouts = numval;  break;
	  case K_TIMEOUTS_USED:
	    side->timeoutsused = numval;  break;
	  case K_FINISHED_TURN:
	    side->finishedturn = numval;  break;
	  case K_DESIGNER:
#ifdef DESIGNERS
	    /* Don't need to call become_designer here. */
	    side->designer = numval;
	    ++numdesigners;
	    compromised = TRUE;
#endif /* DESIGNERS */
	    break;
	  case K_START_WITH:
	    if (side->startwith == NULL) {
		side->startwith = (short *) xmalloc(numutypes * sizeof(short));
	    }
	    u = 0;
	    for (tmp = cdr(bdg); tmp != lispnil; tmp = cdr(tmp)) {
		if (u >= numutypes) break;
		side->startwith[u++] = c_number(car(tmp));
	    }
	    break;
	  case K_NEXT_NUMBERS:
	    u = 0;
	    for (tmp = cdr(bdg); tmp != lispnil; tmp = cdr(tmp)) {
		if (u >= numutypes) break;
		side->counts[u++] = c_number(car(tmp));
	    }
	    break;
	  case K_TECH:
	    u = 0;
	    for (tmp = cdr(bdg); tmp != lispnil; tmp = cdr(tmp)) {
		if (u >= numutypes) break;
		side->tech[u++] = c_number(car(tmp));
	    }
	    break;
	  case K_INIT_TECH:
	    u = 0;
	    for (tmp = cdr(bdg); tmp != lispnil; tmp = cdr(tmp)) {
		if (u >= numutypes) break;
		side->inittech[u++] = c_number(car(tmp));
	    }
	    break;
	  case K_SCORES:
	    /* This will be patched up later, after scorekeepers exist. */
	    side->scores = (short *) cdr(bdg);  break;
	  case K_TERRAIN_VIEW:
	    read_terrain_view(side, cdr(bdg));  break;
	  case K_UNIT_VIEW:
	    read_unit_view(side, cdr(bdg));  break;
	  case K_UNIT_VIEW_DATES:
	    read_unit_view_dates(side, cdr(bdg));  break;
	  default:
	    unknown_property("side", side_desig(side), propname);
	}
    }
    if (side->noun != NULL && side->pluralnoun == NULL) {
	side->pluralnoun = copy_string(plural_form(side->noun));
    }
}

check_name_uniqueness(side, str, kind)
Side *side;
char *str, *kind;
{
    if (name_in_use(side, str)) {
	init_warning("Side %s \"%s\" is already in use", kind, str);
    }
}

/* Helper function to init side view from rle encoding. */

fn_set_terrain_view(x, y, val)
int x, y, val;
{
    set_terrain_view(tmpside, x, y, val);
}

read_terrain_view(side, contents)
Side *side;
Obj *contents;
{
    tmpside = side;

    init_view(side);
    read_layer(contents, fn_set_terrain_view);
}

fn_set_unit_view(x, y, val)
int x, y, val;
{
    set_unit_view(tmpside, x, y, val);
}

read_unit_view(side, contents)
Side *side;
Obj *contents;
{
    tmpside = side;

    init_view(side);
    read_layer(contents, fn_set_unit_view);
}

fn_set_unit_view_date(x, y, val)
int x, y, val;
{
/*    set_unit_view(tmpside, x, y, val); */
}

read_unit_view_dates(side, contents)
Side *side;
Obj *contents;
{
    tmpside = side;

/*    init_view(side); only if unit view missing? */
    read_layer(contents, fn_set_unit_view_date);
}

/* Read doctrine info pertaining to a particular unit type. */

read_utype_doctrine(side, ulist, props)
Side *side;
Obj *ulist, *props;
{
    int u = 0;
    char *propname;
    Obj *bdg, *val;

    if (!consp(ulist)) ulist = cons(ulist, lispnil);

    ulist = eval(ulist);
    if (numberp(car(ulist))) u = c_number(car(ulist));

    for (; props != lispnil; props = cdr(props)) {
	bdg = car(props);
	PARSE_PROPERTY(bdg, propname, val);
	switch (keyword_value(propname)) {
	  case K_EVER_ASK_SIDE:
	    u_doctrine(side, u, everaskside) = c_number(val);  break;
	  case K_AVOID_BAD_TERRAIN:
	    u_doctrine(side, u, avoidbadterrain) = c_number(val);  break;
#if 0
	  case K_EXPLORE_INPORTANCE:
	    u_doctrine(side, u, exploreimportance) = c_number(val);  break;
#endif
	  default:
	    unknown_property("utype doctrine", "", propname);
	}
    }
}

interp_player(form)
Obj *form;
{
    int id = -1;
    Obj *ident = lispnil, *props = cdr(form);
    Player *player = NULL;

    if (props != lispnil) {
	if (!consp(car(props))) {
	    ident = car(props);
	    props = cdr(props);
	}
    }
    if (numberp(ident)) {
	id = c_number(ident);
	player = find_player(id);
    }
    if (player == NULL) {
	player = add_player();
    }
    if (id > 0) player->id = id;
    fill_in_player(player, props);
    Dprintf("  Got player %s\n", player_desig(player));
}

fill_in_player(player, props)
Player *player;
Obj *props;
{
    char *propname;
    Obj *bdg, *tmp, *propval;

    for (; props != lispnil; props = cdr(props)) {
	bdg = car(props);
	PARSE_PROPERTY(bdg, propname, propval);
	switch (keyword_value(propname)) {
	  case K_NAME:
	    player->name = c_string(propval);  break;
	  case K_CONFIG_NAME:
	    player->configname = c_string(propval);  break;
	  case K_DISPLAY_NAME:
	    player->displayname = c_string(propval);  break;
	  case K_AI_TYPE_NAME:
	    player->aitypename = c_string(propval);  break;
	  case K_INITIAL_STRENGTH:
	    player->strength = c_number(propval);  break;
	  default:
	    unknown_property("player", player_desig(player), propname);
	}
    }
    canonicalize_player(player);
}

interp_agreement(form)
Obj *form;
{
    int id;
    char *propname;
    Obj *props = cdr(form), *agid, *bdg, *val;
    Agreement *ag;

    agid = car(props);
    if (numberp(agid)) {
	/* use the number eventually */
	props = cdr(props);
    }
    if (1 /* must create a new agreement object */) {
	ag = create_agreement();
	ag->active = TRUE;  /* default for now */
/*	ag->signers = 0;
	ag->willing = 0;
	ag->knownto = 0; */
    }
    /* Interpret the properties. */
    for (; props != lispnil; props = cdr(props)) {
	bdg = car(props);
	PARSE_PROPERTY(bdg, propname, val);
	switch (keyword_value(propname)) {
	  case K_TYPE_NAME:
	    ag->typename = c_string(val);  break;
	  case K_NAME:
	    ag->name = c_string(val);  break;
	  case K_ACTIVE:
	    ag->active = c_number(val);  break;
	  case K_TERMS:
	    ag->terms = val;  break;
	  case K_SIGNERS:
	    break;
	  default:
	    unknown_property("agreement", "", propname);
	}
    }
}

ttype_from_name(str)
char *str;
{
    int t;

    for_all_terrain_types(t) {
	if (strcmp(str, t_type_name(t)) == 0)
	  return t;
    }
    return NONTTYPE;
}

/* Try to find a unit type named by the string. */

utype_from_name(str)
char *str;
{
    int u;

    for_all_unit_types(u) {
	if (strcmp(str, u_type_name(u)) == 0
	    || (u_short_name(u) && strcmp(str, u_short_name(u)) == 0)
	    || (u_long_name(u) && strcmp(str, u_long_name(u)) == 0))
	  return u;
    }
    /* Try evaluating the symbol too. */
    if (boundp(intern_symbol(str))
	&& symbol_value(intern_symbol(str))->type == UTYPE) {
	return symbol_value(intern_symbol(str))->v.num;
    }
    return NONUTYPE;
}

/* This creates an individual unit and fills in data about it. */

interp_unit(form)
Obj *form;
{
    int type = utype_from_name(c_string(car(form)));
    int nusn = 0, nuid = 0, nuidorig, nutsn, nut = -1, m;
    Obj *props = cdr(form), *bdg, *tmp, *val;
    char *propname;
    Unit *unit;
    extern int nextid;

    /* We must have some unit types or we're screwed. */
    if (numutypes == 0) load_standard_game();
    Dprintf("Restoring a unit, type %d, from ", type);
    Dprintlisp(form);
    Dprintf("\n");
    if (type != NONUTYPE) {
	if ((unit = create_unit(type)) != NULL) {
	    init_supply(unit);  /* should do this later? */
	} else {
	    /* Bad, but not disastrous. */
	    /* (unless restoring a game?) */
	    init_warning("Failed to create a unit!");
	    return;
	}
    } else {
	/* worry about this one? */
    }
    if (numberp(car(props))) {
	unit->prevx = c_number(car(props));
	props = cdr(props);
    }
    if (numberp(car(props))) {
	unit->prevy = c_number(car(props));
	props = cdr(props);
    }
    if (numberp(car(props))) {
	nusn = c_number(car(props));
	props = cdr(props);
    }
    /* Now crunch through optional stuff. */
    /* This unit's props should already be correct or correctable, since
       none of the props below may be found. */
    for (; props != lispnil; props = cdr(props)) {
	bdg = car(props);
	PARSE_PROPERTY(bdg, propname, val);
	switch (keyword_value(propname)) {
	  case K_N:
	    unit->name = c_string(val);  break;
	  case K_SHARP:
	    nuid = c_number(val);  break;
	  case K_S:
	    nusn = c_number(val);  break;
	  case K_AT:
	    unit->prevx = c_number(val);
	    unit->prevy = c_number(cadr(cdr(bdg)));
	    break;
	  case K_NM:
	    unit->number = c_number(val);  break;
	  case K_CP:
	    unit->cp = c_number(val);  break;
	  case K_HP:
	    unit->hp = c_number(val);  break;
	  case K_CXP:
	    unit->cxp = c_number(val);  break;
	  case K_M:
	    /* Will get weird if nummtypes doesn't match, but won't fail. */
	    tmp = cdr(bdg);
	    for_all_material_types(m) {
		if (tmp != lispnil) {
		    unit->supply[m] = c_number(car(tmp));
		    tmp = cdr(tmp);
		}
	    }
	    break;
	  case K_IN:
	    nut = c_number(val);  break;
	  case K_ACT:
	    interp_unit_act(unit, cdr(bdg));  break;
	  case K_PLAN:
	    interp_unit_plan(unit, cdr(bdg));  break;
	  case K_Z:
	    /* (should be prevz?) */
	    unit->z = c_number(val);  break;
	  case K_X:
	    unit->hook = cdr(bdg);  break;
	  default:
	    unknown_property("unit", unit_desig(unit), propname);
	}
    }
    /* If the unit id was given, assign it to the unit, avoiding duplication. */
    if (nuid > 0) {
    	nuidorig = nuid;
    	/* If this id is already in use, try the next one. */
    	while (find_unit(nuid) != NULL) ++nuid;
    	/* Guaranteed distinct, safe to use. */
	unit->id = nuid;
	/* (should stash nuidorig in a slot, use to recalc references) */
	/* Ensure that future random ids won't step on this one. */
	nextid = max(nextid, nuid + 1);
    }
    /* Sleazy trick - will fill in actual transport ptr when all units in. */
    unit->transport = (Unit *) nut;
    /* Set the side of this unit if it is not independent. */
    if (nusn > 0) {
	set_unit_side(unit, side_n(nusn));
    }
    Dprintf("  Got %s\n", unit_desig(unit));
}

interp_unit_act(unit, props)
Unit *unit;
Obj *props;
{
    Obj *bdg, *val;
    char *propname;

    for (; props != lispnil; props = cdr(props)) {
	bdg = car(props);
	PARSE_PROPERTY(bdg, propname, val);
	switch (keyword_value(propname)) {
	  case K_ACP:
	    unit->act->acp = c_number(val);  break;
	  default:
	    unknown_property("unit act", unit_desig(unit), propname);
	}
    }
}

interp_unit_plan(unit, props)
Unit *unit;
Obj *props;
{
    Obj *bdg, *val, *plantype, *tasks;
    char *propname;

	if (unit->plan == NULL) return;  /* should create perhaps? */
	plantype = car(props);
	/* unit->plan->type = lookup_plan_type(plantype); */
    props = cdr(props);
    for (; props != lispnil; props = cdr(props)) {
	bdg = car(props);
	PARSE_PROPERTY(bdg, propname, val);
	switch (keyword_value(propname)) {
	  case K_GOAL:
	  	/* (should read a goal here) */
	  	break;
	  case K_TASKS:
	    for (tasks = cdr(bdg); tasks != lispnil; tasks = cdr(tasks)) {
	    	interp_task(unit, car(tasks));
	    }
	    break;
	  default:
	    unknown_property("unit plan", unit_desig(unit), propname);
	}
    }
}

interp_task(unit, form)
Unit *unit;
Obj *form;
{
	Obj *tasktype;

	tasktype = car(form);
}

/* Manipulate the synthesis method list. */

interp_synth_setter(form)
Obj *form;
{
    Obj *op = cadr(form), *methlist = cddr(form), *tmp;
    char *opstr = c_string(op);

    if (strcmp(opstr, "set") == 0) {
	synthlist = methlist;
    } else if (strcmp(opstr, "append") == 0) {
	for (tmp = methlist; tmp != lispnil; tmp = cdr(tmp)) {
	}
    } else if (strcmp(opstr, "prepend") == 0) {
	for (tmp = methlist; tmp != lispnil; tmp = cdr(tmp)) {
	}
    } else if (strcmp(opstr, "remove") == 0) {
	for (tmp = methlist; tmp != lispnil; tmp = cdr(tmp)) {
	}
    } else {
	init_warning("Unrecognized synth-methods operation `%s'", opstr);
    }
}

/* Manipulate the side library list. */

interp_side_library(form)
Obj *form;
{
    Obj *op = cadr(form), *methlist = cddr(form), *tmp;
    char *opstr = c_string(op);
    extern int totalsideweights;

    if (strcmp(opstr, "set") == 0) {
    totalsideweights = 0;
	sidenamelist = methlist;
    } else if (strcmp(opstr, "append") == 0) {
	for (tmp = methlist; tmp != lispnil; tmp = cdr(tmp)) {
	}
    } else if (strcmp(opstr, "prepend") == 0) {
	for (tmp = methlist; tmp != lispnil; tmp = cdr(tmp)) {
	}
    } else if (strcmp(opstr, "remove") == 0) {
	for (tmp = methlist; tmp != lispnil; tmp = cdr(tmp)) {
	}
    } else {
	init_warning("Unrecognized side-names operation `%s'", opstr);
    }
}

/* Make a namer from the form. */

interp_namer(form)
Obj *form;
{
    Obj *id = cadr(form), *meth = car(cddr(form));

    if (symbolp(id)) {
	setq(id, make_namer(meth));
    }
}

/* Make a scorekeeper from the given form. */

interp_scorekeeper(form)
Obj *form;
{
    int id = 0;
    char *propname;
    Obj *props = cdr(form), *bdg, *val;
    Scorekeeper *sk = NULL;

    if (numberp(car(props))) {
	id = c_number(car(props));
	props = cdr(props);
    }
    if (id > 0) {
	/* (should attempt to find scorekeeper) */
    }
    /* Create a new scorekeeper object if necessary. */
    if (sk == NULL) {
	sk = create_scorekeeper();
	if (id > 0) {
		/* set id of scorekeeper */
	}
    }
    /* Interpret the properties. */
    for (; props != lispnil; props = cdr(props)) {
	bdg = car(props);
	PARSE_PROPERTY(bdg, propname, val);
	switch (keyword_value(propname)) {
	  case K_NAME:
	    sk->name = c_string(val);  break;
	  case K_WHEN:
	    sk->when = val;  break;
	  case K_APPLIES_TO:
	    sk->who = val;  break;
	  case K_TRIGGER:
	    sk->trigger = val;  break;
	  case K_DO:
	    sk->body = val;  break;
	  case K_TRIGGERED:
	    sk->triggered = c_number(val);  break;
	  case K_INITIAL:
	    sk->initial = c_number(val);  break;
	  case K_NOTES:
	    sk->notes = val;  break;
	  default:
	    unknown_property("scorekeeper", "", propname);
	}
    }
}

/* Make a historical event from the form. */

interp_history(form)
Obj *form;
{
}

/* isn't this redundant? */

r_in_area(x, y)
int x, y;
{
    return (between(0, x, area.width-1) && between(0, y, area.height-1));
}

/* Designer is trying to define too many different types. */

too_many_types(typename, maxnum, name)
char *typename;
int maxnum;
Obj *name;
{
    init_warning("Limited to %d types of %s!\n", maxnum, typename);
    sprintlisp(spbuf, name);
    init_warning("(Failed to create type with name `%s')", spbuf);
}

/* Prop name is unknown, either misspelled or misapplied. */

unknown_property(type, inst, name)
char *type, *inst, *name;
{
    init_warning("The %s form %s has no property named %s", type, inst, name);
}

int layerusedefault;
int layerdefault;
int layermultiplier;
int layeradder;
int layerareax, layerareay;
int layerareawidth, layerareaheight;

int ignorespecials;

read_layer(contents, setter)
Obj *contents;
int (*setter)();
{
    int i, t;
    char *str;
    short chartable[256];
    char buf[BUFSIZE];
    Obj *rest, *desc, *nml, *val;

    layerusedefault = FALSE;
    layerdefault = 0;
    layermultiplier = 1;
    layeradder = 0;
    layerareax = layerareay = 0;
    layerareawidth = area.width;  layerareaheight = area.height;
    ignorespecials = FALSE;
    for (rest = contents; rest != lispnil; rest = cdr(rest)) {
	desc = car(rest);
	if (stringp(desc)) {
	    /* Read according to the default (exact literal) encoding. */
	    read_rle(rest, setter, NULL);
	    return;
	} else if (symbolp(desc)) {
	    switch (keyword_value(c_string(desc))) {
	      case K_EXACT:
		/* Read according to the default (exact literal) encoding. */
		read_rle(cdr(rest), setter, NULL);
		return;
	      case K_EVERYWHERE:
		/* should set to a constant value */
		return;
	    }
	} else if (consp(desc) && symbolp(car(desc))) {
	    switch (keyword_value(c_string(car(desc)))) {
	      case K_SUBAREA:
		break;
	      case K_XFORM:
		layermultiplier = c_number(cadr(desc));
		layeradder = c_number(caddr(desc));
		break;
	      case K_BY_BITS:
		break;
	      case K_BY_CHAR:
		/* Assign each char to its corresponding index. */
		for (i = 0; i < 255; ++i) chartable[i] = 0;
		str = c_string(cadr(desc));
		for (i = 0; i < strlen(str); ++i) {
		    chartable[str[i]] = i;
		    /* If special chars in by-char string, flag it. */
		    if (str[i] == '*' || str[i] == ',') ignorespecials = TRUE;
		}
		/* Now read it all in. */
		read_rle(cdr(rest), setter, chartable);
		return;
	      case K_BY_NAME:
		/* Work through list and match names to numbers. */
		for (i = 0; i < 255; ++i) chartable[i] = 0;
		desc = cdr(desc);
		/* Support optional explicit string a la by-char. */
		if (stringp(car(desc))) {
		    str = c_string(car(desc));
		    for (i = 0; i < strlen(str); ++i) chartable[str[i]] = i;
		} else {
		    str = NULL;
		}
		i = 0;
		for (nml = cdr(desc); nml != lispnil; nml = cdr(nml)) {
		    val = eval(car(nml));
		    SYNTAX(val, (numberp(val)
			    || ttypep(val)
			    || mtypep(val)
			    || utypep(val)),
			   "by-name layer key is not a number");
		    chartable[(str ? str[i] : 'a' + i)] = c_number(val);
		    ++i;
		}
		/* Now read it all in. */
		read_rle(cdr(rest), setter, chartable);
		return;
	      default:
		sprintlisp(buf, desc);
		init_warning("Ignoring garbage terrain description %s", buf);
	    }
	}
    }
}

/* General RLE reader.  This basically parses the run lengths and calls
   the function that records what was read. */

read_rle(contents, setter, chartable)
Obj *contents;
int (*setter)();
short *chartable;
{
    char ch, *rowstr;
    int i, x, y, run, val;
    int basex = layerareax, basey = layerareay + layerareaheight - 1;
    Obj *rest, *tmp;

    rest = contents;
    y = basey;
    while (rest != lispnil && y >= 0) {
	/* should error check ... */
	rowstr = c_string(car(rest));
	i = 0;
	x = basex;  /* depends on shape of saved data... */
	while ((ch = rowstr[i++]) != '\0' && x < layerareawidth) {
	    if (isdigit(ch)) {
		run = ch - '0';
		/* will foul up if end of string here! */
		/* should check... */
		while ((ch = rowstr[i++]) != 0 && isdigit(ch)) {
		    run = run * 10 + ch - '0';
		}
		/* if we've got garbled data, skip to the next line */
		if (ch == '\0') goto recovery;
	    } else {
		run = 1;
	    }
	    if (ch == '*' && !ignorespecials) {
		/* A run/large number separator. */
	    } else if (ch == ',' && !ignorespecials) {
		/* This was a value instead of a run length. */
		val = run;
		run = 1;
	    } else if (chartable != NULL) {
		val = chartable[ch];
	    } else if (between('a', ch, '~')) {
		val = ch - 'a';
	    } else if (between(':', ch, '[')) {
		val = ch - ':' + 30;
	    } else {
		/* (should only warn once for each char) */
		init_warning("Garbage char '%c' in a layer", ch);
		val = 0;
	    }
	    val = val * layermultiplier + layeradder;
	    while (run-- > 0 && r_in_area(x, y)) { /* i.e. is in array */
		(*setter)(x, y, val);
		++x;
	    }
	}
      recovery:
	/* Fill-in string may be too short, should subst default for rest */
	rest = cdr(rest);
	y--;
    }
}

#ifdef DEBUGGING
/* Use this to ensure that everything is cool. */

doublecheck_state(side)
{
    Unit *unit;
    Side *loop_side;

    for_all_units(loop_side, unit) {
	if (unit->x < 0 || unit->x >= area.width ||
	    unit->y <= 0 || unit->y >= (area.height - 1) ||
	    unit->hp <= 0) {
	    Dprintf("%s off map hp %d", unit_desig(unit), unit->hp);
	}
    }
}
#endif /* DEBUGGING */

/* Return the value of an extension, or the default if it wasn't found. */

long
extension_value(extensions, key, dflt)
Obj *extensions;
char *key;
long dflt;
{
    Obj *bdg, *val;

    for ( ; extensions != lispnil; extensions = cdr(extensions)) {
	bdg = car(extensions);
	if (consp(bdg)
	    && symbolp(car(bdg))
	    && strcmp(c_string(car(bdg)), key) == 0) {
	    val = cadr(bdg);
	    /* Return number or string directly, Lisp value otherwise. */
	    if (numberp(val)) {
		return c_number(val);
	    } else if (stringp(val)) {
		return ((long) c_string(val));
	    } else {
		return ((long) val);
	    }
	}
    }
    return dflt;
}

set_u_internal_name(u, s) int u; char *s; { utypes[u].iname = s; }
set_u_type_name(u, s) int u; char *s; { utypes[u].name = s; }
set_m_type_name(m, s) int m; char *s; { mtypes[m].name = s; }
set_t_type_name(t, s) int t; char *s; { ttypes[t].name = s; }
