/* Copyright (c) 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Unit task handling. */

#include "conq.h"

#define ut_dies_on(u, t) (ut_vanishes_on(u,t) || ut_wrecks_on(u, t))

#define push_task(unit, task)  \
    task->next = unit->plan->tasks;  \
    unit->plan->tasks = task;

/* A list of available task objects. */

Task *freetasks = NULL;

char *tasktypenames[] = {

#undef  DEF_TASK
#define DEF_TASK(code,NAME,argtypes, fn) NAME,

#include "task.def"

    NULL
};

do_none_task(unit, task)
Unit *unit;
Task *task;
{
}

do_approach_task(unit, task, uid2, dist)
Unit *unit;
Task *task;
int uid2, dist;
{
}

do_build_task(unit, task)
Unit *unit;
Task *task;
{
}

do_capture_unit_task(unit, task)
Unit *unit;
Task *task;
{
}

do_action_task(unit, task)
Unit *unit;
Task *task;
{
}

do_hit_position_task(unit, task)
Unit *unit;
Task *task;
{
}

do_hit_unit_task(unit, task)
Unit *unit;
Task *task;
{
}

do_movedir_task(unit, task)
Unit *unit;
Task *task;
{
}

do_moveto_task(unit, task)
Unit *unit;
Task *task;
{
}

do_occupy_task(unit, task)
Unit *unit;
Task *task;
{
}

do_pickup_task(unit, task)
Unit *unit;
Task *task;
{
}

do_repair_task(unit, task)
Unit *unit;
Task *task;
{
}

do_resupply_task(unit, task)
Unit *unit;
Task *task;
{
}

do_sentry_task(unit, task)
Unit *unit;
Task *task;
{
}

#if 0
(int (*taskfunctions)())[] = {

#undef  DEF_TASK
#define DEF_TASK(code,NAME,argtypes, FN) FN,

#include "task.def"

    NULL
};
#endif

char *task_desig();

void
init_tasks()
{
#if 0
    int i;
    Task *tasks;

    tasks = (Task *) xmalloc(INITMAXUNITS * sizeof(Task));
    freetasks = tasks;
    for (i = 0; i < INITMAXUNITS; ++i) {
      tasks[i].next = &tasks[i+1];
    }
    tasks[INITMAXUNITS-1].next = NULL;
#endif
}

/* Create and return a new task. */

Task *
create_task(type)
TaskType type;
{
	int i;
    Task *task;

    if (freetasks) {
	task = freetasks;
	freetasks = task->next;
    } else {
	task = (Task *) xmalloc(sizeof(Task));
    }
    task->type = type;
    task->execnum = 0;
    task->retrynum = 0;
 /*   for (i = 0; i < 4; ++i) task->args[i] = 0; */
    task->next = NULL;
    return task;
}

/* Perform a single given task. */

TaskOutcome
execute_task(unit, task)
Unit *unit;
Task *task;
{
    int u = unit->type, dist, dx, dy, dir, tx, ty, nx, ny;
    Unit *unit2, *occ;
    Side *us = unit->side;

    if (!alive(unit) || task == NULL) return;
    DMprintf("%s executing task %s\n", unit_desig(unit), task_desig(task));
    /* Count this execution. */
    ++task->execnum;
    switch (task->type) {
      case NO_TASK:
	/* This is a no-op, useful as a placeholder.  Always "succeeds". */
	return TASK_COMPLETE;
      case BUILD_TASK:
	/* Construction results in several kinds of actions as needed. */
	if (0 /* need research */) {
	} else if (0 /* need to tool up */) {
	} else {
	    /* Either look for or work on any incomplete units. */
	    unit2 = NULL;
	    if (task->args[1] != 0) {
		unit2 = find_unit(task->args[1]);
		if (in_play(unit2) && unit2->type == task->args[0]) {
		    if (fullsized(unit2)) {
			++(task->args[2]);
			if (task->args[2] >= task->args[3]) {
			    return TASK_COMPLETE;
			}
		    	unit2 = NULL;
	            } else {
	            	/* press on */
	            }
		} else {
		    unit2 = NULL;
		}
	    }
	    if (unit2 == NULL) {
	    	/* Search for any appropriate incomplete units nearby. */
	    	for_all_occupants(unit, occ) {
	    	    if (in_play(occ)
	    	    	&& !fullsized(occ)
	    	    	&& occ->type == task->args[0]) {
	    	    	unit2 = occ;
	    	    	break;
	    	    }
	    	}
	    }
	    if (unit2 == NULL) {
		if (!valid(check_create_in_action(unit, task->args[0], unit))) {
		    ++task->retrynum;
		    return TASK_FAILED;
		} else if (prep_create_in_action(unit, task->args[0], unit)) {
		    return TASK_PREPPED_ACTION;
		} else {
		    ++task->retrynum;
		    return TASK_FAILED;
		}
	    } else {
		task->args[1] = unit2->id;
		if (in_play(unit2)) {
		    if (fullsized(unit2)) { /* does this ever execute? */
			++(task->args[2]);
			if (task->args[2] >= task->args[3]) {
			    return TASK_COMPLETE;
			} else {
			    return TASK_INCOMPLETE;
			}
		    } else if (!valid(check_build_action(unit, unit2))) {
			++task->retrynum;
			return TASK_FAILED;
		    } else if (prep_build_action(unit, unit2)) {
			return TASK_PREPPED_ACTION;
		    } else {
			++task->retrynum;
			return TASK_FAILED;
		    }
		} else {
		    ++task->retrynum;
		    return TASK_FAILED;
		}
	    }
	}
	break;
      case CAPTURE_UNIT_TASK:
	/* This is to capture a given type/side of unit at a given place. */
	/* (need to be able to say how hard to try) */
	tx = task->args[0];  ty = task->args[1];
	dist = distance(tx, ty, unit->x, unit->y);
	switch (dist) {
	  case 0:
	    /* huh? */
	    return TASK_FAILED;
	  case 1:
	    if ((unit2 = unit_at(tx, ty)) != NULL) {
		if (unit2->side != us) {
		    if (!valid(check_capture_action(unit, unit2))) {
			/* Not possible to capture for some reason. */
			++task->retrynum;
			return TASK_FAILED;
		    } else if (do_capture_action(unit, unit2)) {
			/* We did it! */
			return TASK_COMPLETE;
		    } else {
			/* Try again in a moment. */
			return TASK_INCOMPLETE;
		    }
		} else {
		    /* Our victim apparently saw the light already. */
		    return TASK_COMPLETE;
		}
	    } else {
		/* Odd, nothing was here to capture. */
		++task->retrynum;
		return TASK_FAILED;
	    }
	  default:
	    /* We're too far away to capture directly, add a moveto task. */
	    push_movenear_task(unit, tx, ty, 1);
	    return TASK_INCOMPLETE;
	} 
	break;
      case HIT_POSITION_TASK:
	/* This is to hit a given place. */
	/* (ask for a number of hits?) */
	/* (should do a ranged hit if possible) */
	tx = task->args[0];  ty = task->args[1];
	dist = distance(tx, ty, unit->x, unit->y);
	if (!valid(check_fire_into_action(unit, tx, ty, 0, 0))) {
	    /* Not possible to shoot for some reason. */
	    ++task->retrynum;
	    return TASK_FAILED;
	} else if (prep_fire_into_action(unit, tx, ty, 0, 0)) {
	    return TASK_PREPPED_ACTION;
	} else {
	    /* We're too far away to shoot directly, add a moveto task. */
	    push_moveto_task(unit, tx, ty);
	    return TASK_INCOMPLETE;
	}
	break;
      case HIT_UNIT_TASK:
	/* This is to hit a given type/side of unit at a given place. */
	/* (ask for a number of hits?) */
	/* (should do a ranged hit if possible) */
	tx = task->args[0];  ty = task->args[1];
	dist = distance(tx, ty, unit->x, unit->y);
	switch (dist) {
	  case 0:
	    /* hit our own cell? */
	    return TASK_FAILED;
	  case 1:
	    if ((unit2 = unit_at(tx, ty)) != NULL) {
		if (unit2->side != us) {
		    if (!valid(check_attack_action(unit, unit2, 100))) {
			/* Not possible to attack for some reason. */
			++task->retrynum;
			return TASK_FAILED;
		    } else if (prep_attack_action(unit, unit2, 100)) {
			return TASK_PREPPED_ACTION;
		    } else {
			/* Try again in a moment. */
			return TASK_INCOMPLETE;
		    }
		} else {
		    /* Our victim apparently saw the light already. */
		    return TASK_COMPLETE;
		}
	    } else {
		/* Nothing was here to attack. */
		++task->retrynum;
		return TASK_FAILED;
	    }
	  default:
	    if (mobile(u)
	    	&& !valid(check_fire_at_action(unit, unit_at(tx, ty), 0))) {
		    /* We're too far away to hit directly, add a moveto task. */
		    push_movenear_task(unit, tx, ty, 1);
		    return TASK_INCOMPLETE;
	    } else {
	    	if ((unit2 = unit_at(tx, ty)) != NULL) {
		  if (unit2->side != us) {
		    if (!valid(check_fire_at_action(unit, unit2, 0))) {
			/* Not possible to attack for some reason. */
			++task->retrynum;
			return TASK_FAILED;
		    } else if (!could_hit(u, unit2->type)) {
		    	/* firing actions don't check stupidity of action. */
			++task->retrynum;
			return TASK_FAILED;		    	
		    } else if (prep_fire_at_action(unit, unit2, 0)) {
			return TASK_PREPPED_ACTION;
		    } else {
			/* Try again in a moment. */
			return TASK_INCOMPLETE;
		    }
		  } else {
		    /* Our victim apparently saw the light already. */
		    return TASK_COMPLETE;
		  }
	        } else {
		   /* Nothing was here to attack. */
		   ++task->retrynum;
		   return TASK_FAILED;
	        }
	    }
	}
	break;
      case MOVEDIR_TASK:
	if ((task->args[1])-- > 0) {
	    dir = task->args[0];
	    tx = wrapx(unit->x + dirx[dir]);  ty = unit->y + diry[dir];
	    if (unit_at(tx, ty)) {
	    	for_all_stack(tx, ty, unit2) {
	    	}
	    } else {
	    	prep_move_action(unit, tx, ty, 0, unit);
	    	return TASK_PREPPED_ACTION;
	    }
	} else {
	    return TASK_COMPLETE;
	}
	break;
      case MOVETO_TASK:
	/* This task is to get to a designated location somehow. */
	tx = task->args[0];  ty = task->args[1];
	dist = distance(tx, ty, unit->x, unit->y);
	if (dist <= task->args[2]) {
	    return TASK_COMPLETE;
	}
	switch (dist) {
	  case 0:
	    /* We're there already, nothing more to do. */
	    return TASK_COMPLETE;
	  case 1:
	    /* Adjacent hex, do a single move. */
	    if (unit_at(tx, ty)) {
		/* Interact with the unit at our destination. */
		if (move_onto_unit(unit, tx, ty)) {
		    return TASK_COMPLETE;
		} else {
		    ++task->retrynum;
		    return TASK_FAILED;
		}
	    } else {
		/* Moving into an empty hex. */
		prep_move_action(unit, tx, ty, unit->z, unit);
		return TASK_PREPPED_ACTION;
	    }
	    break;
	  default:
	    /* Still some distance away, pick a way to go. */
	    if ((dir = choose_move_dir(unit, tx, ty, TRUE)) >= 0) {
		nx = wrapx(unit->x + dirx[dir]);  ny = unit->y + diry[dir];
	    	if (unit_at(nx, ny)) {
	    	    for_all_stack(nx, ny, unit2) {
	    	    	if (can_occupy(unit, unit2)) {
	    	    	    if (prep_enter_action(unit, unit2, unit)) {
		    		return TASK_PREPPED_ACTION;
	    	    	    } else {
		    		++task->retrynum;
		   		return TASK_FAILED;
	    	    	    }
	    	    	} else if (unit2->side != unit->side) {
	    	    	    if (prep_attack_action(unit, unit2, 100)) {
		    		return TASK_PREPPED_ACTION;
	    	    	    } else {
		    		++task->retrynum;
		    		return TASK_FAILED;
	    	    	    }
	    	    	}
	    	    }
		    if (prep_move_action(unit, nx, ny, 0, unit)) {
		    	return TASK_PREPPED_ACTION;
		    } else {
		    	/* Fail if none of the prospective dirs is working. */
		    	++task->retrynum;
		    	return TASK_FAILED;
		    }
		} else if (prep_move_action(unit, nx, ny, 0, unit)) {
		    return TASK_PREPPED_ACTION;
		} else {
		    /* Fail if none of the prospective dirs is working. */
		    ++task->retrynum;
		    return TASK_FAILED;
		}
	    } else {
		++task->retrynum;
		return TASK_FAILED;
	    }
	}
	break;
      case SENTRY_TASK:
	if (task->args[0] > 0) {
	    --(task->args[0]);
	    return TASK_INCOMPLETE;
	} else {
	    /* Unit won't necessarily wake up, may just replan and
	       continue sleeping. */
	    return TASK_COMPLETE;
	}
	break;
      default:
	run_warning("Unknown task type %d", task->type);
/*        case_panic("task type", task->type);  */
	break;
    }
    /* post-task analysis - should we wake up or do something? */
    /* have any goals been satisfied? */
    /* if action buffered and done later, must check elsewhere */
    DMprintf("%s couldn't find an action to acomplish task\n",
	    unit_desig(unit));
    /* is this right? */
    ++task->retrynum;
    return TASK_FAILED;
}

/* Implement interaction with unit already in hex. */

/* some of these tests should be attached to assault action setup only */

move_onto_unit(unit, x, y)
Unit *unit;
int x, y;
{
    int success = FALSE;
    int u = unit->type;
    Unit *unit2 = unit_at(x, y);

    if (unit2 == NULL) {
	/* ? */
    } else if (allied_side(unit->side, unit2->side)) {
	/* fail for now */
    } else if (unit->transport != NULL
	       && impassable(unit, unit2->x, unit2->y)
	       && !uu_bridge(unit2->type, unit->type)) {
/*	move_error(unit, "%s can't attack there!",
		   unit_handle(unit->side, unit));  */
    } else if (0 /* side_view(unit->side, unit2->x, unit2->y) == EMPTY */) {
/*	notify(unit->side, "%s spots something!",
	       unit_handle(unit->side, unit)); */
	see_exact(unit->side, unit2->x, unit2->y);
	update_cell_display(unit->side, unit2->x, unit2->y, TRUE);
    } else if (0 /* avoid attacking always */) {
    } else if (!could_hit(u, unit2->type)
	       && !could_capture(u, unit2->type)) {
/*	move_error(unit->side, "%s refuses to attack a %s!",
		   unit_handle(unit->side, unit), u_type_name(unit2->type)); */
    } else if (!enough_ammo(unit, unit2)) {
/*	move_error(unit->side, "%s is out of ammo!",
		   unit_handle(unit->side, unit));  */
    } else {
	/* attack maybe? */
	success = TRUE;
    }
    return success;
}

/* This weird-looking routine computes next directions for moving to a
   given spot.  The basic strategy is to prefer to go in the x or y distance
   that is the greatest difference, so as to even the two displacements out.
   (This leaves more options open if blockage several hexes away.)  Make it
   probabilistic, so repeated travel will eventually trace the envelope of
   possible moves.  The number of directions ranges from 1 to 4, depending
   on whether there is a straight-line path to the dest, and whether we are
   required to take a direct path or are allowed to move in dirs that don't
   the unit any closer (we never increase our distance though).
   Some trickinesses:  if area wraps, must resolve ambiguity about
   getting to the same place going either direction (we pick shortest). */

choose_move_dir(unit, tx, ty, shortest)
Unit *unit;
int tx, ty, shortest;
{
    int avoid = 0;
    int closer;
    int dx, dxa, dy, dist, d1, d2, d3, d4, axis = -1, hextant = -1, tmp;

    dist = distance(unit->x, unit->y, tx, ty);
    dx = tx - unit->x;  dy = ty - unit->y;

    if (area.xwrap) {
	dxa = (tx + area.width) - unit->x;
	if (abs(dx) > abs(dxa)) dx = dxa;
	dxa = (tx - area.width) - unit->x;
	if (abs(dx) > abs(dxa)) dx = dxa;
    }
    if (dx == 0 && dy == 0) {
	return -1;
    }
    axis = hextant = -1;
    if (dx == 0) {
	axis = (dy > 0 ? NE : SW);
    } else if (dy == 0) {
	axis = (dx > 0 ? EAST : WEST);
    } else if (dx == (0 - dy)) {
	axis = (dy > 0 ? NW : SE);
    } else if (dx > 0) {
	hextant = (dy > 0 ? EAST : (abs(dx) > abs(dy) ? SE : SW));
    } else {
	hextant = (dy < 0 ? WEST : (abs(dx) > abs(dy) ? NW : NE));
    }
    if (axis >= 0) {
	d1 = d2 = axis;
    }
    if (hextant >= 0) {
	d1 = left_dir(hextant);
	d2 = hextant;
    }
    d3 = left_dir(d1);
    d4 = right_dir(d2);
    if (flip_coin()) {
        tmp = d1;  d1 = d2;  d2 = tmp;
    }
    if (plausible_move_dir(unit, d1)) {
	return d1;
    } else if (d1 != d2 && plausible_move_dir(unit, d2)) {
	return d2;
    } else if (!(shortest || dist == 1)) {
	if (flip_coin()) {
	    tmp = d3;  d3 = d4;  d4 = tmp;
	}
	if (plausible_move_dir(unit, d3)) {
	    return d3;
	} else if (plausible_move_dir(unit, d4)) {
	    return d4;
	}
    }
    if (possible_move_dir(unit, d1)) {
	return d1;
    } else if (d1 != d2 && possible_move_dir(unit, d2)) {
	return d2;
    } else if (!(shortest || dist == 1)) {
	if (possible_move_dir(unit, d3)) {
	    return d3;
	} else if (possible_move_dir(unit, d4)) {
	    return d4;
	}
    }
    /* inform caller that we can't find a way to go. */
    return -1;
}

/* A heuristic test for whether the given direction is a good one to move in.
   May not be accurate, but only effect is to increase chance of failed
   move actions. */

plausible_move_dir(unit, dir)
Unit *unit;
int dir;
{
    int u = unit->type;
    int ux = unit->x, uy = unit->y;
    int nx = wrapx(unit->x + dirx[dir]), ny = unit->y + diry[dir];

    if (0) {
	/* occupied by a unit that won't share hex */
    } else if (ut_dies_on(u, terrain_at(nx, ny))) {
	/* ... and no usable connection from our hex ... */
	return FALSE;
    } else if (unit->act->mp < ut_mp_to_enter(u, terrain_at(nx, ny))) {
	return FALSE;
    } else {
	return TRUE;
    }
}

possible_move_dir(unit, dir)
Unit *unit;
int dir;
{
    int u = unit->type;
    int ux = unit->x, uy = unit->y;
    int nx = wrapx(unit->x + dirx[dir]), ny = unit->y + diry[dir];

    if (0) {
	/* occupied by a unit that won't share hex */
    } else if (ut_dies_on(u, terrain_at(nx, ny))) {
	/* ... and no usable connection from our hex ... */
	return FALSE;
    } else if (u_speed(u) < ut_mp_to_enter(u, terrain_at(nx, ny))) {
	return FALSE;
    } else {
	return TRUE;
    }
}

void 
free_task(task)
Task *task;
{
    Task *next;

    while (task != NULL) {
	next = task->next;
	task->next = freetasks;
	freetasks = task;
	task = next;
    }
}

be_active(unit)
Unit *unit;
{
    /* Shouldn't be asleep any longer. */
    unit->plan->asleep = FALSE;
    /* We're not in reserve. */
    unit->plan->reserve = FALSE;
    /* Obviously we're no longer waiting to be told what to do. */
    unit->plan->waitingfortasks = FALSE;
}

/* Order a unit to sit quietly for some number of turns. */

order_sentry(unit, n)
Unit *unit;
int n;
{
    unit->plan->tasks = create_task(SENTRY_TASK);
    unit->plan->tasks->args[0] = n;
    /* We're no longer waiting to be told what to do. */
    unit->plan->waitingfortasks = FALSE;
}

Task *
create_moveto_task(unit, x, y)
Unit *unit;
int x, y;
{
    Task *task = create_task(MOVETO_TASK);

    task->args[0] = x;  task->args[1] = y;
    task->args[2] = 0;
    return task;
}

push_moveto_task(unit, x, y)
Unit *unit;
int x, y;
{
    Task *task;

    if (unit == NULL || unit->plan == NULL) {
	run_warning("Trying to move garbage %s", unit_desig(unit));
	return;
    }
    task = create_moveto_task(unit, x, y);
    push_task(unit, task);
    /* Obviously we're no longer waiting to be told what to do. */
    unit->plan->waitingfortasks = FALSE;
}

/* Give the unit a task to move to a given place, erasing every other task. */

order_moveto(unit, x, y)
Unit *unit;
int x, y;
{
    if (unit == NULL || unit->plan == NULL) {
	run_warning("Trying to move garbage %s", unit_desig(unit));
	return;
    }
    unit->plan->tasks = create_moveto_task(unit, x, y);
    be_active(unit);
}

Task *
create_movenear_task(unit, x, y, dist)
Unit *unit;
int x, y, dist;
{
    Task *task = create_task(MOVETO_TASK);

    task->args[0] = x;  task->args[1] = y;
    task->args[2] = dist;
    return task;
}

push_movenear_task(unit, x, y, dist)
Unit *unit;
int x, y, dist;
{
    Task *task;

    if (unit == NULL || unit->plan == NULL) {
	run_warning("Trying to move garbage %s", unit_desig(unit));
	return;
    }
    task = create_movenear_task(unit, x, y, dist);
    push_task(unit, task);
    be_active(unit);
}

/* Give the unit a task to move to a given place, erasing every other task. */

order_movenear(unit, x, y)
Unit *unit;
int x, y;
{
    if (unit == NULL || unit->plan == NULL) {
	run_warning("Trying to move garbage %s", unit_desig(unit));
	return;
    }
    unit->plan->tasks = create_moveto_task(unit, x, y, 1);
    be_active(unit);
}

/* Create a task to move in a given direction for a given distance. */

Task *
create_movedir_task(unit, dir, n)
Unit *unit;
int dir, n;
{
    Task *task = create_task(MOVEDIR_TASK);

    task->args[0] = dir;  task->args[1] = n;
    return task;
}

/* Fill in the given unit with direction-moving orders. */

order_movedir(unit, dir, n)
Unit *unit;
int dir, n;
{
    unit->plan->tasks = create_movedir_task(unit, dir, n);
    unit->plan->tasks->args[0] = dir;  unit->plan->tasks->args[1] = n;
    be_active(unit);
}

/* Order to follow a unit just needs the unit to follow. */

order_follow(unit, leader, n)
Unit *unit, *leader;
int n;
{
#if 0
    unit->plan->orders.type = FOLLOW;
    unit->plan->orders.rept = n;
    unit->plan->orders.p.leader_id = leader->id;;
#endif
}

/* A two-waypoint patrol suffices for many purposes. */
/* Should have a more general patrol routine eventually (> 2 waypoints). */

order_patrol(unit, x0, y0, x1, y1, n)
Unit *unit;
int x0, y0, x1, y1, n;
{
#if 0
    unit->plan->orders.type = PATROL;
    unit->plan->orders.rept = n;
    unit->plan->orders.p.pt[0].x = x0;
    unit->plan->orders.p.pt[0].y = y0;
    unit->plan->orders.p.pt[1].x = x1;
    unit->plan->orders.p.pt[1].y = y1;
#endif
}

/* Wait around to embark on some unit when it comes to our location. */

order_embark(unit, n)
Unit *unit;
int n;
{
}

/* This routine sets up a task to build a unit of the given type. */

Task *
create_build_task(unit, u2, run)
Unit *unit;
int u2, run;
{
    Task *task = create_task(BUILD_TASK);

    /* (should check plausibility first) */
    task->args[0] = u2;
    task->args[1] = 0;
    task->args[2] = 0;
    task->args[3] = run;
    return task;
}

push_build_task(unit, u2, run)
Unit *unit;
int u2, run;
{
    Task *task;

    if (unit == NULL || unit->plan == NULL) {
	run_warning("Trying to build with garbage %s", unit_desig(unit));
	return;
    }
    task = create_build_task(unit, u2, run);
    push_task(unit, task);
    be_active(unit);
}

/* Describe a task succinctly. */

char *taskbuf = NULL;

char *
task_desig(task)
Task *task;
{
    if (taskbuf == NULL) taskbuf = (char *) xmalloc(1000 * sizeof(char));
    if (task) {
	sprintf(taskbuf, "{%s %d,%d,%d,%d *%d",
		tasktypenames[task->type],
		task->args[0], task->args[1],
		task->args[2], task->args[3],
		task->execnum);
	if (task->retrynum > 0) {
	    sprintf(taskbuf+strlen(taskbuf), " %d fail", task->retrynum);
	}
	strcat(taskbuf, "}");
    } else {
	sprintf(taskbuf, "no task");
    }
    return taskbuf;
}
