/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact.  */

/* Graphical routines for the Xlib interface to xconq. */

#include "conq.h"
#include "xconq.h"

#define terrain_visible(side, x, y) \
  (g_see_all() || side->designer || terrain_view(side, x, y) != UNSEEN)

#define units_visible(side, x, y) \
  (g_see_all() || side->designer || cover(side, wrapx(x), y) >= 100)

/* Draw an individual detailed hex, as a row of one, on all maps. */

update_cell_display(side, x, y, rightnow)
Side *side;
int x, y;
int rightnow;
{
    Map *map;

    if (active_display(side)) {
	for_all_maps(side, map) {
	    draw_row(side, map, x, y, 1, TRUE);
	    draw_cursor(side, map);
	}
	if (rightnow) flush_output(side);
    }
}

/* Completely redo a screen, making no assumptions about appearance.
   This is a last-gasp measure, most redrawing should be restricted
   to only the directly affected windows.  Also this shouldn't be
   done without the user's permission, since it will blow away impending
   input. */

redraw(side)
Side *side;
{
    /* should do prompts and help window too */
    show_all_maps(side);
    flush_output(side);
    flush_input(side);
}

/* Put the point x, y in the center of the map. */

void
recenter(side, map, x, y)
Side *side;
Map *map;
int x, y;
{
    if ((x == map->vcx) && (y == map->vcy)) {
	/* no movement needed */
	/* can we assume the state of the display is valid? */
	/* could use oldx,y to fix... */
    } else {
	map->vcx = x;  map->vcy = y;
	set_map_viewport(map);
	show_map(side, map);
    }
}

/* Ensure that given location is visible on the front map.
   We also flush the input because
   any input relating to a different screen is probably worthless. */

put_on_screen(side, x, y)
Side *side;
int x, y;
{

    /* Ugly hack to prevent extra boxes being drawn during init - don't ask!*/
    if (x == 0 && y == 0) return;
    if (!in_middle(side, side->ui->maps, x, y)) {
	recenter(side, side->ui->maps, x, y);
    }
}

/* Decide whether given location is not too close to edge of screen. */
/* We do this because it's a pain to move units when half the adjacent */
/* places aren't even visible.  This routine effectively places a lower */
/* limit of 5x5 for the map window. (I think) */

in_middle(side, map, x, y)
Side *side;
Map *map;
int x, y;
{
    int vcx = map->vcx, vw2 = map->vw2, odd = map->vw % 2;
    int vcy = map->vcy, vh2 = map->vh2;
    int left_hex, right_hex;
    
    /* changed so that units on top and bottom lines do the right thing. */
    /* this still isn't quite right! */
    if (!between(vcy-vh2+2, y, vcy+vh2-2) && between(1, y, area.height-2))
	return FALSE;
    left_hex = vcx - vw2 - ((y - vcy) / 2) - (y > vcy) + 3;
    right_hex = vcx + vw2 + odd - ((y - vcy) / 2) + ((y < vcy) && odd)
      - ((y > vcy) && !odd) - 3;

    if (left_hex < 0)
      return !between(right_hex, x, left_hex + area.width);
    else if (right_hex >= area.width)
      return !between(right_hex - area.width, x, left_hex);
    return between(left_hex, x, right_hex);
}

/* Decide if the position is visible in the front map. */
/* (should be "on some visible map"?) */

in_view(side, x, y)
Side *side;
int x, y;
{
    Map *map = side->ui->maps;

    int vx = map->vx, vy = map->vy;
    int vw = map->vw, vh = map->vh;
    int vcx = map->vcx, vw2 = map->vw2, odd = map->vw % 2;
    int vcy = map->vcy, vh2 = map->vh2;
    int left_hex, right_hex;
    
    /* changed so that units on top and bottom lines do the right thing. */
    /* this still isn't quite right! */
    if (!between(vy, y, vy+vh) && between(0, y, area.height-1))
	return FALSE;
    left_hex = vx - ((y - vy) / 2) - 1;
    right_hex = vx + vw - ((y - vy) / 2) + 1;

    if (left_hex < 0)
      return !between(right_hex, x, left_hex + area.width);
    else if (right_hex >= area.width)
      return !between(right_hex - area.width, x, left_hex);
    return between(left_hex, x, right_hex);
}

/* Transform map coordinates into screen coordinates, relative to the given
   side.  Allow for cylindricalness and number of pixels in a hex. */

xform(side, map, x, y, sxp, syp)
Side *side;
Map *map;
int x, y, *sxp, *syp;
{
    if (in_area(x, y)) {
	/* Compute the pixel positions. */
	*sxp = ((map->hw * (x - map->vx)) + (map->hw * (y - map->vy)) / 2);
	*syp = map->hch * ((map->vy + map->vh) - y - 1);
    } else {
/*	fprintf(stderr, "%d,%d shouldn't be xformed!\n", x, y);
	abort();  */
	*sxp = *syp = 1;
    }
}

xform_unit(side, map, unit, sxp, syp, swp, shp)
Side *side;
Map *map;
Unit *unit;
int *sxp, *syp, *swp, *shp;
{
    int num, n, sq;
    int x = unit->x, y = unit->y;
    int sx1, sy1, sw1, sh1;
    Unit *unit2;

    if (unit->transport == NULL) {
	xform(side, map, x, y, sxp, syp);
	num = num_units_at(x, y);
	if (num == 0) { *swp = *shp = *sxp = *syp = 1; return; }
	n = 0;
	for_all_stack(x, y, unit2) {
	    if (unit2 == unit) break;
	    ++n;
	}
	sq = isqrt(num);
	if (sq * sq < num) ++sq;
	*swp = map->uw / sq;  *shp = map->uh / sq;
	*sxp += (map->hw - map->uw) / 2 + *swp * (n % sq);
	*syp += (map->hh - map->uh) / 2 + *shp * (n / sq);
    } else {
	xform_unit(map, unit->transport, &sx1, &sy1, &sw1, &sh1);
	xform_occupant(map, unit->transport, unit, sx1, sy1, sw1, sh1, sxp, syp, swp, shp);
    }
}

xform_unit_self(side, map, unit, sxp, syp, swp, shp)
Side *side;
Map *map;
Unit *unit;
int *sxp, *syp, *swp, *shp;
{
    int num, n, sq;
    int x = unit->x, y = unit->y;
    int sx1, sy1, sw1, sh1;
    Unit *unit2;
    
    if (unit->transport == NULL) {
	if (unit->occupant == NULL) {
	    xform_unit(map, unit, sxp, syp, swp, shp);
	} else {
	    xform_unit(map, unit, &sx1, &sy1, &sw1, &sh1);
	    xform_occupant(map, unit, unit, sx1, sy1, sw1, sh1, sxp, syp, swp, shp);
	}
    } else {
	xform_unit(map, unit->transport, &sx1, &sy1, &sw1, &sh1);
	xform_occupant(map, unit->transport, unit, sx1, sy1, sw1, sh1, sxp, syp, swp, shp);
    }
}

xform_occupant(side, map, transport, unit, sx, sy, sw, sh, sxp, syp, swp, shp)
Side *side;
Map *map;
Unit *transport, *unit;
int sx, sy, sw, sh, *sxp, *syp, *swp, *shp;
{
    int num, n, sq;
    Unit *unit2;

    num = num_occupants(transport);
    if (num == 0) { *swp = *shp = *sxp = *syp = 1; return; }
    ++num;
    if (transport == unit) {
	n = 0;
    } else {
	/* Count the transport as being in the box with its occupants. */
	/* Figure out the position of this unit among all the occupants. */
	n = 1;
	for_all_occupants(transport, unit2) {
	    if (unit2 == unit) break;
	    ++n;
	}
    }
    sq = isqrt(num);
    if (sq * sq < num) ++sq;
    *swp = sw / sq;  *shp = sh / sq;
    *sxp = sx + *swp * (n % sq);  *syp = sy + *shp * (n / sq);
}

/* Un-transform screen coordinates (as supplied by mouse perhaps) into
   map coordinates.  Compute both a hex and a closest direction, just
   in case the caller needs to know.   This routine actually doesn't
   pay attention to hexagon boundaries, using only box-shaped areas,
   but in practice this seems to be OK. */

deform(side, map, sx, sy, xp, yp, dirp)
Side *side;
Map *map;
int sx, sy, *xp, *yp, *dirp;
{
    int adjust;

    /* Scale to hex coords then flip y. */
    *yp = (map->vy + map->vh) - (sy / map->hch) - 1;
    /* Even/odd row adjustment gizmo (value in pixels). */
    adjust = (((*yp - map->vy) & 1) ? (map->hw / 2) : 0);
    /* Scale to hex then offset and correct for hex obliqueness. */
    *xp = map->vx + ((sx - adjust) / map->hw) - (*yp - map->vy) / 2;
    /* Wrap coords as usual. */
    if (area.xwrap) *xp = wrapx(*xp);
    /* Compute the closest edge, as a direction.  Note that this doesn't
       return all dirs (since would return adj hex pos with opposite dir -
       valid because also refers to the same edge). */
    if (((sx - adjust) % map->hw) > map->hw/2) {
	*dirp = ((sy % map->hch < (map->hh - map->hch)) ? NE : EAST);
    } else {
	*dirp = ((sy % map->hch < (map->hh - map->hch)) ? NW : WEST);
    }
    Dprintf("Pixel %d,%d (relative %d,%d) -> hex %d,%d dir %d\n",
	    sx, sy, (sx - adjust) % map->hw, sy % map->hch, *xp, *yp, *dirp); 
}

/* Draw all the maps that are currently up. */

show_all_maps(side)
Side *side;
{
    Map *map;

    for_all_maps(side, map) {
	show_map(side, map);
    }
}

/* Display a map and all of its paraphernalia. */

show_map(side, map)
Side *side;
Map *map;
{
    XClearWindow(side->ui->dpy, map->infowin);
    XClearWindow(side->ui->dpy, map->viewwin);
    /* draw an outline of unseen part of world? */
    show_controls(side, map);
    show_map_info(side, map);
    show_map_view(side, map);
    draw_cursor(side, map);
    flush_output(side);
    draw_scrollbars(side, map);
    if (side_lost(side)) {
	draw_map_scratchout(side, map);
    }
}

/* Draw the view proper. */

show_map_view(side, map)
Side *side;
Map *map;
{
    int y1, y2, y, x1, x2, adj;

    if (map->vcx < 0 || map->vcy < 0) {
	map->vcx = map->vcy = 2;  /* a hack */
	set_map_viewport(map);
    }
    DGprintf("Displaying map view at %d,%d, size %d,%d, center %d,%d\n",
	     map->vx, map->vy, map->vw, map->vh, map->vcx, map->vcy);
    /* Compute top and bottom rows to be displayed. */
    /* Include rows that will only be partly drawn. */
    y1 = min(map->vy + map->vh, area.height - 1);
    y2 = map->vy;
    for (y = y1; y >= y2; --y) {
	/* Adjust the right and left bounds to fill the viewport as
	   much as possible, without going too far (the drawing code
	   will clip, but clipped drawing is still expensive). */
	/* could test by drawing viewport rect as lines... */
	adj = (y - map->vy) / 2;
	/* If the area doesn't wrap, then we might have to stop
	       drawing before we reach the edge of the viewport. */
	    /* (is this really reliable?) */
	    if (area.xwrap) {
		x1 = map->vx - adj - 1;
		x2 = map->vx + map->vw - adj + 1;
	    } else {
		int halfheight = area.height / 2;

		x1 = map->vx - adj - 1;
		x2 = map->vx + map->vw - adj + 1;  
		x1 = max(0, min(x1, area.width-1)); /* use limitx instead? */
		x2 = max(0, min(x2, area.width));
		if (x1 + y > area.width + halfheight)
		  continue; /* x1 = area.width + halfheight - y;  */
		if (x2 + y > area.width + halfheight)
		  x2 = area.width + halfheight - y;
		if (x1 + y < 1 * halfheight)
		  x1 = 1 * halfheight - y;
		if (x2 + y < 1 * halfheight)
		  continue; /* x2 = 1 * halfheight - 1 - y;  */
		
	    }
	    draw_row(side, map, x1, y, x2 - x1, FALSE);
	    /* (should test for input events here, would respond better) */
	}
}

/* Return the *predominant* color within the given hex. */

long
hex_color_at(side, map, x, y)
Side *side;
Map *map;
int x, y;
{
    if (!side->ui->monochrome && map->drawterrain && !map->drawhexpats) {
	/* account for snow? */
	return hex_color(side, x, y);
    } else {
	return side->ui->bgcolor;
    }
}

/* The basic map drawing routine does an entire row at a time, which yields
   order-of-magnitude speedups (!).  Actual nature of speedup is device
   specific, this routine concentrates on getting all the different
   display elements on the screen. */

draw_row(side, map, x0, y0, len, clearit)
Side *side;
Map *map;
int x0, y0, len, clearit;
{
    int empty = TRUE;
    char ch;
    int i = 0, x, sx, sy, b, c;

    if (len >= 1000) run_error("Display too wide!");

#if 0
    if (0 /* clearit */) {
	xform(side, map, x0, y0, &sx, &sy);
	{
	    int i;
	    char hexbuf[1000];

	    for (i = 0; i < len; ++i) {
		hexbuf[i] = 255; /* ? */
	    }
	    /* Draw a collection of bg-colored hexes. */
	    draw_uniform_segment(side, map->viewwin,
				 sx, sy, hexbuf, len, map->power,
				 /* need a generic fill "terrain type" */
				 side->ui->usewhat[map->power][0],
				 side->ui->bgcolor, TRUE);
	}
    }
#endif
    if (map->drawterrain) {
	draw_terrain_row(side, map, x0, y0, len);
    }
    if (elevations_defined() && map->drawelevations) {
	draw_elevations(side, map, x0, y0, len);
    }
    /* Draw sparse things on top of the basic row. */
    /* The relative ordering of these is quite important.  Note that
       each should be prepared to run independently also, since the
       other displays might have been turned off. */
    if (numbordtypes > 0 && map->drawterrain) {
	for_all_terrain_types(b) {
	    if (t_is_border(b)) {
		for (x = x0; x < x0 + len; ++x) {
		    draw_borders(side, map, x, y0, b);
		}
	    }
	}
    }
    /* This isn't in the right place to draw roads underneath */
    /* the units of the previous row... */
    if (numconntypes > 0 && map->drawterrain) {
	for_all_terrain_types(c) {
	    if (t_is_connection(c)) {
		for (x = x0; x < x0 + len; ++x) {
		    draw_connections(side, map, x, y0, c);
		}
	    }
	}
    }
    if (people_sides_defined() && map->drawpeople && map->power > 2) {
	for (x = x0; x < x0 + len; ++x) {
	    draw_people(side, map, x, y0);
	}
    }
    if (map->drawunits) {
	for (x = x0; x < x0 + len; ++x) {
	    draw_units(side, map, x, y0);
	}
    }
    if (map->drawnames) {
	for (x = x0; x < x0 + len; ++x) {
	    if (!inside_area(x, y0)) continue;
	    draw_legend(side, map, x, y0);
	}
    }
}

/* Return the color of a terrain icon overlaying a colored hex. */

terricon_color(side, x, y)
Side *side;
int x, y;
{
    return (side->ui->bgcolor);
}

enum whattouse
hex_style(side, x, y, power)
Side *side;
int x, y, power;
{
    return (terrain_visible(side, x, y) ?
	    side->ui->usewhat[power][terrain_at(x, y)] :
	    dontdraw);
}

Pixmap
hex_pattern(side, x, y, power)
Side *side;
int x, y, power;
{
    return None;
#if 0
    return (terrain_visible(side, x, y) ?
	    side->ui->terrpics[power][terrain_at(x, y)] :
	    None);
#endif
}

/* Return the color of the hex. (for color displays only) */

long
hex_color(side, x, y)
Side *side;
int x, y;
{
    long color;

    color = (terrain_visible(side, x, y) ?
	     side->ui->hexcolor[terrain_at(wrap(x), y)] :
	     side->ui->bgcolor);
    return color;
}

/* This interfaces higher-level drawing decisions to the rendition of
   individual pieces of display.  The rendering technique chosen depends
   on what the init code has decided is appropriate given what it found
   during init and what magnification the display is at.

   This routine is performance-critical;  any improvements will probably
   have a noticeable effect on the display. */

draw_terrain_row(side, map, x0, y0, len)
Side *side;
Map *map;
int x0, y0, len;
{
    int empty;
    int x, x1, t, sx, sy, w, h, p = map->power;
    int dogrid = map->drawgrid, dofill = map->drawhexpats;
    int i = 0;
    long color, segcolor;
    enum whattouse style, segstyle;
    Pixmap pat, segpat;
    Display *dpy = side->ui->dpy;
    GC gc = side->ui->icongc;
    Window win = map->viewwin;
    char hexbuf[1000];
    Image *timg;

    x1 = x0;
    segcolor = hex_color(side, x0, y0);
    segstyle = hex_style(side, x0, y0, p);
    segpat = hex_pattern(side, wrap(x0), y0, p);
    for (x = x0; x < x0 + len + 1; ++x) {
	t = terrain_at(wrap(x), y0);
	color = hex_color(side, wrap(x), y0);
	style = hex_style(side, wrap(x), y0, p);
	pat = hex_pattern(side, wrap(x), y0, p);
	/* Decide if the run is over and we need to dump some output. */
	if (x == x0+len
	    || color != segcolor
	    || style != segstyle
	    || pat != segpat
	    || segstyle == usepictures
	    || segstyle == usepolygons) {
	    /* don't draw anything that would match the window's bg */
	    if (segstyle != dontdraw && segcolor != side->ui->bgcolor) {
		xform(side, map, x1, y0, &sx, &sy);
		XSetForeground(dpy, gc, segcolor);
		/* no background? */
		switch (style) {
		  case useblocks:
		    /* could maybe adjust to go along middle of hex borders? */
		    XSetClipMask(dpy, gc, None);
		    w = i * hws[p];  h = hcs[p];
		    timg = best_image(side, &(side->ui->timages[t]), w, h);
		    if (timg != NULL && timg->mono != None && side->ui->monochrome) {
			XSetFillStyle(dpy, gc, FillTiled);
			XSetTile(dpy, gc, timg->mono);
			XSetTSOrigin(dpy, gc, 0, 0);
			XFillRectangle(dpy, win, gc, sx, sy, w, h);
			XSetFillStyle(dpy, gc, FillSolid);
		    } else {
			XFillRectangle(dpy, win, gc, sx, sy, w, h);
		    }
		    break;
		  case usepictures:
		    if (dofill) {
			XSetClipMask(dpy, gc, segpat);
		    } else if (dogrid) {
			XSetClipMask(dpy, gc, side->ui->bhexpics[p]);
		    } else {
			XSetClipMask(dpy, gc, side->ui->hexpics[p]);
		    }
		    XSetClipOrigin(dpy, gc, sx, sy);
		    XFillRectangle(dpy, win, gc, sx, sy, hws[p], hhs[p]);
		    break;
		  case usefontchars:
		    if (dofill) {
			hexbuf[i] = side->ui->terrchars[p][t];
		    } else if (dogrid) {
			hexbuf[i] = 'o';
		    } else {
			hexbuf[i] = 'O';
		    }
/*		    sy += side->ui->iconfont->max_bounds.ascent; */
		    XSetClipMask(dpy, gc, None);
		    XDrawString(dpy, win, gc, sx, sy, hexbuf, i);
		    break;
		  case usepolygons:
		    /* No clipping when we're drawing big polygons. */
		    XSetClipMask(dpy, gc, None);
		    timg = best_image(side, &(side->ui->timages[t]), w, h);
		    /* If we do patterns, use tiling within big hexes. */
		    if (timg != NULL && timg->mono != None && side->ui->monochrome) {
			XSetFillStyle(dpy, gc, FillTiled);
			XSetTile(dpy, gc, timg->mono);
			XSetTSOrigin(dpy, gc, 0, 0);
			draw_hex_polygon(side, win, gc, sx, sy, p, dogrid);
			XSetFillStyle(dpy, gc, FillSolid);
		    } else {
			draw_hex_polygon(side, win, gc, sx, sy, p, dogrid);
		    }
		}
	    }
	    i = 0;
	    x1 = x;
	    segcolor = color;
	    segstyle = style;
	    segpat = pat;
	}
	++i;
    }
}

draw_hex_icon(side, win, sx, sy, t, power, color)
Side *side;
Window win;
int sx, sy, t, power, color;
{
#if 0
    char terrbuf[2];
	
    terrbuf[0] = 0;
	draw_uniform_segment(side, side->ui->help,
			     sx, sy,
			     terrbuf, 1, power, 2, color, 0, 0);
#endif
}

/* Do the grody work of drawing very large polygons accurately. */

draw_hex_polygon(side, win, gc, sx, sy, power, dogrid)
Side *side;
Window win;
GC gc;
int sx, sy, power, dogrid;
{
    XPoint points[6];
    int hw = hws[power], hh = hhs[power], delt = (hhs[power] - hcs[power]);
    int ew = (dogrid ? 1 : 0);
	    
    points[0].x = sx + hw / 2;        points[0].y = sy;
    points[1].x = hw / 2 - ew;        points[1].y = delt - ew;
    points[2].x = 0;                  points[2].y = hh - 2 * delt - ew;
    points[3].x = 0 - (hw / 2 - ew);  points[3].y = delt - ew;
    points[4].x = 0 - (hw / 2 - ew);  points[4].y = 0 - (delt - ew);
    points[5].x = 0;                  points[5].y = 0 - (hh - 2 * delt - ew);
    XFillPolygon(side->ui->dpy, win, gc, points, 6, Convex, CoordModePrevious);
    XFlush(side->ui->dpy);
}

draw_elevations(side, map, x0, y0, len)
Side *side;
Map *map;
int x0, y0, len;
{
    int x, sx, sy;

    if (map->power < 3) return;

    for (x = x0; x < x0 + len + 1; ++x) {
	sprintf(spbuf, "%d", elev_at(x, y0));
	xform(side, map, x, y0, &sx, &sy);
	draw_text(side, map->viewwin, sx + 5, sy + map->uh / 2, spbuf, side->ui->fgcolor);
    }
}

/* Draw a single unit icon as appropriate.  This *also* has a bunch of
   details to worry about: centering of icon in hex, clearing a rectangular
   area for the icon, picking a color for the unit, using either a bitmap
   or font char, and adding a side emblem.

   Must also be careful not to draw black-on-black for units in space. */

draw_units(side, map, x, y)
Side *side;
Map *map;
int x, y;
{
    int tview, uview;
    int sx, sy, sw, sh, ucolor, hcolor, n, p = map->power, u, s;
    int terr = terrain_at(wrap(x), y);
    Side *side2;
    int num;
    int e = -1, b, c;
    Unit *unit;

    if (units_visible(side, x, y)) {
	num = num_units_at(x, y);
	if (num <= 0) {
	} else if (map->uw <= 16) {
	    xform(side, map, x, y, &sx, &sy);
	    /* Adjust to unit part of cell. */
	    sx += (map->hw - map->uw) / 2;  sy += (map->hh - map->uh) / 2;
	    if ((unit = unit_at(x, y)) != NULL) {
		draw_unit_image(side, map->viewwin, sx, sy, map->uw, map->uh,
				unit->type,
				side->ui->bgcolor, side->ui->bgcolor,
				side_number(unit->side));
		/* (should indicate presence of other units in some way) */
	    }
	} else {
	    for_all_stack(x, y, unit) {
		xform_unit(side, map, unit, &sx, &sy, &sw, &sh);
		draw_unit_and_occs(side, map, unit, sx, sy, sw, sh);
	    }
	}
    } else {
	if ((uview = unit_view(side, x, y)) != EMPTY) {
	    xform(side, map, x, y, &sx, &sy);
	    u = vtype(uview);
	    s = vside(uview);
	    sx += (map->hw - map->hh) / 2;  sy += (map->hh - map->uh) / 2;
	    draw_unit_image(side, map->viewwin, sx, sy, map->uw, map->uh,
			    u, side->ui->bgcolor, side->ui->bgcolor, s);
	}
    }
}

draw_unit_and_occs(side, map, unit, sx, sy, sw, sh)
Side *side;
Map *map;
Unit *unit;
int sx, sy, sw, sh;
{
    int u = unit->type, s = side_number(unit->side), sx2, sy2, sw2, sh2;
    Unit *occ;

    if (unit->occupant == NULL) {
	draw_unit_image(side, map->viewwin, sx, sy, sw, sh, u, s);
    } else {
	xform_occupant(side, map, unit, unit, sx, sy, sw, sh, &sx2, &sy2, &sw2, &sh2);
#if 0
		SetRect(&tmprect, sx, sy, sx+sw, sy+sh);
		InsetRect(&tmprect, sw2/2, sh2/2);
		FrameRect(&tmprect);
#endif
	draw_unit_image(side, map->window, sx2, sy2, sw2, sh2, u, side->ui->bgcolor, side->ui->bgcolor, s);
	for_all_occupants(unit, occ) {
	    xform_occupant(side, map, unit, occ, sx, sy, sw, sh, &sx2, &sy2, &sw2, &sh2);
	    draw_unit_and_occs(map, occ, sx2, sy2, sw2, sh2);
	}
    }
}

/* Indicate what kind of people are living in the given hex. */

draw_people(side, map, x, y)
Side *side;
Map *map;
int x, y;
{
#if 0
    viewdata view = side_view(side, wrap(x), y);
    int sx, sy;
    Side *side2;
    int pop, pcolor;
	
    pop = people_side_at(wrap(x), y);
    if (pop != NOBODY) {
	side2 = side_n(pop);
	/* need to work on rules for color of people emblem */
	pcolor = (allied_side(side, side2) ? side->ui->owncolor :
		  (enemy_side(side, side2) ? side->ui->enemycolor :
		   side->ui->neutcolor));
#if 0
	if (pcolor == side->ui->owncolor && 
	    (0  ||
	     side->ui->monochrome ||
	     (0 /* map->showmode == TERRICONS */)))
	  pcolor = side->ui->fgcolor;
#endif
	xform(side, map, x, y, &sx, &sy);
	/* compute pcolor inside drawing routine? */
	/* should this do a mask or bg also? */
	draw_people_emblem(side, map->viewwin, sx, sy, side2);
    }
#endif
}

/* Draw three borders of the given hex. */

/* (do we need another routine to repair all six borders?) */

draw_borders(side, map, x, y, b)
Side *side;
Map *map;
int x, y, b;
{
    int sx, sy;

    if (bwid[map->power] == 0) return;
    xform(side, map, x, y, &sx, &sy);
    if (border_at(x, y, WEST, b)
	&& seen_border(side, x, y, WEST)) {
	draw_border_line(side, map->viewwin,
			 sx, sy, WEST, map->power, side->ui->hexcolor[b]);
    }
    if (border_at(x, y, NW, b)
	&& seen_border(side, x, y, NW)) {
	draw_border_line(side, map->viewwin,
			 sx, sy, NW,   map->power, side->ui->hexcolor[b]);
    }
    if (border_at(x, y, NE, b)
	&& seen_border(side, x, y, NE)) {
	draw_border_line(side, map->viewwin,
			 sx, sy, NE,   map->power, side->ui->hexcolor[b]);
    }
}

/* Draw a single border on the given edge of the given hex. */

draw_border_line(side, win, sx, sy, dir, power, color)
Side *side;
Window win;
int sx, sy, dir, power, color;
{
    if (bsx[power][1] == 0) return;
    /* (Can this be skipped if color same as adj hexes?) */
    XSetForeground(side->ui->dpy, side->ui->varicongc, color);
    XSetClipMask(side->ui->dpy, side->ui->varicongc, None);
    XSetLineAttributes(side->ui->dpy, side->ui->varicongc,
		       bwid[power], LineSolid, CapButt, JoinMiter); 
    XDrawLine(side->ui->dpy, win, side->ui->varicongc,
	      sx + bsx[power][dir], sy + bsy[power][dir],
	      sx + bsx[power][dir+1], sy + bsy[power][dir+1]);
}

/* Draw three connections of the given hex. */

/* Actually this draws all six half-connections.  It also only draws the
   connection if the underlying terrain is different. */

draw_connections(side, map, x, y, c)
Side *side;
Map *map;
int x, y, c;
{
    int sx, sy, dir, t = terrain_at(x, y);

    if (bwid[map->power] == 0) return;
    if (!terrain_visible(side, x, y)) return;
    xform(side, map, x, y, &sx, &sy);
    for_all_directions(dir) {
	if (connection_at(x, y, dir, c)) {
	    draw_connection_line(side, map->viewwin,
				 sx, sy, dir,
				 map->power, side->ui->hexcolor[c]);
	}
    }
}

/* Draw a single line representing connection terrain.  It goes from
   the center of the hex right to the edge. */

draw_connection_line(side, win, sx, sy, dir, power, color)
Side *side;
Window win;
int sx, sy, dir, power, color;
{
    int cx = hws[power] / 2, cy = hhs[power] / 2;

    if (lsx[power][0] == 0) return;
    XSetForeground(side->ui->dpy, side->ui->varicongc, color);
    XSetClipMask(side->ui->dpy, side->ui->varicongc, None);
    XSetLineAttributes(side->ui->dpy, side->ui->varicongc,
		       bwid[power], LineSolid, CapButt, JoinMiter); 
    XDrawLine(side->ui->dpy, win, side->ui->varicongc,
	      sx + cx, sy + cy,
	      sx + cx + lsx[power][dir], sy + cy + lsy[power][dir]);
}

/* Draw any text that should be associated with this hex. */

/* (could precompute what the string will lap over and move or truncate str),
   should be deterministic for each mag, so redraw doesn't scramble */

/* do geofeatures, label at a hex with nothing else, and declared as the
   feature's "center" */

/* Black/white text should be consistent for each period, use mask only
   to fix difficulties. */

draw_legend(side, map, x, y)
Side *side;
Map *map;
int x, y;
{
    int hasdark = FALSE, haslight = FALSE, maskit = FALSE;
    int xx, numhexes = 1, color, sx, sy;
    char legend[100];
    Unit *unit;

    if (!inside_area(x, y)) return;
    if (!terrain_visible(side, x, y)) return;
    /* Draw a unit's name or number. */
    /* (This will only do top unit in hex, what about others?) */
    if (map->drawunits && units_visible(side, x, y)) {
	if ((unit = unit_at(x, y)) != NULL) {
	    if (unit->name != NULL) {
		strcpy(legend, unit->name);
	    } else if (unit->number > 0) {
		sprintf(legend, "%d%s",
			unit->number, ordinal_suffix(unit->number));
	    } else {
		/* Turns out we have nothing to make a legend with. */
		return;
	    }
	    /* Look at all the hexes that the legend will overlap. */
	    numhexes = (strlen(legend) * 8) / hws[map->power];
	    for (xx = x; xx < x + numhexes; ++xx) {
		color = hex_color_at(side, map, xx, y);
		if (low_contrast(side, color, side->ui->fgcolor)) {
		    haslight = TRUE;
		}
		if (low_contrast(side, color, side->ui->bgcolor)) {
		    hasdark = TRUE;
		}
		/* (should also account for borders and connections?) */
		/* (should also see which other hexes have legends, decide
		   how to move this one) */
	    }
	    /* if any unexplored stuff, use a mask too */
	    /* some periods might be better with all white text */
	    color = side->ui->bgcolor;
	    if (hasdark) maskit = TRUE;
	    /* (should also be done for any "noisy" background) */
	    if (side->ui->monochrome) maskit = TRUE;
	    xform(side, map, x, y, &sx, &sy);
	    draw_legend_text(side, map->viewwin,
			     /* or derive from unit's loc within hex... */
			     sx + map->hw/2 + map->uw/2 - 1,
			     sy + min(map->hh/2 - map->uh/2 + 1, map->hh - 1),
			     map->power, legend, color, maskit);
	}
    } else {
	/* feature object should specify legend's position */
	extern GeoFeature *feature_at();
	GeoFeature *feature;

	feature = feature_at(x, y);
	if (feature) {
	    if (feature->size == 1) {
		int pixlen;

		strcpy(legend, feature_name_at(x, y));
		pixlen = strlen(legend) * 8;
		xform(side, map, x, y, &sx, &sy);
		draw_legend_text(side, map->viewwin,
				 sx + 1 + (pixlen > map->hw ? 2 :
					   (map->hw/2 - pixlen / 2)),
				 sy - 6 + map->hh / 2,
				 map->power, legend, side->ui->fgcolor, TRUE);
	    }
	}
    }
}

/* Map legends are stencils usually. */

draw_legend_text(side, win, sx, sy, power, str, color, maskit)
Side *side;
Window win;
int sx, sy, power, color, maskit;
char *str;
{
    sy += (side->ui->ulegendfonts[power][0])->max_bounds.ascent;
    XSetFont(side->ui->dpy, side->ui->ltextgc,
	     (side->ui->ulegendfonts[power][0])->fid);
    XSetForeground(side->ui->dpy, side->ui->ltextgc, color);
    if (maskit) {
	XSetBackground(side->ui->dpy, side->ui->ltextgc, 
		       (color == side->ui->bgcolor ? side->ui->fgcolor :
			side->ui->bgcolor));
	XDrawImageString(side->ui->dpy, win, side->ui->ltextgc,
			 sx, sy, str, strlen(str));
    } else {
	XDrawString(side->ui->dpy, win, side->ui->ltextgc,
		    sx, sy, str, strlen(str));
    }
}

/* Display the control panel.  This is highly device-specific, so this
   routine just passes control through. */

/* (should flush or make worthwhile) */

show_controls(side, map)
Side *side;
Map *map;
{
    draw_controls(side, map);
}

/* Do this to change the boxes without disturbing anything else. */

update_submap_boxes(side, map)
Side *side;
Map *map;
{
    undraw_submap_boxes(side, map);
    draw_submap_boxes(side, map);
}

draw_submap_boxes(side, map)
Side *side;
Map *map;
{
    int mm2, sx1, sy1, sx2, sy2;
    Map *submap;

    for_all_maps(side, submap) {
	if (submap->supermap == map) {
	    mm2 = hws[map->power] / 2;
	    xform(side, map, submap->vx, submap->vy, &sx1, &sy1);
	    xform(side, map,
		  submap->vx + submap->vw - submap->vh / 2,
		  submap->vy + submap->vh,
		  &sx2, &sy2);
	    sx1 += mm2;  sy1 -= mm2;  sx2 += mm2;  sy2 += mm2;
	    invert_box(side, map->viewwin, sx1, sy1, sx2, sy2, mm2);
#if 0
	    /* Draw a second box if we're wrapped around. */
	    invert_box(side, map, submap->vcx - area.width, submap->vcy);
#endif
	}
    }
}

undraw_submap_boxes(side, map)
Side *side;
Map *map;
{
    int mm2, sx1, sy1, sx2, sy2;
    Map *submap;

    for_all_maps(side, submap) {
	if (submap->supermap == map) {
	    mm2 = hws[map->power] / 2;
	    xform(side, map, submap->vx, submap->vy, &sx1, &sy1);
	    xform(side, map,
		  submap->vx + submap->vw, submap->vy + submap->vh,
		  &sx2, &sy2);
	    sx1 += mm2;  sy1 -= mm2;  sx2 += mm2;  sy2 += mm2;
	    invert_box(side, map->viewwin, sx1, sy1, sx2, sy2, mm2);
	}
    }
}

/* Draw an inverted outline box on the given window. */
/* (isn't this a prim?) */

invert_box(side, win, sx1, sy1, sx2, sy2, mm2)
Side *side;
Window win;
int sx1, sy1, sx2, sy2, mm2;
{
    XSetFunction(side->ui->dpy, side->ui->gc, GXinvert);
    /* is this next call really necessary? */
    XSetLineAttributes(side->ui->dpy, side->ui->gc,
		       1, LineSolid, CapButt, JoinMiter);
    /* do drawrect instead? */
    XDrawLine(side->ui->dpy, win, side->ui->gc,
	      sx1, sy1, sx2, sy1);
    XDrawLine(side->ui->dpy, win, side->ui->gc,
	      sx2, sy1-1, sx2, sy2+1);
    XDrawLine(side->ui->dpy, win, side->ui->gc,
	      sx2, sy2, sx1, sy2);
    XDrawLine(side->ui->dpy, win, side->ui->gc,
	      sx1, sy2+1, sx1, sy1-1);
    XSetFunction(side->ui->dpy, side->ui->gc, GXcopy);
}

/* Cursor drawing also draws the unit in some other color if it's not the
   "top-level" unit in a hex. */

/* "color unders" here are not correct */

draw_cursor(side, map)
Side *side;
Map *map;
{
    int sx, sy;

    if (active_display(side)
	&& map->curunit
	&& alive(map->curunit)
	&& map->curunit->x >= 0) {
	xform(side, map, map->curunit->x, map->curunit->y, &sx, &sy);
	/* Maybe highlight the unit that the cursor is showing. */
	if (map->curunit != NULL && map->curunit->transport != NULL) {
	    if (side->ui->monochrome) {
		draw_hex_icon(side, map->viewwin,
			      sx, sy,
			      terrain_at(map->curunit->x, map->curunit->y),
			      map->power, side->ui->bgcolor);
		sx += (map->hw - map->hh) / 2;  sy += (map->hh - map->uh) / 2;
		draw_unit_image(side, map->viewwin,
				sx, sy, map->uw, map->uh,
				map->curunit->type,
				side->ui->fgcolor, side->ui->bgcolor,
				-1);
	    } else {
		sx += (map->hw - map->hh) / 2;  sy += (map->hh - map->uh) / 2;
		draw_unit_image(side, map->viewwin,
				sx, sy, map->uw, map->uh,
				map->curunit->type,
				side->ui->diffcolor, -1,
				-1);
	    }
	}
	/* Draw the cursor icon proper. */
	draw_cursor_icon(side, map, sx, sy);
    }
}

/* Get rid of cursor by redrawing the hex. */
/* (Overkill, should only do for particular map) */

erase_cursor(side, map)
Side *side;
Map *map;
{
    if (map->curunit) {
	int lastx = map->curunit->x, lasty = map->curunit->y;

	map->curunit = NULL;
	draw_row(side, map, lastx, lasty, 1, TRUE);
    }
}

/* Draw the scrollbars for the given map. */

/* These should be made a little prettier, also add extra gizmos. */

draw_scrollbars(side, map)
Side *side;
Map *map;
{
    int hpos, hbox, vpos, vbox;

    XSetForeground(side->ui->dpy, side->ui->gc, side->ui->fgcolor);
    XClearWindow(side->ui->dpy, map->hscrollbar);
    if (map->vw < area.width) {
	hpos = (map->vx * map->pxw) / area.width;
	hbox = (map->vw * map->pxw) / area.width;
	XFillRectangle(side->ui->dpy, map->hscrollbar, side->ui->gc,
		       hpos, 1, hbox, sbarwid - 2);
	/* other half if we're wrapping in x. */
	if (area.xwrap) {
	    XFillRectangle(side->ui->dpy, map->hscrollbar, side->ui->gc,
			   hpos - map->pxw, 1, hbox, sbarwid-2);
	}
    }
    XClearWindow(side->ui->dpy, map->vscrollbar);
    if (map->vh < area.height) {
	vpos = (map->vy * map->pxh) / area.height;
	vbox = (map->vh * map->pxh) / area.height;
	XFillRectangle(side->ui->dpy, map->vscrollbar, side->ui->gc,
		       1, map->pxh - vpos - vbox, sbarwid - 2, vbox);
    }
}

/* The cursor icon highlights a unit of interest - should be called a
   "selection highlight" or something similar. */

draw_cursor_icon(side, map, sx, sy)
Side *side;
Map *map;
int sx, sy;
{
    Display *dpy = side->ui->dpy;
    GC gc = side->ui->unitgc;

    /* should use some flags here */
    if (map->power == 4) {
	XSetClipOrigin(dpy, gc, sx, sy);
	XSetForeground(dpy, gc, side->ui->bgcolor);
	XSetClipMask(dpy, gc, side->ui->boxmask);
	XFillRectangle(dpy, map->viewwin, gc, sx, sy, map->hw, map->hh);
	XSetForeground(dpy, gc, side->ui->fgcolor);
	XSetClipMask(dpy, gc, side->ui->boxcurs);
	XFillRectangle(dpy, map->viewwin, gc, sx, sy, map->hw, map->hh);
    } else {
	int dx = (map->hw - map->uw) / 2, dy = (map->hh - map->uh) / 2;

	XSetClipMask(dpy, gc, None);
	XSetForeground(dpy, gc, side->ui->bgcolor);
	XDrawRectangle(dpy, map->viewwin, gc,
		       sx+dx, sy+dy, map->uw, map->uh);
	XSetForeground(dpy, gc, side->ui->fgcolor);
	XDrawRectangle(dpy, map->viewwin, gc, sx+dx, sy+dy, map->uw - 1, map->uh - 1);
    }
}

/* Draw something representing the population in the given hex.  Their side
   emblem alone is a good approximation. */

draw_people_emblem(side, win, sx, sy, power, side2)
Side *side, *side2;
Window win;
int power, sx, sy;
{
}

/* Draw a splat visible to both sides at a given location.  Several splats
   available, depending on the seriousness of the hit.  Make an extra-flashy
   display when The Bomb goes off.  Because of the time delays involved, we
   have to update both sides' displays more or less simultaneously.  */

draw_blast(unit, es, hit)
Unit *unit;
Side *es;
int hit;
{
#if 0
    int ux = unit->x, uy = unit->y, sx, sy, i, btype;
    Side *us = unit->side, *side;
    Map *map;

    if (0 /* hit >= period_nukehit() */) {
#if 0
	/* or all sides? */
	if (active_display(us)) invert_whole_map(us);
	if (active_display(es)) invert_whole_map(es);
	sleep(1);
	/* may need a time delay if X is too speedy */
	if (active_display(us)) invert_whole_map(us);
	if (active_display(es)) invert_whole_map(es);
	for (i = 0; i < 4; ++i) {
	    if (active_display(us)) draw_mushroom(us, ux, uy, i);
	    if (active_display(es)) draw_mushroom(es, ux, uy, i);
	    if (i != 2 && (active_display(us) || active_display(es))) sleep(1);
	}
	if (active_display(us) || active_display(es)) sleep(1);
#endif
    } else {
      int color;
      char buf[1];

	btype = ((hit >= unit->hp) ? 2 : ((hit > 0) ? 1 : 0));
	for_all_sides(side) {
	    if (active_display(side) && cover(side, ux, uy) > 0) {
		for_all_maps(side, map) {
		    if (map->power >= 4) { /* what other tests? */
			xform(side, map, ux, uy, &sx, &sy);
			/* should be able to draw different size blasts
			   for each map zoom */
			color = (side == us ? side->ui->enemycolor :
				 (side == es ? side->ui->owncolor :
				  side->ui->neutcolor));
			XSetForeground(side->ui->dpy, side->ui->varicongc, color);
			if (0 /* ever worthwhile to use font? */) {

			  buf[0] = (btype == 0 ? 'b' : (btype == 1 ? 'c' : 'd'));
			  sy += side->ui->iconfont->max_bounds.ascent;
			  XDrawString(side->ui->dpy, map->viewwin, side->ui->varicongc,
				      sx, sy, buf, 1);
			} else {
			  XSetClipMask(side->ui->dpy, side->ui->varicongc,
				       side->ui->hitpics[btype]);
			  XSetClipOrigin(side->ui->dpy, side->ui->varicongc,
					 sx, sy);
			  XFillRectangle(side->ui->dpy, map->viewwin, side->ui->varicongc,
					 sx, sy, 20, 22);
			}
			flush_output(side);
		      } else {
			/* just flicker the screen a little? */
		      }
		  }
	      }
	  }
    }
#endif
}

/* Flash the player's screen in an unmistakable way. */

invert_whole_map(side)
Side *side;
{
#if 0
    int sw = side->ui->vw * side->ui->hw, sh = side->ui->vh * side->ui->hh;

    XFillRectangle(side->ui->dpy, side->ui->front->viewwin, side->ui->flashgc, 0, 0, sw, sh);
    flush_output(side);
#endif
}

/* Draw just one of the mushroom cloud shapes. */

draw_mushroom(side, map, x, y, i)
Side *side;
Map *map;
int x, y, i;
{
#if 0
    int sx, sy;
    int color;

    color = ((side->ui->monochrome || i == 3) ? side->ui->fgcolor : side->ui->bgcolor);
    xform(side, map, x, y, &sx, &sy);
    XSetForeground(side->ui->dpy, side->ui->unitgc, color);
    XSetClipMask(side->ui->dpy, side->ui->unitgc, side->ui->bombpics[i]);
    XSetClipOrigin(side->ui->dpy, side->ui->unitgc, sx-BH/4, sy-BH/2);
    XFillRectangle(side->ui->dpy, side->ui->front->viewwin, side->ui->unitgc, sx-BH/4, sy-BH/2, BH, BH);
    flush_output(side);
#endif
}

/* Draw the control panel associated with the given map.  It is organized
   as two columns, but a tool can extend over both if so specified. */

draw_controls(side, map)
Side *side;
Map *map;
{
    int i, tx, ty, tw, color;
    int adjx, adjy = (conhgt - 16) / 2;
    int fg = side->ui->fgcolor, bg = side->ui->bgcolor;
    Pixmap pic;

    XClearWindow(side->ui->dpy, map->controls);
    for (i = 0; i < NUMCONTROLS; ++i) {
	ty = (i / 2) * conhgt;
	/* Decide how to draw the control. */
	if (controltypes[i].name != NULL && map->controlbs[i].visible) {
	    /* Draw the top dividing line. */
	    XSetForeground(side->ui->dpy, side->ui->gc, fg);
	    XDrawLine(side->ui->dpy, map->controls, side->ui->gc,
		      0, ty, conwid, ty);
	    /* Decide if we're doing two tools in this row or just one. */
	    tw = conwid / 2;
	    if ((i % 2) == 0) {
		/* First (or only) tool of a side-by-side pair. */
		tx = 0;
		if (controltypes[i+1].name == NULL) {
		    /* Only one in this row; will center its picture. */
		    tw = conwid;
		}
	    } else {
		/* Second tool of a side-by-side pair. */
		tx = conwid / 2;
		/* Draw a vertical dividing line between. */
		XDrawLine(side->ui->dpy, map->controls, side->ui->gc,
			  tx, ty, tx, ty + conhgt);
	    }
	    adjx = (tw - 16) / 2;
	    /* Draw the background of the tool's box. */
	    XSetForeground(side->ui->dpy, side->ui->gc, bg);
	    XFillRectangle(side->ui->dpy, map->controls, side->ui->gc,
			   tx + 1, ty + 1, tw - 1, conhgt - 1);
	    /* Draw its selection state. */
	    XSetForeground(side->ui->dpy, side->ui->gc,
			   map->controlbs[i].selected ? fg : bg);
	    XFillRectangle(side->ui->dpy, map->controls, side->ui->gc,
			   tx + 1, ty + 2, tw - 2, conhgt - 3);
	    /* Draw either text or a picture in the box. */
	    if (map->controlbs[i].text != NULL) {
		XSetForeground(side->ui->dpy, side->ui->gc,
			       map->controlbs[i].selected ? bg : fg);
		XSetBackground(side->ui->dpy, side->ui->gc,
			       map->controlbs[i].selected ? fg : bg);
		/* Only do the first 5 chars. */
		XDrawString(side->ui->dpy, map->controls, side->ui->gc, 
			    tx + margin, ty + adjy + 10,
			    map->controlbs[i].text, 5);
	    } else if (map->controlbs[i].drawer != NULL) {
		(*(map->controlbs[i].drawer))(side, map, tx + adjx, ty + adjy);
	    } else {
		/* Now draw the tool picture.  This draws a solid color
		   box, if no picture has been defined. */
		pic = side->ui->toolpics[i];
		/* Use a instance-specific picture if it has been set up. */
		if (map->controlbs[i].pic != None) {
		    pic = map->controlbs[i].pic;
		}
		XSetClipMask(side->ui->dpy, side->ui->unitgc, pic);
		XSetClipOrigin(side->ui->dpy, side->ui->unitgc,
			       tx + adjx, ty + adjy);
		/* Pick a color. */
		if (map->controlbs[i].color >= 0) {
		    color = map->controlbs[i].color;
		} else if (map->controlbs[i].selected) {
		    /* If selected, box already drawn in fg color. */
		    color = bg;
		} else if (map->controlbs[i].enabled) {
		    color = fg;
		} else {  /* visible but disabled */
		    /* (should be a dimmer color) */
		    color = side->ui->graycolor;
		}
		XSetForeground(side->ui->dpy, side->ui->unitgc, color);
		XFillRectangle(side->ui->dpy, map->controls, side->ui->unitgc,
			       tx + adjx, ty + adjy, 16, 16);
	    }
	    /* Draw the bottom dividing line. */
	    /* (this is usually redundant...) */
	    XSetForeground(side->ui->dpy, side->ui->gc, fg);
	    XDrawLine(side->ui->dpy, map->controls, side->ui->gc,
		      0, ty + conhgt, conwid, ty + conhgt);
	}
    }
    /* Fix the mess we made. */
    XSetForeground(side->ui->dpy, side->ui->gc, fg);
    XSetBackground(side->ui->dpy, side->ui->gc, bg);
}

/* Return the actual dimensions of the unit icon that will be used at the
   given magnification. */

unit_icon_width(side, power, u)
Side *side;
int power, u;
{
    return side->ui->unitw[power][u];
}

unit_icon_height(side, power, u)
Side *side;
int power, u;
{
    return side->ui->unith[power][u];
}

/* Splash a unit image (either bitmap or font char) onto some window. */

/* (Add support for full-color icons?) */

draw_unit_image(side, win, sx, sy, sw, sh, u, fg, bg, s2)
Side *side;
Window win;
int sx, sy, sw, sh, u, fg, bg, s2;
{
    char buf[1];
    int sx2, sy2, ew, eh;
    Image *uimg;
    Display *dpy = side->ui->dpy;
    GC gc = side->ui->unitgc;

    uimg = best_image(side, &(side->ui->uimages[u]), sw, sh);
    if (uimg) {
	sx += (sw - uimg->w) / 2;  sy += (sh - uimg->h) / 2;
	sw = uimg->w;  sh = uimg->h;
	if (uimg->mono != None) {
	    XSetClipOrigin(dpy, gc, sx, sy);
	    XSetForeground(dpy, gc, side->ui->fgcolor);
	    if (uimg->mask != None) {
		XSetClipMask(dpy, gc, uimg->mask);
	    } else {
/*		XSetClipMask(dpy, gc, None);  */
		XSetClipMask(dpy, gc, uimg->mono); /* for now */
	    }
	    XFillRectangle(dpy, win, gc, sx, sy, uimg->w, uimg->h);
	    XSetForeground(dpy, gc, side->ui->bgcolor);
	    XSetClipMask(dpy, gc, uimg->mono);
	    XFillRectangle(dpy, win, gc, sx, sy, uimg->w, uimg->h);
	} else if (uimg->monochar) {
	    XSetForeground(dpy, gc, side->ui->fgcolor);
	    buf[0] = uimg->monochar;
/*	    sy += uimg->font->max_bounds.ascent;  */
	    /* need to set font as the one in use here? */
	    XDrawString(dpy, win, gc, sx, sy, buf, 1);
	} else {
	    XSetForeground(dpy, gc, side->ui->fgcolor);
	    XSetClipMask(dpy, gc, None);
	    XFillRectangle(dpy, win, gc, sx, sy, uimg->w, uimg->h);
	}
    } else {
	/* a bad error? */
	XSetClipOrigin(dpy, gc, sx, sy);
	/* Pretty desperate, or else we're at a scale that doesn't matter. */
	/* should actually clip to hex boundaries, maybe even inside
	   border terrain if necessary? */
	XSetClipMask(dpy, gc, None);
	/* Draw an edge. */
	if (sw >= 4) {
	    XSetForeground(dpy, gc, side->ui->fgcolor);
	    XFillRectangle(dpy, win, gc, sx, sy, sw, sh);
	}
	/* Draw a filled box inside the edge. */
	XSetForeground(dpy, gc, side->ui->bgcolor);
	XFillRectangle(dpy, win, gc, sx + 1, sy + 1, sw - 2, sh - 2);
	/* Might be cool to draw unit type name or char */
    }
    if (between(0, s2, numsides)) {
	ew = min(sw, max(8, sw / 4));  eh = min(sh, max(8, sh / 4));
	draw_side_emblem(side, win, sx + sw - ew, sy, ew, eh, side_n(s2));
    }
}

/* Draw an emblem identifying the given side.  If a side does not have a
   distinguishing emblem, fall back on some defaults. */

/* This needs to do something with masks or backgrounds or something... */

draw_side_emblem(side, win, ex, ey, ew, eh, side2)
Side *side, *side2;
Window win;
int ex, ey, ew, eh;
{
    int s2 = side_number(side2);
    long fgcolor, bgcolor;
    Image *eimg;
    Pixmap epic;
    Display *dpy = side->ui->dpy;

    eimg = best_image(side, &(side->ui->eimages[s2]), ew, eh);
    /* Decide on the colors to use with the emblem. */
    if (side->ui->numcolors[s2] > 0) {
	fgcolor = side->ui->colors[s2][0];
    } else {
	fgcolor = side->ui->fgcolor;
    }
    if (side->ui->numcolors[s2] > 1) {
	bgcolor = side->ui->colors[s2][1];
    } else {
	bgcolor = side->ui->bgcolor;
    }
#if 0
    /* Draw a mask or background for the emblem. */
    if (0 /* mask defined */) {
	/* Assume mask guarantees definition and contrast. */
	XSetForeground(dpy, side->ui->varicongc, bgcolor);
	XSetClipMask(dpy, side->ui->varicongc, None);
	XFillRectangle(dpy, win, side->ui->varicongc,
		       ex, ey, ew, eh);
    } else {
	/* Without a mask, we need to draw a box (use fgcolor for now). */
/*	bgcolor = (low_contrast(side, fgcolor, side->ui->bgcolor) ?
		   side->ui->fgcolor : side->ui->bgcolor); */
	XSetForeground(dpy, side->ui->varicongc, side->ui->fgcolor);
	XSetClipMask(dpy, side->ui->varicongc, None);
	/* Not good to expand by two pixels... */
	XFillRectangle(dpy, win, side->ui->varicongc, ex, ey, ew, eh);
    }
#endif
    /* Now draw the emblem proper.  Note that we should *always* draw
       it in its declared color. */
    XSetForeground(dpy, side->ui->varicongc, fgcolor);
    if (eimg) {
	if (eimg->mono != None) {
	    XSetClipMask(dpy, side->ui->varicongc, eimg->mono);
	    XSetClipOrigin(dpy, side->ui->varicongc, ex, ey);
	    XFillRectangle(dpy, win, side->ui->varicongc,
			   ex, ey, ew, eh);
	    return;
	}
    }
    /* Default is to draw a blob. */
    if (ew > 2 && eh > 2) {
	XFillRectangle(dpy, win, side->ui->varicongc, ex+1, ey+1, ew-2, eh-2);
    }
}

/* Also draw a big 'X' through the losing side's maps. */

/* (This should draw fatter lines and invert rather than overwrite.) */

draw_map_scratchout(side, map)
Side *side;
Map *map;
{
    XDrawLine(side->ui->dpy, map->viewwin, side->ui->gc,
	      0, 0, map->pxw, map->pxh);
    XDrawLine(side->ui->dpy, map->viewwin, side->ui->gc,
	      map->pxw, 0, 0, map->pxh);
}

/* Beep the beeper! */

beep(side)
Side *side;
{
    XBell(side->ui->dpy, side->ui->screen);
}

/* Little routines to pop up the help window and make it go away again */
/* They only get called when display is in use. */

reveal_help(side)
Side *side;
{
    if (side->ui->curhelpnode == NULL) {
	side->ui->curhelpnode = firsthelpnode;
    }
    /* Do we really need to wait until after popup? */
/*    update_help_content(side);  */
}

/* Generate a legend for the pictures of terrain and units. */

/* (should try to use "natural size" rather than power = 4) */

legend_help(side)
Side *side;
{
#if 0
    int t, u, dir, tcol = 3 * margin + hws[4];
    int spacing = hhs[4] + 4 * margin;
    int offset = hhs[4]/4, second = 20*10;

    second += 2 * (margin + hws[4]);
    tcol += 2 * (margin + hws[4]);
    /* List all the terrain patterns/colors. */
    /* (All these should have bgs or be outlined or something) */
    for_all_terrain_types(t) {
	if (t_is_hex(t)) {
#if 0
	    draw_hex_icon(side, side->ui->help,
			  margin, t * spacing, t, 4,
			  (side->ui->monochrome ? side->ui->fgcolor :
			   side->ui->hexcolor[t]));
#endif
	}
	if (t_is_border(t)) {
	    for_all_directions(dir) {
#if 0
		draw_border_line(side, side->ui->help,
				 2 * margin + hws[4],
				 t * spacing, dir,
				 4, side->ui->hexcolor[t]);
#endif
	    }
	}
	if (t_is_connection(t)) {
	    for_all_directions(dir) {
#if 0
		draw_connection_line(side, side->ui->help,
				     3 * margin + 2 * hws[4],
				     t * spacing, dir,
				     4, side->ui->hexcolor[t]);
#endif
	    }
	}
#if 0
	draw_fg_text(side, side->ui->help,
		     tcol, t * spacing + offset, t_type_name(t));
    }
#endif
    /* List all the unit types. */
    for_all_unit_types(u) {
#if 0
	/* (Need to do at preferred magnification) */
	draw_unit_icon(side, side->ui->help,
		       second, u * spacing, u, 4,
		       side->ui->fgcolor, side->ui->bgcolor);
	sprintf(tmpbuf, "%s (%s)", u_type_name(u), u_short_name(u));
	draw_fg_text(side, side->ui->help,
		     tcol + second, u * spacing + offset, tmpbuf);
#endif
    }
#endif
}

Image *
best_image(side, imf, w, h)
Side *side;
ImageFamily *imf;
int w, h;
{
    int i;
    Image *img, *fallback = NULL;

    for (i = 0; i < imf->numsizes; ++i) {
	img = &(imf->images[i]);
	if (w >= img->minw && h >= img->minh) {
	    if (!fallback || (img->w > fallback->w && img->h > fallback->h)) {
		fallback = img;
	    }
	}
	if (w <= img->maxw && h <= img->maxh) {
	    return img;
	}
    }
    return fallback;
}
