/* Copyright (c) 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact.  */

/* Xlib xconq map handling. */

#include "conq.h"
#include "xconq.h"

/* Create a generic map object. */

create_map(side, power, super)
Side *side;
int power;
Map *super;
{
    int x, y, r, sx, sy, i;
    long winfg, winbg;
    char *retain;
    Map *map = (Map *) xmalloc(sizeof(Map));
    XSetWindowAttributes winattrs;

    DGprintf("Creating map, mag power %d\n", power);
    set_map_power(map, power);
    /* (these should come from geometry spec) */
    sx = 10;  sy = 50;
    map->totalw = 600;  map->totalh = 500;
    subdivide_map(map);
    pick_a_focus(side, &x, &y);
    map->vcx = x;  map->vcy = y;
    set_map_viewport(map);
    /* Set default values for the display controls. */
    map->drawterrain = TRUE;
    map->drawgrid = TRUE;  /* should get from a resource */
    map->drawhexpats = side->ui->monochrome;
    map->drawunits = TRUE;
    map->drawnames = (map->power >= 3);
    map->drawpeople = TRUE;
    map->drawelevations = FALSE;
    for_all_material_types(r) map->drawresources[r] = FALSE;
    map->drawtemp = FALSE;
    map->drawweather = FALSE;
    map->curunit = NULL;
#ifdef DESIGNERS
    map->curttype = 0;
    map->curutype = 0;
    map->cursidenumber = 0;
#endif /* DESIGNERS */
    /* Newest map goes on the front of the list. */
    map->next = side->ui->maps;
    side->ui->maps = map;
    /* Make this be the front map until otherwise directed. */
    side->ui->front = map;
    winfg = side->ui->fgcolor;  winbg = side->ui->bgcolor;
    map->window = (Window)
      XCreateSimpleWindow(side->ui->dpy, side->ui->rootwin,
			  sx, sy, map->totalw, map->totalh,
			  1, winfg, winbg);
    XSelectInput(side->ui->dpy, map->window,
		 /* why only a subset of events? */
		 KeyPressMask|ExposureMask|StructureNotifyMask);
    map->msg = (Window)
      XCreateSimpleWindow(side->ui->dpy, map->window,
			  fudge, fudge,
			  map->msgw, map->msgh / 2,
			  1, winfg, winbg);
    XSelectInput(side->ui->dpy, map->msg, everything);
    map->promptwin = (Window)
      XCreateSimpleWindow(side->ui->dpy, map->window,
			  fudge, fudge + map->msgh / 2,
			  map->msgw, map->msgh / 2,
			  1, winfg, winbg);
    XSelectInput(side->ui->dpy, map->promptwin, everything);
    map->game = (Window)
      XCreateSimpleWindow(side->ui->dpy, map->window,
			  map->msgw + 2 * fudge, fudge,
			  map->totalw - map->msgw - 3 * fudge, map->msgh,
			  1, winfg, winbg);
    XSelectInput(side->ui->dpy, map->msg, everything);
    map->controls = (Window)
      XCreateSimpleWindow(side->ui->dpy, map->window,
			  fudge, map->msgh + 2 * fudge,
			  conwid, map->totalh - map->msgh - 3 * fudge,
			  1, winfg, winbg);
    XSelectInput(side->ui->dpy, map->controls, everything);
    map->infowin = (Window)
      XCreateSimpleWindow(side->ui->dpy, map->window,
			  conwid + 2 * fudge, map->msgh + 2 * fudge,
			  map->pxw + fudge + sbarwid, map->infoh,
			  1, winfg, winbg);
    XSelectInput(side->ui->dpy, map->infowin, everything);
    map->viewwin = (Window)
      XCreateSimpleWindow(side->ui->dpy, map->window,
			  conwid + 2 * fudge,
			  map->msgh + map->infoh + 3 * fudge,
			  map->pxw, map->pxh,
			  1, winfg, winbg);
    XSelectInput(side->ui->dpy, map->viewwin, everything);
    map->hscrollbar = (Window)
      XCreateSimpleWindow(side->ui->dpy, map->window,
			  conwid + 2 * fudge,
			  map->msgh + map->infoh + map->pxh + 4 * fudge,
			  map->pxw, sbarwid,
			  1, winfg, winbg);
    XSelectInput(side->ui->dpy, map->hscrollbar, everything);
    map->vscrollbar = (Window)
      XCreateSimpleWindow(side->ui->dpy, map->window,
			  conwid + map->pxw + 3 * fudge,
			  map->msgh + map->infoh + 3 * fudge,
			  sbarwid, map->pxh,
			  1, winfg, winbg);
    XSelectInput(side->ui->dpy, map->vscrollbar, everything);
    map->sides = (Window)
      XCreateSimpleWindow(side->ui->dpy, map->window,
			  map->leftw + fudge,
			  map->msgh + 2 * fudge,
			  map->list1w, map->list1h,
			  1, winfg, winbg);
    XSelectInput(side->ui->dpy, map->sides, everything);
    map->list = (Window)
      XCreateSimpleWindow(side->ui->dpy, map->window,
			  map->leftw + fudge,
			  map->msgh + map->list1h + 3 * fudge,
			  map->list2w, map->list2h,
			  1, winfg, winbg);
    XSelectInput(side->ui->dpy, map->list, everything);
    XStoreName(side->ui->dpy, map->window, "Xconq Map");
    config_map_controls(side, map);
    /* Have the view itself be a retained window, cuts down on redraws */
    if ((retain = XGetDefault(side->ui->dpy, PROGRAMNAME, "Retain")) == NULL
	|| (*retain != 'N' && *retain != 'n')
	/* strcasecmp(retain, "no") */) {
	winattrs.backing_store = 1;
	XChangeWindowAttributes(side->ui->dpy, map->viewwin,
				CWBackingStore, &winattrs);
    }
    /* Draw this map's bounding box on its supermap. */
    map->supermap = super;
}

/* Given the total width and height of a map window, calculate the sizes
   of the different subareas. */

subdivide_map(map)
Map *map;
{
    /* (should be smarter?) */
    map->msgw = map->totalw - 100;  map->msgh = 30;
    map->leftfrac = 70;
    map->leftw = (map->leftfrac * map->totalw) / 100 ;
    map->pxw = map->leftw - conwid - sbarwid - 3 * fudge;
    map->pxh = map->totalh - map->msgh - 123 - sbarwid - 5 * fudge;
    map->infow = map->pxw;  map->infoh = 123;
    map->list1frac = 30;
    map->listsh = map->totalh - map->msgh - 3 * fudge;
    map->list1w = map->totalw - map->leftw - 2 * fudge;
    map->list1h = (map->list1frac * map->listsh) / 100;
    map->list2w = map->list1w;
    map->list2h = map->listsh - map->list1h - fudge;
}

/* Set the state of all the regular control panel items. */

config_map_controls(side, map)
Side *side;
Map *map;
{
    int i;

    map->curtool = LOOKTOOL;
    map->curshifttool = UNITMOVETOOL;
    for (i = 0; i < NUMCONTROLS; ++i) {
	map->controlbs[i].visible = TRUE;
	map->controlbs[i].enabled = TRUE;
	map->controlbs[i].selected = FALSE;
	map->controlbs[i].text = NULL;
	map->controlbs[i].color = -1;
	map->controlbs[i].pic = None;
	map->controlbs[i].drawer = NULL;
    } 
    map->controlbs[LOOK].selected = TRUE;
    map->controlbs[SHOWTERR].selected = map->drawterrain;
    map->controlbs[SHOWTERR].text = "Terr";
    map->controlbs[SHOWPATS].selected = map->drawhexpats;
    map->controlbs[SHOWPATS].text = "Pats";
    map->controlbs[SHOWEDGES].selected = map->drawgrid;
    map->controlbs[SHOWEDGES].text = "Grid";
    map->controlbs[SHOWUNITS].selected = map->drawunits;
    map->controlbs[SHOWUNITS].text = "Units";
    map->controlbs[SHOWNAMES].selected = map->drawnames;
    map->controlbs[SHOWNAMES].text = "Names";
    map->controlbs[SHOWPEOPLE].selected = map->drawpeople;
    map->controlbs[SHOWPEOPLE].text = "People";
    map->controlbs[SHOWELEVS].selected = map->drawelevations;
    map->controlbs[SHOWELEVS].text = "Elevs";
    map->controlbs[SHOWTEMP].selected = map->drawtemp;
    map->controlbs[SHOWTEMP].text = "Temp";
#ifdef DESIGNERS
    config_designer_controls(side, map);
#endif /* DESIGNERS */
}

#ifdef DESIGNERS

draw_utype_in_control(side, map, sx, sy)
Side *side;
Map *map;
int sx, sy;
{
    draw_unit_image(side, map->viewwin, sx, sy, 16, 16,
		    map->curutype, side->ui->fgcolor, side->ui->bgcolor, -1);
}

/* Set/reset the states of the designer controls. */

config_designer_controls(side, map)
Side *side;
Map *map;
{
    int i;

    for (i = 0; i < NUMCONTROLS; ++i) {
	/* If a selected tool is designer-only, switch to another. */
	if (!side->designer
	    && controltypes[i].fordesigners
	    && controltypes[i].tool >= 0
	    && map->controlbs[i].selected) {
	    map->controlbs[i].selected = TRUE;
	    map->controlbs[LOOK].selected = TRUE;
	}
	if (controltypes[i].fordesigners) {
	    map->controlbs[i].visible = side->designer;
	    map->controlbs[i].enabled = side->designer;
	    /* Selection status should not be affected by default. */
	}
    }
    /* Make sure these have something reasonable in them. */
    /* need to indicate "no type" better, also fix for b/w */
    map->controlbs[CURTTYPE].color =
      side->ui->hexcolor[map->curttype];
    map->controlbs[CURUTYPE].drawer = draw_utype_in_control;
    if (map->controlbs[CURSIDE].text == NULL) {
	/* should use emblem instead? */
	sprintf(tmpbuf, "%s", side_desig(side_n(map->cursidenumber)));
	map->controlbs[CURSIDE].text = copy_string(tmpbuf);
    }
}
#endif /* DESIGNERS */

/* Given a map with already-set pixel size and desired viewport center, compute
   reasonable values for its size and position in world coordinates.
   This is really just caching, can be run at any time. */

set_map_viewport(map)
Map *map;
{
    /* Compute the width and height. */
    map->vw = min(area.width, map->pxw / map->hw);
    /* what about the hh - hch correction here? */
    map->vh = min(area.height, map->pxh / map->hch + 1);
    /* Compute some cached values that are used a lot (still?). */
    map->vw2 = map->vw / 2;  map->vh2 = map->vh / 2;
    /* Since we don't want to let the viewport stare into the void
       too much, the code here keeps the position limited to reasonable
       values, *except* that the left and right sides of a hexagon may be
       exposed, so other code has to make sure that nonsensical positions
       are not displayed. */
    if (between(0, map->vcx, area.width)) {
	map->vx = map->vcx - map->vw2 + map->vh2;
    } else {
	map->vx = area.width / 2;
    }
    if (between(0, map->vcy, area.height)) {
	map->vy = map->vcy - map->vh2;
    } else {
	map->vy = 1; 
    }
    /* Now push the viewport into a valid position. */
    if (map->vy < 0)
      map->vy = 0;
    if (map->vy > area.height - map->vh)
      map->vy = area.height - map->vh;
    if (area.xwrap) {
	if (map->vx < 0)
	  map->vx = (map->vx + 10 * area.width) % area.width;
	/* vx can't possibly get too large. */
    } else {
	/* No x wrapping. */
	if (map->vy < area.height / 2) {
	    /* Bottom half of hexagon. */
	    if (map->vx < area.height/2)
	      map->vx = area.height / 2;
	    if (map->vx > (area.width - map->vw))
	      map->vx = area.width - map->vw;
	} else {
	    /* Top half of hexagon. */
	}
    }
    DGprintf("Set map %x viewport to be %dx%d @ %d,%d (nom center %d,%d)\n",
	     map, map->vw, map->vh, map->vx, map->vy, map->vcx, map->vcy);
}

/* Change the view area of the given map to the given dimensions.  This is
   just the drawing area proper, doesn't have to account for the sizes of
   decorations. */

resize_map(side, map, w, h)
Side *side;
Map *map;
int w, h;
{
    unsigned int mask;
    XWindowChanges changes;
    
    if (map->totalw == w && map->totalh == h) return;
    DGprintf("Want to change from %dx%d to %dx%d pixel map window\n",
	     map->totalw, map->totalh, w, h);
    map->totalw = w;  map->totalh = h;
    /* Calculate the new sizes and positions of everything. */
    subdivide_map(map);
    /* Don't need to config everything, just what changes. */
    /* Do the "topline" windows. */
    mask = CWWidth;
    changes.width = map->msgw;
    XConfigureWindow(side->ui->dpy, map->msg, mask, &changes);
    mask = CWX;
    changes.x = map->msgw + 2 * fudge;
    XConfigureWindow(side->ui->dpy, map->game, mask, &changes);
    /* Do the control panel. */
    mask = CWHeight;
    changes.height = map->totalh - map->msgh - 2 * fudge;
    XConfigureWindow(side->ui->dpy, map->controls, mask, &changes);
    /* Do the info and view windows. */
    mask = CWWidth;
    changes.width = map->pxw;
    XConfigureWindow(side->ui->dpy, map->infowin, mask, &changes);
    mask = CWWidth|CWHeight;
    changes.width = map->pxw;  changes.height = map->pxh;
    XConfigureWindow(side->ui->dpy, map->viewwin, mask, &changes);
    /* Do the scrollbars. */
    mask = CWY|CWWidth;
    changes.y = map->msgh + map->infoh + map->pxh + 4 * fudge;
    changes.width = map->pxw;
    XConfigureWindow(side->ui->dpy, map->hscrollbar, mask, &changes);
    mask = CWX|CWHeight;
    changes.x = conwid + map->pxw + 3 * fudge;
    changes.height = map->pxh;
    XConfigureWindow(side->ui->dpy, map->vscrollbar, mask, &changes);
    /* Do the lists. */
    mask = CWX|CWWidth|CWHeight;
    changes.x = map->leftw + fudge;
    changes.width = map->list1w;  changes.height = map->list1h;
    XConfigureWindow(side->ui->dpy, map->sides, mask, &changes);
    mask = CWX|CWY|CWWidth|CWHeight;
    changes.x = map->leftw + fudge;
    changes.y = map->msgh + map->list1h + 3 * fudge;
    changes.width = map->list1w;  changes.height = map->list1h;
    XConfigureWindow(side->ui->dpy, map->list, mask, &changes);
    /* Recalculate and redisplay contents. */
    set_map_viewport(map);
    show_map(side, map);
}

destroy_map(side, map)
Side *side;
Map *map;
{
    /* find map and remove it from the list of maps */
    /* also put some other map in front if this one was the front one */
}
